#############################################################################
# psl2_plugin.tcl by The Mad Monk.
# PgcEdit plugin to make the creation of DVD Decrypter's PSL2 files easier.
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "PSL2 (ARccOS)" menu will be added in the Plugins menu.
# You need PgcEdit 0.6.1 or more, DVD Decrypter 3.5.4.0, and optionally
# VobBlanker 2.0.1.0 or more to use this plugin.
#############################################################################
# Version history: Use the plugins menu
#############################################################################


# Add Plugins -> PSL2 submenu
if {! [winfo exists .mainmenu.plugins.psl2]} {
	.mainmenu.plugins add cascade -label "PSL2 (ARccOS)" -menu .mainmenu.plugins.psl2
	menu .mainmenu.plugins.psl2
	.mainmenu.plugins.psl2 add command -label "Wizard" \
		-command {::plugins::psl2::wizard}

	.mainmenu.plugins.psl2 add cascade -label "Advanced mode" -menu .mainmenu.plugins.psl2.adv
	menu .mainmenu.plugins.psl2.adv
		.mainmenu.plugins.psl2.adv add command -label "Quick Open DVD" \
			-command {::plugins::psl2::open_dvd_wo_messages true}
		.mainmenu.plugins.psl2.adv add command -label "Find protected titles" \
			-command {::plugins::psl2::find_protected_titles false}
		.mainmenu.plugins.psl2.adv add separator
		.mainmenu.plugins.psl2.adv add command -label "Before rip: Generate PSL2 file" \
			-command {::plugins::psl2::main_gui false}
		.mainmenu.plugins.psl2.adv add command -label "After rip: Remove unreferenced Titles" \
			-command {::plugins::psl2::remove_unref_titles}
		.mainmenu.plugins.psl2.adv add command -label "After rip: Remove crap" \
			-command {::plugins::psl2::main_gui true}
		.mainmenu.plugins.psl2.adv add command -label "Finally: Cleanup VOBs with FixVTS" \
			-command {::plugins::psl2::fix_processed_vobs}
		.mainmenu.plugins.psl2.adv add command -label "Or: Cleanup VOBs with VobBlanker" \
			-command {::plugins::psl2::clean_processed_vobs}

	.mainmenu.plugins.psl2 add cascade -label "Tools" -menu .mainmenu.plugins.psl2.tools
	menu .mainmenu.plugins.psl2.tools
		.mainmenu.plugins.psl2.tools add command -label "Dump absolute cells LBAs" \
			-command {::plugins::psl2::display_absolute_lbas}
		.mainmenu.plugins.psl2.tools add command -label "Display PGC starting time (SCR)" \
			-command {::plugins::psl2::display_starting_scr}
		.mainmenu.plugins.psl2.tools add command -label "Display CRC32 code" \
			-command {::plugins::psl2::display_crc32}
		.mainmenu.plugins.psl2.tools add command -label "Clean cell commands" \
			-command {::plugins::psl2::clean_cell_commands}

	if {! [catch {package require registry}]} {
		.mainmenu.plugins.psl2 add cascade -label "DVD Decrypter" -menu .mainmenu.plugins.psl2.dvdd
		menu .mainmenu.plugins.psl2.dvdd
			.mainmenu.plugins.psl2.dvdd add command -label "Launch DVD Decrypter (File mode)" \
				-command {::plugins::psl2::launch_dvdd FILE}
			.mainmenu.plugins.psl2.dvdd add command -label "Launch DVD Decrypter (IFO mode)" \
				-command {::plugins::psl2::launch_dvdd IFO}
			.mainmenu.plugins.psl2.dvdd add separator
			.mainmenu.plugins.psl2.dvdd add command -label "Reset DVD Decrypter rip settings" \
				-command {::plugins::psl2::reset_dvdd_settings}
	}
	.mainmenu.plugins.psl2 add cascade -label "FixVTS" -menu .mainmenu.plugins.psl2.fixvts
	menu .mainmenu.plugins.psl2.fixvts
		.mainmenu.plugins.psl2.fixvts add command -label "Launch FixVTS in background" \
				-command {set n [::plugins::psl2::find_exe FixVTS]; if {$n!=""} {exec -- $n &}}
		.mainmenu.plugins.psl2.fixvts add command -label "Process current Title domain with FixVTS" \
				-command {::plugins::psl2::fixvts_current_domain}

	.mainmenu.plugins.psl2 add cascade -label "VobBlanker" -menu .mainmenu.plugins.psl2.vb
	menu .mainmenu.plugins.psl2.vb
		.mainmenu.plugins.psl2.vb add command -label "Launch VobBlanker in background" \
				-command {set n [::plugins::psl2::find_exe VobBlanker]; if {$n!=""} {exec -- $n &}}
		.mainmenu.plugins.psl2.vb add command -label "Process current Title domain with VB" \
				-command {::plugins::psl2::vobblanker_current_domain}

	.mainmenu.plugins.psl2 add cascade -label "Help" -menu .mainmenu.plugins.psl2.help
	menu .mainmenu.plugins.psl2.help
		.mainmenu.plugins.psl2.help add command -label "Online User Guide (short version)" \
				-command {::main::display_html "http://forum.digital-digest.com/showthread.php?t=56831"}
		.mainmenu.plugins.psl2.help add command -label "Online User Guide (full version)" \
				-command {::main::display_html "http://forum.digital-digest.com/showthread.php?t=56078"}
		.mainmenu.plugins.psl2.help add separator
		.mainmenu.plugins.psl2.help add command -label "Version history" \
			-command ::plugins::psl2::versionhistory
		.mainmenu.plugins.psl2.help add command -label "About" -command {
			tk_messageBox  -title "PSL2 (ARccOS) plugin" -icon info \
				-message "PSL2 (ARccOS) plugin for PgcEdit 0.6.0 or more.\nTo be used with DVD Decrypter 3.5.4.0.\nVersion $::plugins::psl2::version, by The Mad Monk\nEnjoy!"
	}
}


namespace eval ::plugins {}
namespace eval ::plugins::psl2 {

set version "2.14"

set remap_cell 1	;# Default state of the remap cell option (1 or 0)
set remap_cellnum 1	;# Default remapping cell number (normally 1)
set DVDD_IO_SoftwareReadErrorRetries 3	;# Default number of read error retries for DVD Decrypter

# display version history
proc versionhistory {} {
	set txt {PSL2 plugin: Version history

V2.14:
- Added an option on the first page of the wizard to start it automatically
  when PgcEdit is launched.

V2.13:
- The commands with a condition "if ( gprm(x) & 0 ) then ..." are now correctly
  removed from the list of post commands by the clean cell commands routine.
  (Fix needed to do WWE Wrestling Superstars)
- Clean Cell Commands:  When a sprm(4), sprm(5) or sprm(6) is used in an IF
  condition of a cell command, it is now replaced by his current value.  Makes
  it easier to read the command.  (Currently, this new cleanup is not applied
  when the PGC is in a menu domain.)

V2.12:
- Fixed a bug in the wizard: When there are too many protected titles to be
  displayed in the wizard window in phase 3, step 1, the Next button was not
  displayed, and it was impossible to continue the process.  (Cf. Toy Story 2,
  10th. Aniversary, with 104 protected PGCs!)
- Added the function "Advanced mode -> After rip: Remove unreferenced Titles"
  to remove all fake unreferenced titles at the end of the title domains before
  using the standard Remove Crap pass.  It is sometimes necessary to do it so
  to be able to remove crap on all PGCs. The wizard doesn't need this function.
- Little bug fixed in the Remove Crap GUI. The function crashed when the Return
  key was used instead of the Remove Marked Cells button to accept the dialog.

V2.11:
- Support for VobBlanker 2.1. The format of the project (INI) file has changed
  since 2.1.  The plugin is now compatible with this new format, as well as
  with older versions of VobBlanker.
- Added a VobBlanker menu, with a function to launch VB in the background with
  the default user options, and a function to process the current title domain
  with the specific 'remove crap' options.
- The destination folder used to rip the DVD during phase 1 of the wizard is
  now automatically selected by default in DVD Decrypter during phase 2.
  Previously, a wrong folder was sometimes used.
- Added a bunch of DVD Decrypter download links in the psl2_install.txt file.

V2.10:
- If there were illegal characters for a filename in the DVD-Text name, the
  default filename for the PSL file was illegal, and it was not possible to
  save the PSL file.  Now, the illegal characters are remapped to "_".

V2.9:
- Fixed a little bug introduced in v2.8 in the wizard:  the VobBlanker option
  Force "Fix always VOB int pointers" was impossible to select.

V2.8:
- Added an option to process all DVD files with FixVTS.  It's normally not
  necessary, but this option can be used to remove some unreferenced material
  in non-ARccOS protected titlesets or menus.
- FixVTS beta 1.22 included.  This release fixes a bug when the last cell in a
  VOB file was removed.

V2.7:
- When post-processing the VOBs with VobBlanker v2.1 or more, his new option
  to renumber the VOB/Cell IDs is used instead of the option to fill the gaps
  with tiny black cells.  (Warning: the wrong method may be used if there are
  several versions of VobBlanker installed on the computer!)
- In some rare cases ("Dark Water" zone 1), the plugin was unable to recognize
  some titles as protected.  Fixed.

V2.6:
- Clean cell commands function enhanced to remove also the new fake conditions
  "if { gprm(X) >= 0 }" and "if { gprm(X) <= 65535 }".  Note that this is only
  a cosmetic change.  The remove crap method worked well in previous versions.
- The read speed of DVD Decrypter is now forced to MAX when it is launched by
  the wizard and when using the "Reset DVD Decrypter rip settings" function.
  This can speed up the rip process of an ARccOS disc considerably.

V2.5:
- Wizard: The user is not prompted any more to browse to the ripped DVD folder
  when phase 3 (remove crap) is called immediately after phase 1 (first rip).
  This happens when there were no read errors during the rip.  The DVD folder
  name is now retrieved automatically from the DVD Decrypter log.

V2.4:
- Bug fixed: When using VobBlanker, the "Use blank cells in unref. middle ones"
  option was unticked, and "Use blank cells in unref. last ones" was ticked.
  Now, it's exactly the opposite.  That should fix the problems with wrong
  audio file sizes and impossibility to shrink the main movie in DVD Shrink.
- The Reset DVD Decrypter rip settings function now resets also the Device /
  Disable Media Change Notification option to ON (which is the default value.)
  This option is now also temporarily forced when DVD Decrypter is called by
  the wizard.

V2.3:
- The cell number used for the remapping option in advanced mode was increased
  by 1 (ie, by default, cell 2 was used instead of cell 1.)
- Fixed a rare bug when trying to locate the original DVD Decrypter log file.

V2.2:
- Changed the error message when a non ARccOS protected DVD is opened in the
  wizard.
- The button to use upon error is now labelled "Close" instead of "Cancel".
- Some minor changes in the messages of the wizard.

V2.1:
- Changed the messages when opening the DVD in the wizard to insist to open the
  DVD FOLDER, with the files ripped on hard disc!
- Added a warning in the "Display CRC32 code" function to be sure that the user
  has disabled AnyDVD and DVD43, and that the original DVD is used.

V2.0:
- Added the wizard.
- Menu reorganized in sub-menus.
- Fixed the "Madagascar" removing crap problem, where only cells 1 and 2 were
  automatically selected, instead of cells 1 to 3 or 4.
- Added the menu "Show CRC32 of DVD" to display the CRC32 code used by DVD
  Decrypter to identify the DVD.  You must select the ORIGINAL DVD file!
  This function may help selecting the right PSL file to download on the net.
- Added the menu Quick Open DVD, to open a DVD, skipping most of the dialogs
  opened normally by PgcEdit.
- Added the menu to open the short guide in your web browser.
- Changed some of the VobBlanker options when removing crap in the VOBs.
- Code reorganized. ;-)

V1.16:
- Added support to use FixVTS (by jeanl) instead of VobBlanker in the final
  pass.  The option to launch VobBlanker from the Remove Crap pass has been
  removed.  You must now use either the FixVTS or VobBlanker menu option.
- Added also the menus to launch FixVTS as a standalone application, and to
  process the current Title domain with FixVTS.
- Added "Reset DVD Decrypter rip settings" menu to set the DVD Decrypter
  options related to the ARccOS protections to reasonable defaults.
- Fixed a bug after the VobBlanker processing when the original protected title
  has been completely removed by the Remove Crap pass.
- Added Install.txt in the archive with instructions on how to install PgcEdit
  and the plugin files.

V1.15:
- Bug when the VOB/Cell ID to patch was not the first one in the VOB fixed.

V1.14:
- Some unused cell commands were not removed in some case.  Fixed.
- It is now possible to remove crap on titles without selecting any cell to
  remove.  (It's necessary to clean title 94 of Madagascar.)
- Identify and offer to remove completely the fake titles used in the newer
  ARccOS protections.  (Madagascar title 94)

V1.13:
- The last cell of the PGC was wrongly included by the Find Unplayed Cells when
  this cell was used to allow the user to skip to the end of the title with the
  remote control, and was preceded by a LinkTailPGC cell-command.

V1.12:
- Remove crap can now be used on a DVD ripped by DVDFab Decrypter. (Previously
  the Find All Cells to Remove function did not identify the first cell as a
  cell to remove.)
- Remove crap now adds the missing entries in the VTS_VOBU_ADMAP tables so that
  it is now possible to use DVD Shrink and similar programs without having to
  process the domain with VobBlanker.

V1.11:
- Added "Tool: Display PGC starting time (SCR)" menu to help resynchronize the
  subpics extracted by SubRip or similar programs.
- Fix: VobBlanker's backup folder was not always saved at the right location.
- Fix: An Infinite Loop error message was displayed in some cases when finding
  the unplayed cells.
- Fix: The number of deleted cells commands was wrong in the message after a
  Clean Cell Commands.
- PgcEdit v 0.6.1 or more is now needed to use this plugin.

V1.10:
- Fixed crash if the Open File dialog was cancelled after Generate PSL File ->
  Append to Existing PSL2 file.

V1.9:
- Workaround for the difficult problem of the VOB/Cell ID of the ARccOS cells
  being reused in the played cells.  (Cf. Seinfeld season 4, DVD 4, VTS 3.)
- In the Remove Crap pass, new button "Automatically find all cells to remove"
  to automate the job of finding the right cells.  (The last tiny cells should
  not be removed because those cells are usually beginning with an Open GOP.)
- New menu option to process the VOBs with VobBlanker after the whole job has
  been done in several titlesets.  Useful to process episodic DVDs at once, or
  to come back to VobBlanker after some manual work in PgcEdit.
- An incremental backup is now automatically made at the end of the VobBlanker
  pass, to be sure to keep the IFOs and VOBs synchronized.
- New menu "Online User Guide".  (Link to "PSL2 plugin for PgcEdit - A User
  Guide" at Digital Video Forum, by Cynthia)
- The "Find Cells with bad VOBU addressed" function do not automatically select
  the cells anymore. In some cases, some cells that must be kept were selected!
- Several little bugs fixed and some improvments.

V1.8:
- New Append button in the Generate PSL2 File GUI, to be used when there are
  several protected titlesets in the same DVD (ARccOS protection on episodes)
  Note: the DVD Description and the path to original PSL2 file fields are not
  used when the Append button is selected.
- New "Find the cells with bad VOBU addresses in the VTS_VOBU_ADMAP table"
  button to searches for discrepancies in the VOBU table and select and mark
  the corresponding cell(s) with a light yellow background.  Seems those
  discrepancies are always in a cell with a read error.
- When there are several PGCs in the same domain using the same removed cells,
  the Remove Crap function is now automatically restarted to process also the
  remaining PGSs.  Note: the user must still manually select at least one
  protected PGC in each protected titleset.
- New feature to add automatically the cells of other PGCs belonging to the
  same protected area in the output.  In such case, a warning is issued.
- It is now possible to right-click on a cell to preview it. (Use this feature
  only when doing the Remove Crap pass!)
- There are now tooltips on the cells checkboxes to show the cell command
  (before and after the cleanup of the fake conditions)
- New "Tool: Find protected titles" menu to display the list of protected
  titles and to select automatically the first or next protected title.
- Added the Version History menu.

V1.7:
- Added checks for fake <=, =>, &, < and > conditions in cell commands.
  The cell commands with those conditions were not checked in previous versions.
  Should fix the problem with Seinfeld season 4 disc 4.
  See here: http://forum.digital-digest.com/showthread.php?t=56189
- It is now impossible to remove all cells of the current title during the
  Remove Crap pass.  See bug report by Blutach here:
  http://forum.digital-digest.com/showpost.php?p=306225&postcount=41
- The program number is now displayed in the main table of cells.

V1.6:
- Small bug fixed: it was possible to accept the Remove Crap GUI without
  selecting any cell to remove.
- Mew Remove Crap button to locate the tiny blank, unuseful cells in Program 1.
- New Remove Crap option to launch VobBlanker automatically to process the
  main title VOBs.  Pressing PROCESS!! when VB starts then QUIT when the job
  is done is sufficient.  If the plugin cannot find VobBlanker.exe, the user is
  prompted to browse to it (the first time only.)
- New message to remind the user to save the DVD after the Remove Crap pass
  (Only if the Launch VobBlanker option is not set.)

V1.5:
- The option to highlight the cell with error handles now all errors
  referenced in the log file.
- Changed some error messages to be more user friendly.
- Added source code comments.

V1.4:
- Hopefully first bug-free version.
}
	::main::showtext .psl2history \
		"PSL2 plugin v$::plugins::psl2::version: Version history" $txt
}


####################### main GUI #####################
# main GUI function: Generate the PSL file, or,
# if crap is true, do the remove crap post-processing.
proc main_gui {crap} {
	show_msg ""

	if {[catch {if {![test_version 0.6.1 "PSL2 Plugin"]} return}]} return ""

	# Init internal values, + safety checks
	if {![info exists ::dvddir] || $::scratchbook_active} {
		bell
		show_msg "Select the ARccOS protected Title"
		return ""
	}
	set type [lindex $::editor_cmds_from 0]
	if {$type == "menu"} {
		bell
		show_msg "Select the ARccOS protected Title"
		return ""
	}
	set indexes [split [lindex $::editor_cmds_from 1] ,]
	set vts [lindex $indexes 0]
	if {$vts == 0} {
		bell
		show_msg "Dummy PGC!"
		return ""
	}
	set pgc [lindex $indexes 1]
	set numcells  $::pgcs($vts,$pgc,numcells)
	if {$numcells == 0} {
		bell
		show_msg "Dummy PGC!"
		return ""
	}

	# If this function is called for the first time, or
	# if a new DVD has been opened since the last use fo this function,
	# then forget all cleaned VTS (to disable the post-processing with VobBlanker)
	if {![info exists ::plugins::psl2::old_dvddir] || \
		$::plugins::psl2::old_dvddir != $::dvddir} {
		set ::plugins::psl2::cleaned_titles_list {}
	}

	set cellpbtable $::pgcs($vts,$pgc,cellpbtable)
	set vobfile "VTS_[format %02d $vts]_1.VOB"
	# initialize table of programs: prog(cell) = program
	for {set i 0} {$i < $::pgcs($vts,$pgc,numprograms)} {incr i} {
		set prog([get_int ::pgcs($vts,$pgc,programmap) $i 1]) [expr $i+1]
	}

	# Retrieve cell commands (as Tcl List of hex codes)
	set cellcommands [get_type_cmds cell ::pgcs($vts,$pgc)]
	# And generate another list with clean cell commands.
	set cleaned [clean_cell_commands_sub pgc $vts $pgc \
			[string toupper [get_type_cmds pre ::pgcs($vts,$pgc)]] \
			[string toupper $cellcommands] $numcells $cellpbtable]
	set cleancellcmds [lindex $cleaned 0]
	set cleancellpbtable [lindex $cleaned 1]

	# If the current PGC is a fake PGC, simply remove it
	if {$crap && [remove_if_fake_pgc $vts $pgc false]} {
		return "OK"
	}

	# build GUI
	set w ".psl2editor"
	if {[winfo exists $w]} {
		destroy $w
	}
	toplevel $w -class Dialog
	wm withdraw $w
	wm protocol $w WM_DELETE_WINDOW {set ::plugins::psl2::psl2done 0}
	bind $w <Escape> "set ::plugins::psl2::psl2done 0"
	if {$crap} {
		bind $w <Return> "set ::plugins::psl2::psl2done -1"
	} else {
		bind $w <Return> "set ::plugins::psl2::psl2done 1"
	}
	if {$::config(tk) != "aqua"} {wm transient $w .}
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}

	if {!$crap} {
		wm title $w "PSL2 plugin - Generate PSL2 file (before rip)"
		frame $w.f1 -borderwidth 2 -relief groove
		pack  $w.f1 -padx 4 -pady 4 -expand 1 -fill x
		label $w.f1.lab -text "DVD description:\n(Movie name - DVD zone - DVD country or language, edition, rental/store version...)" -justify left
		pack  $w.f1.lab -anchor w
		entry $w.f1.e -textvariable ::plugins::psl2::name -validate key \
			-validatecommand {::plugins::psl2::validate_dvd_name {%P}}
		pack  $w.f1.e -expand 1 -fill x -padx 4 -pady 2
		set pady 0
	} else {
		wm title $w "PSL2 plugin - Remove crap (after rip)"
		set pady 4
	}

	frame $w.mf -borderwidth 2 -relief groove
	pack  $w.mf -padx 4 -pady $pady -fill x -anchor w
	if {!$crap} {
		label $w.mf.lab -text "Check the ARccOS protected cells:\nNOTE: A cell command exists in the cells marked with \"!\".  See the tooltip." -justify left
	} else {
		label $w.mf.lab -text "Check the ARccOS and the first tiny blank dummy cells:\nNOTES: Right-click on a cell below to preview it.\nA cell command exists in the cells marked with \"!\".  See the tooltip." -justify left
	}
	pack  $w.mf.lab -anchor w
	frame $w.mf.f
	pack  $w.mf.f -padx 3 -pady 4 -anchor w
	for {set c 1} {$c <= $numcells} {incr c} {
		set cellcmdnum [get_int cellpbtable [expr {(($c - 1) * 24) +3}] 1]
		if {$cellcmdnum > 0} {
			set mark "!"
		} else {
			set mark ""
		}
		if {[info exists prog($c)]} {
			set lab "$c $mark (Pg.$prog($c))"
		} else {
			set lab "$c $mark      "
		}
		checkbutton $w.mf.f.c_$c -variable ::plugins::psl2::cb($c) -text $lab -bd 0 -pady 0 -anchor w
		grid $w.mf.f.c_$c -row [expr ($c-1) % 10] -column [expr ($c-1) / 10] -sticky ew -padx 1
		if {![info exists ::plugins::psl2::cb($c)]} {
			set ::plugins::psl2::cb($c) 0
		}
		bind $w.mf.f.c_$c <3> "::preview::preview2 pgc $vts,$pgc $c false false 0"
		if {$cellcmdnum > 0} {
			if {$cellcmdnum <= [llength $cellcommands]} {
				set origcellcmd [::disasm::disasm [lindex $cellcommands [expr $cellcmdnum-1]]]
				set cleancellcmdnum [get_int cleancellpbtable [expr {(($c - 1) * 24) +3}] 1]
				if {$cleancellcmdnum > 0} {
					set cleancellcmd [::disasm::disasm [lindex $cleancellcmds [expr $cleancellcmdnum-1]]]
				} else {
					set cleancellcmd "none!"
				}
				if {$origcellcmd == $cleancellcmd} {
					::balloon::set_balloon $w.mf.f.c_$c $origcellcmd
				} else {
					::balloon::set_balloon $w.mf.f.c_$c "original:   $origcellcmd\ncleaned: $cleancellcmd"
				}
			} else {
				::balloon::set_balloon $w.mf.f.c_$c "Cell command $cellcmdnum not defined!"
			}
		}
		set lastremoved false
	}
	if {$crap} {
		button $w.mf.b0 -pady 0 -text "Automatically find all cells to remove" \
			-command "
				::plugins::psl2::err {}
				set ec \$::plugins::psl2::entry_cell
				if {!\[string is integer \$ec\]||\$ec==0||\$ec=={}} {
					::plugins::psl2::err {Illegal entry cell value}
					focus $w.mf.ff.f.en
				} else {
					::plugins::psl2::find_arcoss_cells $vts $pgc $numcells [list $cellpbtable] ::pgcs($vts,$pgc) [list $::pgcs($vts,$pgc,programmap)] \$ec
				}
				::plugins::psl2::find_tiny_cells $numcells [list $cellpbtable] [list $::pgcs($vts,$pgc,programmap)] true
			"
		pack   $w.mf.b0 -padx 4 -pady 4 -fill x -expand 1
	}
	frame  $w.mf.ff -bd 2 -relief groove
	pack   $w.mf.ff -padx 4 -fill x -expand 1
	frame  $w.mf.ff.f
	pack   $w.mf.ff.f -pady 2 -anchor w
	label  $w.mf.ff.f.l1 -text "The PGC is entered at cell #"
	pack   $w.mf.ff.f.l1 -side left
	entry  $w.mf.ff.f.en -textvariable ::plugins::psl2::entry_cell -width 3 -validate all \
		-validatecommand "::ced::value_entry_validate 1 $numcells {%d} {%P} 1"
	pack   $w.mf.ff.f.en -side left -padx 4
	label  $w.mf.ff.f.l2 -text "(Should be 1 for this PGC)"
	pack   $w.mf.ff.f.l2 -side left
	button $w.mf.ff.b2 -pady 0 -text "Find the unplayed cells by analysing the PGC structure" \
		-command "::plugins::psl2::err {}; set ec \$::plugins::psl2::entry_cell; if {!\[string is integer \$ec\]||\$ec==0||\$ec=={}} {::plugins::psl2::err {Illegal entry cell value};focus $w.mf.ff.f.en} else {::plugins::psl2::find_arcoss_cells $vts $pgc $numcells [list $cellpbtable] ::pgcs($vts,$pgc) [list $::pgcs($vts,$pgc,programmap)] \$ec}"
	pack   $w.mf.ff.b2 -padx 4 -pady 2 -fill x -expand 1

	if {$crap} {
		button $w.mf.b2 -pady 0 -text "Find tiny cells ( < 1 sec.) in Program 1" \
			-command "::plugins::psl2::find_tiny_cells $numcells [list $cellpbtable] [list $::pgcs($vts,$pgc,programmap)] false"
		pack   $w.mf.b2 -padx 4 -pady 4 -fill x -expand 1
	} else {
		button $w.mf.b1 -pady 0 -text "Find the cells with logged CIRC errors from the DVD Decrypter log file" \
			-command "::plugins::psl2::find_cells_w_errors_from_log $vobfile $numcells [list $cellpbtable]"
		pack   $w.mf.b1 -padx 4 -pady 2 -fill x -expand 1
		button $w.mf.b3 -pady 0 -text "Find the cells with bad VOBU addresses in the VTS_VOBU_ADMAP table" \
			-command "::plugins::psl2::find_cells_w_bad_vobus $vts $numcells [list $cellpbtable]"
		pack   $w.mf.b3 -padx 4 -pady 2 -fill x -expand 1

		frame $w.f2 -borderwidth 2 -relief groove
		pack  $w.f2 -padx 4 -pady 4 -expand 1 -fill x
		frame $w.f2.f1
		pack  $w.f2.f1 -fill x -expand 1 -pady 2
		label $w.f2.f1.lab -text "Start LBA of the current Domain ($vobfile):"
		pack  $w.f2.f1.lab -side left
		entry $w.f2.f1.r -textvariable ::plugins::psl2::domainlba -width 7 -validate all \
			-validatecommand "::ced::value_entry_validate 0 4173822 {%d} {%P} 1"
		pack  $w.f2.f1.r -side left -padx 2
		button $w.f2.f1.b -text "Retrieve from log file" -pady 0 -command \
			"set tmp \[::plugins::psl2::read_domain_lba_from_log $vobfile\]; if {\$tmp != {}} {set ::plugins::psl2::domainlba \$tmp}"
		pack   $w.f2.f1.b -side left -padx 4 -fill both -expand 1
		if {[winfo screenheight  .] > 600} {
			set txt "Open the DVD in DVD Decrypter, right-click on \"$vobfile\", and select Properties."
			set txt "${txt}\nThe starting LBA is the first number on the \"LBA:\" line."
			set txt "${txt}\nOr click on the button to retrive this value from the DVD Decrypter logfile."
			label $w.f2.lab -text $txt -justify left
			pack  $w.f2.lab -anchor w
		}

		frame $w.f3 -borderwidth 2 -relief groove
		pack  $w.f3 -padx 4 -expand 1 -fill x
		label $w.f3.lab -text "Path to original DVD Decrypter PSL2 file:"
		pack  $w.f3.lab -anchor w
		frame $w.f3.f1
		pack  $w.f3.f1 -fill x -padx 2 -pady 2 -expand 1
		entry $w.f3.f1.r -textvariable ::plugins::psl2::psl2path
		pack  $w.f3.f1.r -side left -padx 2 -fill x -expand 1
		button $w.f3.f1.b -text "..." -width 2 -pady 0 -command {
			set default_pslfile [::plugins::psl2::find_last_pslfile]
			set tmp [tk_getOpenFile -title "Open original DVD Decrypter PSL2 file" \
				-initialdir  [file dirname $default_pslfile] \
				-initialfile [file tail $default_pslfile] \
				-filetypes {{"PSL files" {*.psl}} {"All files" {*.*}}}]
			if {$tmp != ""} {
				set tmp [file nativename $tmp]
				set ::plugins::psl2::psl2path $tmp
				set ::plugins::psl2::default_pslfile $tmp
			}
		}
		pack   $w.f3.f1.b -side left -padx 2 -fill y
		set txt "If you don't provide the original PSL2 file, you will have to edit the output manually!"
		label  $w.f3.lab2 -text $txt -justify left
		pack   $w.f3.lab2 -anchor w

		frame $w.f4 -borderwidth 2 -relief groove
		pack  $w.f4 -padx 4 -pady 4 -expand 1 -fill x
		frame $w.f4.f1
		pack  $w.f4.f1 -fill x -expand 1
		label $w.f4.f1.lab -text "Recommended PSL2 output option:"
		pack  $w.f4.f1.lab -side left
		checkbutton $w.f4.f1.cb -text "Remap protected cells to valid cell #" -variable ::plugins::psl2::remap_cell \
			-command {if $::plugins::psl2::remap_cell {set s normal} else {set s disabled};.psl2editor.f4.f1.e configure -state $s}
		pack  $w.f4.f1.cb -side left
		entry $w.f4.f1.e -textvariable ::plugins::psl2::remap_cellnum -width 3 -validate all \
			-validatecommand "::ced::value_entry_validate 1 $numcells {%d} {%P} 1"
		pack  $w.f4.f1.e -side left
		if {! $::plugins::psl2::remap_cell} {$w.f4.f1.e configure -state disabled}
	}
	label  $w.err -foreground red
	pack   $w.err -fill x -pady 0

	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	if {$crap} {
		button $w.fb.rem -text " Remove marked cells and fake cell cmds " -command {set ::plugins::psl2::psl2done -1}
		pack   $w.fb.rem -side left -padx 2 -fill x -expand 1
	} else {
		button $w.fb.ap -text "  Append to existing PSL2 file  " -command {set ::plugins::psl2::psl2done 2}
		pack   $w.fb.ap -side left -padx 2 -fill x -expand 1
		button $w.fb.ok -text "  Save PSL2 file  " -command {set ::plugins::psl2::psl2done 1}
		pack   $w.fb.ok -side left -padx 2 -fill x -expand 1
		$w.f1.e icursor end
		focus $w.f1.e
	}
	button $w.fb.cancel -text "  Cancel  " -command {set ::plugins::psl2::psl2done 0}
	pack   $w.fb.cancel -side left -padx 2 -fill x -expand 1
	if {$::config(platform) == "mac"} {
		pack [frame $w.fb.macpad] -side left -padx 6
	}

	# If this function is called for the first time, or
	# if a new DVD has been opened since the last use fo this function, or
	# it the DVD has been modified,
	# Reinit GUI.
	if {![info exists ::plugins::psl2::old_dvddir] || \
		$::plugins::psl2::old_dvddir != $::dvddir || $::dvd_modified} {
		set ::plugins::psl2::old_dvddir $::dvddir

		set ::plugins::psl2::name [build_dvd_name true]

		# clear PSL output path
		set ::plugins::psl2::psl2path ""
	}
	# Clear cells checkboxes
	for {set c 1} {$c <= $numcells} {incr c} {
		set ::plugins::psl2::cb($c) 0
	}

	# init other GUI variables
	set ::plugins::psl2::domainlba ""
	set ::plugins::psl2::entry_cell 1

	# if there are LinkCN pre-commands, change the "(Should be 1 for this PGC)"
	# message to display the list of possible entry cells.
	set entry_cells [::plugins::psl2::find_possible_entry_cells [get_type_cmds pre ::pgcs($vts,$pgc)] $numcells $::pgcs($vts,$pgc,programmap)]
	if {[llength $entry_cells] > 0} {
		.psl2editor.mf.ff.f.l2 configure -text "May be 1, [join $entry_cells ", "].  Try all values!" -foreground red
	}

	# GUI cosmetics
	update
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry $w +$wx+$wy
	wm deiconify $w

	# wait until the user click the Process or Cancel button
	set ::plugins::psl2::psl2done -1
	while {1} {
		tkwait variable ::plugins::psl2::psl2done
		if {$::plugins::psl2::psl2done == 0} {
			destroy $w
			return ""
		}
		# In Save PSL mode, the DVD name must be provided
		if {$::plugins::psl2::psl2done == 1 && \
				([string range $::plugins::psl2::name 0 6] == " - Zone" || $::plugins::psl2::name == "")} {
			err "Please enter the DVD name and description!"
			focus $w.f1.e
			continue
		}
		# Some cells must be selected.
		set cellsselected false
		for {set c 1} {$c <= $numcells} {incr c} {
			if {$::plugins::psl2::cb($c)} {
				set cellsselected true
				break
			}
		}
		if {! $cellsselected} {
			if {$::plugins::psl2::psl2done == -1} {
				set txt "You haven't selected any protected cell!\n"
				set txt "${txt}Are you sure you want to continue?"
				set r [tk_messageBox -title "PSL2 plugin" -icon question -message $txt \
						-type yesno -default no]
			} else {
				err "You must select some cells!"
				set r "no"
			}
			if {$r == "no"} {
				focus $w.mf.f
				continue
			}
		}
		# If mode = Create PSL file, then the Starting LBA must be provided,
		# and the remap cell number must be a valid, non-protected cell.
		if {$::plugins::psl2::psl2done >= 1} {
			if {$::plugins::psl2::domainlba == ""} {
				err "The starting LBA must be provided!"
				catch {focus $w.f2.f1.r}
				continue
			}
			if {$::plugins::psl2::remap_cell} {
			 	if {$::plugins::psl2::remap_cellnum == "" || \
					$::plugins::psl2::remap_cellnum == 0 || \
					[string toupper $::plugins::psl2::remap_cellnum] == "0X"} {
					err "Remap cell number: illegal value!"
					catch {focus $w.f4.f1.e}
					continue
				}
				if {$::plugins::psl2::cb($::plugins::psl2::remap_cellnum) == 1} {
					err "Remap cell number: cannot use a protected cell!"
					catch {focus $w.f4.f1.e}
					continue
				}
			}
		} else {
			# In remove crap, the user must leave at least one cell
			set allcellsselected true
			for {set c 1} {$c <= $numcells} {incr c} {
				if {! $::plugins::psl2::cb($c)} {
					set allcellsselected false
					break
				}
			}
			if {$allcellsselected} {
				err "You cannot remove all cells!"
				catch {focus $w.mf.f}
				continue
			}
		}
		break
	}
	# close GUI
	destroy $w

	if {$::plugins::psl2::psl2done > 0} {
		###############################################
		# Pre-processing mode only: Generate PSL file #
		###############################################

		# If original PSL2 path is not empty...
		set crc32 ""
		set capacity ""
		set origline1 ""
		if {$::plugins::psl2::psl2path != ""} {
			set origpsl2list [get_headers_from_orig_psldata \
					[readFile $::plugins::psl2::psl2path auto]]
			set crc32     [lindex $origpsl2list 0]
			set capacity  [lindex $origpsl2list 1]
			set origline1 [lindex $origpsl2list 2]
		}
		if {$crc32 == "" && $capacity == ""} {
			set crc32 "<ENTER CRC32 OF VIDEO_TS.IFO HERE>"
			set capacity "<OR ENTER DVD CAPACITY HERE>"
		}

		# Build output file
		set out0 "0:$crc32,$capacity,$::plugins::psl2::name\n$origline1\n"

		set data [generate_psl2_data $vts $pgc $::plugins::psl2::domainlba \
				$::plugins::psl2::entry_cell $::plugins::psl2::remap_cell \
				[expr $::plugins::psl2::remap_cellnum -1] false]

		####################
		# Save output file #
		####################

		# Build default output filename
		if {[info exists ::plugins::psl2::psl2pathNew]} {
			set default_pslfile $::plugins::psl2::psl2pathNew
		} else {
			set default_pslfile [::plugins::psl2::find_last_pslfile]
			if {[string first "_NEW" $default_pslfile] == -1} {
				set default_pslfile [file rootname $default_pslfile]_new.PSL
			}
		}

		# In Generate PSL2 mode...
		if {$::plugins::psl2::psl2done == 1} {
			# and save the new PSL2 file
			set outfilename [tk_getSaveFile -title "Save new PSL2 file for DVD Decrypter" \
				-initialdir  [file dirname $default_pslfile] \
				-initialfile [file tail $default_pslfile] \
				-filetypes {{"PSL files" {*.psl}} {"All files" {*.*}}}]
			if {$outfilename != ""} {
				set outfilename [file nativename $outfilename]
				if {[string toupper [file extension $outfilename]] != ".PSL"} {
					set outfilename "$outfilename.PSL"
				}
				set ::plugins::psl2::psl2pathNew $outfilename
				set ::plugins::psl2::default_pslfile $outfilename
			}
			# Build full output by concatenating parts together.
			set out "$out0\n$data"
			# strip multiple blank lines
			set out [reorder_psl_file $out]
		} else {
			# Open previous PSL2 file
			set outfilename [tk_getOpenFile -title "Open previous PSL2 file for appending" \
				-initialdir  [file dirname $default_pslfile] \
				-initialfile [file tail $default_pslfile] \
				-filetypes {{"PSL files" {*.psl}} {"All files" {*.*}}}]
			if {$outfilename == ""} {
				return ""
			}
			set outfilename [file nativename $outfilename]
			if {![file exists $outfilename]} {
				bell
				return ""
			}
			set ::plugins::psl2::psl2pathNew $outfilename
			set ::plugins::psl2::default_pslfile $outfilename
			set original [readFile $outfilename auto]
			if {$original == ""} {
				bell
				return ""
			}
			# concat new 1: and 6: commands to existing commands.
			set out "$original\n$data\n"
			# strip multiple blank lines
			set out [reorder_psl_file $out]
		}
		# save file
		if {$outfilename != ""} {
			saveFile $outfilename auto $out
		}
		# and show the resulting data.
		::main::showtext .psl2file "PSL2 file" $out -defaultfilename $outfilename
		return $out
	} else {
		#####################################
		# Post-processing mode: Remove crap #
		#####################################

		# keep track of the other PGCs to process later
		set tmp [warn_if_more_pgcs $vts $pgc $numcells $cellpbtable]
		set morepgcnums [lindex $tmp 0]
		set morepgctext [lindex $tmp 1]

		set num_nav_packs_fixed [remove_crap $vts $pgc $::plugins::psl2::entry_cell false]

		if {[llength $morepgcnums] == 0} {
			update idletasks
			after 100 {::plugins::psl2::find_protected_titles true}
			return "OK"
		}
		if {[llength $morepgcnums] > 0} {
			set newpgc [lindex $morepgcnums 0]
			set msg "Cleanup done on the current PGC.\n\n"
			set msg "${msg}There are some other PGCs using the same cells in the same titleset:\n$morepgctext\n"
			set msg "${msg}The Remove Crap function will now be restarted to process the PGC $newpgc:\n"
			set msg "${msg}[lindex [split $morepgctext \n] 0]"
			set icon info
			if {$num_nav_packs_fixed > 0} {
				set msg "${msg}\n\nWARNING:  Since the VOB has already been modified, you MUST save the DVD\n"
				set msg "${msg}after the cleanup, or launch the FixVTS or VobBlanker pass!"
				set icon warning
			}
			tk_messageBox -title "PSL2 plugin" -icon $icon -message $msg
			::main::display_pgc pgc $vts $newpgc
			update idletasks
			after 500 {::plugins::psl2::main_gui true}
		}
		return "OK"
	}
}
# prohibits the use of "," in the DVD name field of the GUI.
proc validate_dvd_name {name} {
	if {[string first , $name]>=0} {
		bell
		return false
	}
	return true
}

########## GUI dependant functions, directly called by the main GUI ###########
# Finds and get the domain LBA of the VOB file specified in argument
# in the logfile text data.
# Displays an error if the LBA is not found, and a warning if there
# are several occurences of the filename in the log file. (This is
# possible if the user has made several rips w/o clearing the log.)
proc read_domain_lba_from_log {vobfile {logdata ""}} {
	if {$logdata == ""} {
		set logdata [input_log_file]
	}
	if {$logdata == ""} {return ""}

	set domainstart ""
	set nummatches 0
	set index [string first "Decrypting ${vobfile}..." $logdata]
	while {$index != -1} {
		incr nummatches
		set logdata [string range $logdata [expr $index + 33] end]
		set domainstart [lindex $logdata 0]
		set index [string first "Decrypting ${vobfile}..." $logdata]
	}
	if {$nummatches == 0} {
		tk_messageBox -title "Error!" -icon error -message \
			"First LBA for file $vobfile not found."
		return ""
	}
	if {$nummatches > 1} {
		tk_messageBox -title "Warning!" -icon warning -message \
			"File $vobfile is referenced $nummatches times in the log file.\nUsing last reference."
	}
	return $domainstart
}

# Get log file data interactively
proc input_log_file {} {
	set default_pslfile [::plugins::psl2::find_last_logfile]
	set logfile [tk_getOpenFile -title "Open original DVD Decrypter log file" \
		-initialdir  [file dirname $default_pslfile] \
		-initialfile [file tail $default_pslfile] \
		-filetypes {{"Log files" {*.log}} {"All files" {*.*}}}]
	if {$logfile == ""} {return ""}
	set ::plugins::psl2::default_logfile  $logfile
	set log [readFile $logfile auto]
	return $log
}

# proc called from the "Dump absolute cells LBAs" function to input the start LBA of VOB.
proc input_starting_lba {vobfile} {
	set txt "Start LBA of the current Domain ?\n"
	set txt "${txt}\nYou can get this value from DVD Decrypter:"
	set txt "${txt}\nOpen the original DVD, right-click on \"$vobfile\", and select Properties."
	set txt "${txt}\nThe starting LBA is the first number on the \"LBA: ...\" line.\n"
	set txt "${txt}\nYou may also retrieve this value from the DVD Decrypter log file:"
	set txt "${txt}\nJust leave this field empty and select \"Use or get from DVDD log file\".\n"
	set domainstart [::stringreq::stringreq -window .psl2req -title "Domain starting LBA" \
		-text $txt -def "" -width 7 -geometry c -cancelreturn "CANCEL" \
		-callback {::ced::value_entry_validate 0 4173822 {%d} {%P} 1} \
		-ok "Use or get from DVDD log file"]
	if {$domainstart == "CANCEL"} {return ""}
	# If value is empty and OK button was clicked, get LBA from log file
	if {$domainstart == ""} {
		set domainstart [read_domain_lba_from_log $vobfile]
	}
	return $domainstart
}

# get CRC32, DVD Capacity from the 0: line
# and get first original 1: line from the original PSL data
proc get_headers_from_orig_psldata {psl2data} {
	set crc32 ""
	set capacity ""
	set lastline ""
	set index [string first "0:" $psl2data]
	if {$index != -1} {
		set tmp [string range $psl2data [expr $index+2] end]
		set index [string first "," $tmp]
		if {$index > 0} {
			set crc32 [string range $tmp 0 [expr $index-1]]
			set tmp [string range $tmp [expr $index+1] end]
			set index [string first "," $tmp]
			if {$index > 0} {
				set capacity [string range $tmp 0 [expr $index-1]]
			}
		}
	}
	# get original 1: line (if any).  It holds the sector numbers of the last, fake VTS.
	set index [string first "1:" $psl2data]
	if {$index != -1} {
		set lastline [string range $psl2data $index end]
		set index [string first "\n" $lastline]
		if {$index != -1} {
			set lastline [string range $lastline 0 [expr $index-1]]
		}
	}
	return [list $crc32 $capacity $lastline]
}

# Find and highlight cells with errors referenced in the logfile.
proc find_cells_w_errors_from_log {vobfile numcells cellpbtable} {
	# get logfine interactively
	set logdata [input_log_file]
	if {$logdata == ""} {return ""}

	# retrieve start LBA of VOB of current domain
	set domainstart [read_domain_lba_from_log $vobfile $logdata]
	if {$domainstart == ""} {return ""}

	# Get index on first read error message
	set index [string first "Failed to read Sector" $logdata]
	if {$index == -1} {
		err "No read error found in log!"
		return ""
	}

	# Build lists of start and end LBAs for each cell.
	for {set c 1} {$c <= $numcells} {incr c} {
		set cidx [expr (($c-1)*24)+8]
		set vobustart($c) [expr [get_int cellpbtable $cidx 4] + $domainstart]
		incr cidx 12
		set vobuend($c) [expr [get_int cellpbtable $cidx 4] + $domainstart]
		set checked($c) 0
	}

	set err_outside 0
	set in_this_pgc 0
	# While there are errors in the file
	while {$index != -1} {
		# remove whatever is before the error message in the log text
		set logdata [string range $logdata [expr $index + 22] end]
		# retrieve LBA number of the error
		set errorlba [lindex $logdata 0]
		set cellfound 0
		# searches in which cell this LBA is, and highlight and select it.
		for {set c 1} {$c <= $numcells} {incr c} {
			if {$vobustart($c)<=$errorlba && $vobuend($c)>=$errorlba} {
				set cellfound 1
				if {! $checked($c)} {
					set checked($c) 1
					set ::plugins::psl2::cb($c) 1
					.psl2editor.mf.f.c_$c configure -foreground red
					set in_this_pgc 1
					break
				}
			}
		}
		if {!$cellfound} {
			incr err_outside
		}
		# Get index on next read error message, and loop back
		set index [string first "Failed to read Sector" $logdata]
	}
	# Display warnings, if any.
	if {!$in_this_pgc} {
		if {$err_outside > 1} {
			err "The $err_outside read errors are not in this PGC!"
		} else {
			err "The read error referenced in the log is not in this PGC!"
		}
		return $c
	} else {
		if {$err_outside > 1} {
			err "There are $err_outside read errors outside this PGC!" -noterr
		} elseif {$err_outside > 0} {
			err "There is one read error outside this PGC!" -noterr
		}
		return ""
	}
}

# Searches for discrepancies in the VTS_VOBU_ADMAP table and
# mark the corresponding cells in the main GUI with a yellow background
proc find_cells_w_bad_vobus {vts numcells cellpbtable} {
	# scan the VTS_VOBU_ADMAP tables to find the VOBUs with a LBA < previous one
	set table $::raw_sectors($vts,VTS_VOBU_ADMAP)
	set endoftable [get_int table 0 4]
	set last 0
	set badvobus {}
	for {set i 4} {$i<$endoftable} {incr i 4} {
		set lba [get_int table $i 4]
		if {$lba<$last} {
			lappend badvobus $lba
		}
		set last $lba
	}
	if {$badvobus == {}} {
		err "No bad VOBUs found in the VTS_VOBU_ADMAP table of VTS $vts." -noterr
		return
	}

	# mark and select the cell(s) with a bad VOBU
	set in_this_pgc 0
	for {set c 1} {$c <= $numcells} {incr c} {
		set cidx [expr (($c-1)*24)+8]
		set vobustart [get_int cellpbtable $cidx 4]
		incr cidx 12
		set vobuend [get_int cellpbtable $cidx 4]
		foreach lba $badvobus {
			if {$lba >= $vobustart && $lba <= $vobuend} {
				#set ::plugins::psl2::cb($c) 1
				.psl2editor.mf.f.c_$c configure -background lightyellow
				set in_this_pgc 1
			}
		}
	}
	if {! $in_this_pgc} {
		err "The discrepancies in the VTS_VOBU_ADMAP table are not in this PGC." -noterr
	}
}

# check length of cells in program 1, and mark them if length of cell < 1 sec.
# New version: selects all tiny cells from program 1 before the last non-seamless cell.
proc find_tiny_cells {numcells cellpbtable programmap remove_crap} {
	# find first cell of program 2
	set cellprogram2 [get_int programmap 1 1]

	# compute the last cell to process/remove
	set maxcelltoremove [expr $cellprogram2-2]
	if {$remove_crap} {
		set lastlb 0
		for {set c 1} {$c <= $cellprogram2} {incr c} {
			set cidx [expr ($c-1)*24]
			if {[expr [get_int cellpbtable $cidx 1] & 8] == 0} {
				set lastlb $c
			}
		}
		if {[expr $lastlb-1] < $maxcelltoremove} {
			set maxcelltoremove [expr $lastlb-1]
		}
	}

	# check length of cells in program 1, and mark them if length of cell < 1 sec.
	set tinycells {}
	for {set c $maxcelltoremove} {$c > 0} {incr c -1} {
		set cidx [expr ($c-1)*24]
		if {[get_int cellpbtable [expr $cidx+4] 3]==0} {
			lappend tinycells $c
			if {$remove_crap} {
				set ::plugins::psl2::cb($c) 1
			} else {
				.psl2editor.mf.f.c_$c configure -foreground blue
			}
		}
		# if no cells selected and cells 1 or 2 are tiny cells and
		# they have the restricted cell flag set, remove them also.
		# (for titles incorrectly stripped of bad cells by DVDFab Decrypter.)
		if {$remove_crap} {
			if {$c < 3 && $::plugins::psl2::cb($c) == 0 &&
					[expr [get_int cellpbtable [expr (($c-1)*24)+1] 1] & 0x20] == [expr 0x20] &&
					[get_int cellpbtable [expr (($c-1)*24)+4] 3]==0} {
				set ::plugins::psl2::cb($c) 1
				lappend tinycells $c
			}
		}
	}

	return $tinycells
}

# Display error message in red in the plugin's GUI,
# or, if not opened, pops up a dialog.
proc err {msg {noterr ""}} {
	if {$msg == ""} {
		catch {.psl2editor.err configure -text ""}
	} else {
		if {$noterr == ""} {
			bell
			if {[catch {.psl2editor.err configure -text "Error:  $msg" -foreground red}]} {
				tk_messageBox -title "PSL2 (ARccOS) plugin" -icon error -message $msg
			} else {
				after 6000 {catch {.psl2editor.err configure -text ""}}
			}
		} else {
			if {[catch {.psl2editor.err configure -text $msg -foreground blue}]} {
				tk_messageBox -title "PSL2 (ARccOS) plugin" -icon info -message $msg
			} else {
				after 6000 {catch {.psl2editor.err configure -text ""}}
			}
		}
	}
}

# trick to force the GUI to refresh
proc refresh_gui {{delay 100}} {
	update idletasks
	after 100 {set ::plugins::psl2::wiz::refresh_gui_pause 1}
	tkwait variable ::plugins::psl2::wiz::refresh_gui_pause
	return
}

############################### WIZARD ##############################
# Wizard.
proc wizard {} {
	if {[catch {if {![test_version 0.6.1 "PSL2 Plugin"]} return}]} return ""

	if {[info exists ::dvddir] && $::dvd_modified} {
		set txt "The current DVD has been modified but not yet saved.\n"
		set txt "${txt}The wizard needs to load the ripped DVD, and will NOT\nsave the current DVD automatically.\n\n"
		set txt "${txt}Continue anyway?"
		set r [tk_messageBox -title "PSL2 plugin - Wizard" -icon question -message $txt \
				-type yesno -default no]
		if {$r == "no"} return
	}

	# build GUI
	set w ".psl2wizard"
	if {[winfo exists $w]} {
		bell
		return
	}

	# destroy old GUI variables
	catch {namespace delete ::plugins::psl2::wiz}
	namespace eval ::plugins::psl2::wiz {}

	toplevel $w
	wm withdraw $w
	wm protocol $w WM_DELETE_WINDOW {set ::plugins::psl2::wiz::wizarddone "Cancel"}
	bind $w <Escape> {
		if {[winfo ismapped .psl2wizard.bf.b1] && [.psl2wizard.bf.b1 cget -state]=="normal"} {
			set ::plugins::psl2::wiz::wizarddone [.psl2wizard.bf.b1 cget -text]
		}
	}
	bind $w <Return> {
		if {[winfo ismapped .psl2wizard.bf.b2] && [.psl2wizard.bf.b2 cget -state]=="normal"} {
			set ::plugins::psl2::wiz::wizarddone [.psl2wizard.bf.b2 cget -text]
		}
	}
	frame $w.f1 -bd 2 -relief groove -padx 4 -pady 4
	pack  $w.f1 -padx 4 -pady 4 -expand 1 -fill both
	frame $w.f1.mf
	pack  $w.f1.mf -side left -expand 1 -fill x

	frame $w.bf
	pack  $w.bf -pady 2 -anchor e
	button $w.bf.b1 -width 12 -text "Cancel" \
		-command {set ::plugins::psl2::wiz::wizarddone [.psl2wizard.bf.b1 cget -text]}
	pack  $w.bf.b1 -side left -padx 4
	button $w.bf.b2 -width 12 -text "Next" \
		-command {set ::plugins::psl2::wiz::wizarddone [.psl2wizard.bf.b2 cget -text]}
	pack  $w.bf.b2 -side left -padx 4

	# Determine minimum window size, and center window on screen, then open it
	set maxx 0
	set maxy 0
	foreach step {0_1 0_2 1_1 1_2_a 1_2_b 1_3a_1 1_3a_2 1_3b
			2_prep1 2_prep2 2_prep3 2_1_gui 2_1_load 2_2 2_3 2_4 2_5_1 2_5_2 2_6
			3_prep1_gui 3_prep1_load 3_1 3_2 3_3a_1 3_3a_2 3_3a_3 3_3b_1 3_3b_2 3_3b_3 3_4
			Finished} {
		wiz_$step
		update
		set tmp [winfo reqwidth  $w]
		if {$tmp > $maxx} {set maxx $tmp}
		set tmp [winfo reqheight $w]
		if {$tmp > $maxy} {set maxy $tmp}
	}
	wm geometry $w ${maxx}x$maxy
	wm minsize $w $maxx $maxy
	wm maxsize $w $maxx $maxy
	wm resizable $w 0 0
	wm iconify .

	######## Init some useful variables ########
	set dvdd_keyname {HKEY_CURRENT_USER\Software\DVD Decrypter}
	set dvdd_logfilename [file nativename [file join $::config(tempdir) "DVDDecrypter.log"]]
	set phase2done false
	set dvd_dirname_sure false
	set req_pid 0
	set step 0_1

	######## Phase 0: select main phase ########

	# Welcome and autostart toggle
	if {$step == "0_1"} {
		if {! [info exists ::plugins::psl2::autostart]} {
			set ::plugins::psl2::autostart true
		}
		set old_autostart $::plugins::psl2::autostart
		wiz_0_1
		wm deiconify $w
		set step [wiz_wait_button 0_2]
		if {$::plugins::psl2::autostart != $old_autostart} {
			# see if the psl2_plugin.cfg file exists
			set configfile [file join $::config(config_directory) "psl2_plugin.cfg"]
			if {[file exists $configfile]} {
				set txt [readFile $configfile auto]
				set txt [string map {"set ::plugins::psl2::autostart true\n" "" "set ::plugins::psl2::autostart false\n" ""} $txt]
				set txt [string trim $txt \n]
				if {$txt != ""} {set txt "$txt\n"}
			} else {
				set txt ""
			}
			set txt "${txt}set ::plugins::psl2::autostart $::plugins::psl2::autostart"
			saveFile $configfile auto $txt
		}
	}
	
	# select main phase
	if {$step == "0_2"} {
		wiz_0_2
		set step [wiz_wait_button Next]
		if {$step != "Cancel"} {
			switch $::plugins::psl2::wiz::wizphase {
				1 {set step 1_1}
				2 {set step 2_prep1}
				3 {set step 3_prep1}
			}
		}
	}

	######## Phase 1: First rip of the DVD ########

	# Insert DVD
	if {$step == "1_1"} {
		wiz_$step
		set step [wiz_wait_button 1_2_a]
	}

	# Explain how to use DVDD: Click Abort upon I/O error, and continue the rip.
	if {$step == "1_2_a"} {
		wiz_$step
		set step [wiz_wait_button 1_2_b]
	}

	# Launch DVDD and rip
	if {$step == "1_2_b"} {
		# remember the DVDD setups we need to change
		wiz_$step

		set txt "Remember:\n\nIf you see an I/O error dialog during the rip,\n"
		set txt "${txt}click the \"Abort\" button, and continue to rip!\n\n"
		set txt "${txt}When the rip is finished, close DVD Decrypter."
		set req_pid [open_background_msgbox "PSL2 Plugin - Wizard - Phase 1, step 2" $txt 2000]

		# launch DVDD
		catch {file delete -force $dvdd_logfilename}
		set_dvdd_regkeys set 0
		refresh_gui
		wm iconify $w	;# optional!
		set rc [launch_dvdd FILE false]
		set_dvdd_regkeys restore
		set req_pid [close_background_msgbox $req_pid]
		wm deiconify $w
		if {$rc != 0} {
			set step "Error"
			set errmsg "Cannot launch DVD Decrypter!\nReinstalling DVD Decrypter may solve the problem."
		} else {
			set logdata [readFile $dvdd_logfilename auto]
			if {$logdata == ""} {
				set step "Error"
				set errmsg "Cannot open or read the DVD Decrypyer log file \"$dvdd_logfilename\"!"
			}
		}
		if {$step != "Error"} {
			if {[string first "Operation Successfully Completed" $logdata] == -1} {
				set step "Error"
				set errmsg "DVD Decrypter aborted prematurely !"
			}
		}
		if {$step != "Error"} {
			set ::plugins::psl2::default_logfile $dvdd_logfilename
			# generate filename for output log file, and keep DVD folder name to open it in phase 2
			set idx [expr [string first "Destination Folder:" $logdata] + 20]
			set tmp [string range $logdata $idx [expr $idx+200]]
			set tmp [string range $tmp 0 [expr [string first \n $tmp] - 1]]
			set dvd_dirname [file nativename $tmp]
			set dvd_dirname_sure true
			if {[file tail $tmp] == "VIDEO_TS"} {
				set tmp [file dirname $tmp]
			}
			set output_logfilename [file join $tmp "DVDDecrypter.log"]
			if {[string first "Failed to read Sector" $logdata] != -1} {
				set step 1_3a
			} else {
				set step 1_3b
			}
		}
	}

	# I/O errors: retrieve CRC32 of VIDEO_TS.IFO, and region code mask
	if {$step == "1_3a"} {
		wiz_1_3a_1
		refresh_gui
		# Get the drive letter from the DVDD log
		set idx [string first "Source Device:" $logdata]
		if {$idx == -1} {
			set step "Error"
			set errmsg "Cannot retrieve the drive letter from the DVD Decrypter log!"
		}
		if {$step != "Error"} {
			set tmp [string range $logdata $idx [expr $idx+100]]
			set idx [string first ":)" $tmp]
			incr idx -1
			if {$idx < 0} {
				set step "Error"
				set errmsg "Cannot retrieve the drive letter from the DVD Decrypter log!"
			}
		}
		# compute crc32 of VIDEO_TS.IFO using crc32.exe
		if {$step != "Error"} {
			set drive [string range $tmp $idx $idx]:\\
			set crc32 [crc32 [file join $drive VIDEO_TS/VIDEO_TS.IFO]]
			if {$crc32 == ""} {
				set step "Error"
				set errmsg "Cannot retrieve the CRC32 from ${drive}VIDEO_TS\\VIDEO_TS.IFO !"
			}
		}
		# get region code mask from VIDEO_TS.IFO
		if {$step != "Error"} {
			set tmp [readFile [file join ${drive} "VIDEO_TS/VIDEO_TS.IFO"] auto]
			if {$tmp != ""} {
				set regionmask [get_int tmp 0x23 1]
			}
			unset tmp
		}
		# save log
		if {$step != "Error"} {
			set ::plugins::psl2::wiz::save_log 1
			wiz_1_3a_2 [file nativename $output_logfilename]
			set step [wiz_wait_button 2_1]
			if {$::plugins::psl2::wiz::save_log} {
				saveFile $output_logfilename auto $logdata
				if {! [catch {package require registry}]} {
					set keyname {HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU\log}
					if {! [catch {set mrulist [registry get $keyname MRUList]}] && $mrulist != ""} {
						catch {registry set $keyname [string range $mrulist 0 0] [file nativename $output_logfilename]}
					}
				}
				set ::plugins::psl2::default_logfile $output_logfilename
			}
		}
	}

	# No I/O errors
	if {$step == "1_3b"} {
		# assume we don't need the log file
		set ::plugins::psl2::wiz::save_log 0
		# verify if there are problems in the log file
		if {[string first "Pack Header Not Found" $logdata] != -1} {
			set ::plugins::psl2::wiz::save_log 1
		}
		if {$::plugins::psl2::wiz::save_log} {
			set txt "Warning: Errors reported in the DVD Decrypyer log file!\n"
			set txt "${txt}Are you sure AnyDVD or DVD43 was not running when during the rip?\n\n"
			set txt "${txt}The wizard will continue anyway, but at your own risk!"
			tk_messageBox -title "PSL2 plugin - Wizard" -icon warning -message $txt
		}

		wiz_$step [file nativename $output_logfilename]
		set step [wiz_wait_button 3_prep1]
		if {$::plugins::psl2::wiz::save_log} {
			saveFile $output_logfilename auto $logdata
			if {! [catch {package require registry}]} {
				set keyname {HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU\log}
				if {! [catch {set mrulist [registry get $keyname MRUList]}] && $mrulist != ""} {
					catch {registry set $keyname [string range $mrulist 0 0] [file nativename $output_logfilename]}
				}
			}
			set ::plugins::psl2::default_logfile $output_logfilename
		}
	}

	# we have now the LOG file (in $dvdd_logfilename and $logdata)
	# and the crc32 value.
	# It's time to begin phase 2.

	######## Phase 2: Generate PSL file ########

	# Step called only from the first wizard screen,
	# to input the original log filename
	if {$step == "2_prep1"} {
		wiz_$step
		set default_logfile [find_last_logfile]
		set dvdd_logfilename [tk_getOpenFile \
			-title "Open original DVD Decrypter log file" \
			-initialdir  [file dirname $default_logfile] \
			-initialfile [file tail $default_logfile] \
			-parent $w \
			-filetypes {{"LOG files" {*.log}} {"All files" {*.*}}}]
		if {$dvdd_logfilename == ""} {
			set step "Cancel"
		} else {
			set dvdd_logfilename [file nativename $dvdd_logfilename]
			set logdata [readFile $dvdd_logfilename auto]
			if {$logdata == ""} {
				set step "Error"
				set errmsg "Cannot open or read the original DVD Decrypyer log file\n\"$dvdd_logfilename\"!"
			} else {
				set logdata [readFile $dvdd_logfilename auto]
				if {$logdata == ""} {
					set step "Error"
					set errmsg "Cannot open or read the DVD Decrypyer log file \"$dvdd_logfilename\"!"
				}
				if {$step != "Error"} {
					if {[string first "Operation Successfully Completed" $logdata] == -1} {
						set step "Error"
						set errmsg "Log file incomplete !\nYou must let DVD Decrypter save all files it can rip !"
					}
				}
				set ::plugins::psl2::default_logfile $dvdd_logfilename

				# retrieve DVD folder name as default for next step
				set idx [expr [string first "Destination Folder:" $logdata] + 20]
				set tmp [string range $logdata $idx [expr $idx+200]]
				set tmp [string range $tmp 0 [expr [string first \n $tmp] - 1]]
				set dvd_dirname [file nativename $tmp]
				set dvd_dirname_sure false
				set ::config(initialdir) $dvd_dirname

				set step 2_prep2
			}
		}
	}

	# Step called only from the first wizard screen,
	# to read the CRC32 and the region code mask in the original DVD
	if {$step == "2_prep2"} {
		set warning false
		while 1 {
			wiz_2_prep2 $warning
			set step [wiz_wait_button 2_1]
			if {$step == "Cancel"} break
			set driveslist [find_dvd_video_drives]
			if {[llength $driveslist] == 0} {
				set warning true
				bell
			} elseif {[llength $driveslist] > 1} {
				# More than one DVD-Video loaded: ask which drive to use
				# retrieve default drive from log
				set ::plugins::psl2::wiz::drive [lindex $driveslist 0]
				set sourceselected false
				set idx [string first "Source Device:" $logdata]
				if {$idx != -1} {
					set tmp [string range $logdata $idx [expr $idx+100]]
					set idx [string first ":)" $tmp]
					incr idx -1
					if {$idx > -1} {
						set driveletter [string range $tmp $idx $idx]
						if {[string first $driveletter $driveslist] != -1} {
							set ::plugins::psl2::wiz::drive $driveletter:\\
							set sourceselected true
						}
					}
				}
				# ask which drive to use, and continue
				wiz_2_prep3 $driveslist $sourceselected
				set step [wiz_wait_button 2_1]
				if {$step != "Cancel"} {
					set driveslist [list $::plugins::psl2::wiz::drive]
				} else {
					break
				}
			}
			if {[llength $driveslist] == 1} {
				set step 2_1
				set drive [lindex $driveslist 0]
				set crc32 [crc32 [file join $drive VIDEO_TS/VIDEO_TS.IFO]]
				if {$crc32 == ""} {
					set step "Error"
					set errmsg "Cannot retrieve the CRC32 from ${drive}VIDEO_TS\\VIDEO_TS.IFO !"
				}
				# get region code mask from VIDEO_TS.IFO
				if {$step != "Error"} {
					set tmp [readFile [file join ${drive} "VIDEO_TS/VIDEO_TS.IFO"] auto]
					if {$tmp != ""} {
						set regionmask [get_int tmp 0x23 1]
					}
					unset tmp
				}
				break
			}
		}
	}

	# Open new DVD in PgcEdit
	if {$step == "2_1"} {
		while {1} {
			wm iconify .
			wm deiconify $w
			if {! $dvd_dirname_sure} {
				wiz_2_1_gui		;# If DVD folder might have changed, get it interactively.
			} else {
				wiz_2_1_load	;# else, load DVD directly
			}
			set step [wiz_wait_button 2_1]
			if {$step != "Cancel"} {
				wm deiconify .
				set ::config(initialdir) $dvd_dirname
				if {! $dvd_dirname_sure} {
					set rc [open_dvd_wo_messages true]
					if {$rc == 0 && ! [file writable [file join $::dvddir VIDEO_TS.IFO]]} {
						set txt "Error:\nYou must select the DVD FOLDER RIPPED ON YOUR HARD DISC!\n\nPlease retry..."
						tk_messageBox -message $txt -icon error -title "PSL2 plugin"
						continue
					}
				} else {
					set ::dvddir $dvd_dirname
					set rc [open_dvd_wo_messages false]
				}
				if {$rc != 0} {
					set step "Error"
					set errmsg "Error opening the DVD!"
					break
				} else {
					set ::config(initialdir) $::dvddir
					set step 2_2
					break
				}
			} else {
				break
			}
		}
		wm iconify .
		wm deiconify $w
	}

	# Input DVD name.
	if {$step == "2_2"} {
		set name [build_dvd_name false]
		if {[info exists regionmask]} {
			set regioncodes ""
			for {set i 0} {$i < 6} {incr i} {
				set bitmask [expr {round(pow(2,$i))}]
				if {[expr {$regionmask & $bitmask}] == 0} {
					set regioncodes "${regioncodes} [expr $i+1]"
				}
			}
			set name "$name - Zone"
			if {[llength $regioncodes] > 1} {
				set name "${name}s"
			}
			set name "${name}$regioncodes"
		}
		set ::plugins::psl2::wiz::name $name
		wiz_$step
		set step [wiz_wait_button 2_3]
	}

	# Generate PSL file.
	if {$step == "2_3"} {
		wiz_$step
		refresh_gui
		wm deiconify .
		# generate list of protected PGCs
		set protectedtitleslist [find_protected_titles false false]
		if {[llength $protectedtitleslist] == 0} {
			set step "Error"
			set errmsg "Cannot find any protected title in this DVD!\n\nIf you see this window, the DVD is probably not protected by ARccOS.\nYou had just an ordinary I/O error. Try to clean the disc or replace it.\n(Or have you already applied phase 3 of the wizard or the Remove Crap manual pass?)"
		}

		if {$step != "Error"} {
			set outputpsldata ""
			set errorslist {}
			# for each PGC
			foreach protectedtitle $protectedtitleslist {
				set vts [lindex $protectedtitle 0]
				set pgc [lindex $protectedtitle 1]
				set numcells $::pgcs($vts,$pgc,numcells)

				# select the PGC in PgcEdit
				::main::display_pgc pgc $vts $pgc

				set tmp [read_domain_lba_from_log "VTS_[format %02d $vts]_1.VOB" $logdata]
				if {$tmp != ""} {
					set domainlba $tmp
				} else {
					lappend errorslist $protectedtitle
					continue	;# TODO: This is abnormal => display error message!
				}

				# Clear cells selected flags
				for {set c 1} {$c <= $numcells} {incr c} {
					set ::plugins::psl2::cb($c) 0
				}
				find_arcoss_cells $vts $pgc $::pgcs($vts,$pgc,numcells) $::pgcs($vts,$pgc,cellpbtable) \
						::pgcs($vts,$pgc) $::pgcs($vts,$pgc,programmap) 1

				set outputpsldata $outputpsldata\n[generate_psl2_data $vts $pgc $domainlba 1 true 0 true]
			}
			set outputpsldata [reorder_psl_file \
				"0:$crc32,0,$::plugins::psl2::wiz::name\n$outputpsldata"]

			if {$errorslist != ""} {
				set errmsg "Cannot retrieve VOB start sector from original DVD Decrypter log file\nfor VTS"
				foreach i $errorslist {
					set errmsg "$errmsg [lindex $i 0],"
				}
				set errmsg [string trim $errmsg ,].
				if {[llength $errorslist] == [llength $protectedtitleslist]} {
					set errmsg "$errmsg\nThe output PSL file is therefore empty!"
					set step "Error"
				} else {
					set errmsg "$errmsg\nThe output PSL file may be incomplete!\n\nContinuing anyway..."
					tk_messageBox -message $errmsg -icon error -title "PSL2 plugin"
					set step 2_4
				}
			} else {
				set step 2_4
			}
		}
		wm iconify .
		wm deiconify $w
	}

	unset -nocomplain logdata
	unset -nocomplain psldata

	# Save output PSL.
	if {$step == "2_4"} {
		wiz_$step

		set name [::utils::get_General_Name]
		if {$name != ""} {
			# remove illegal characters in output filename
			set name [string map {: _ / _ \\ _ | _ < _ > _ \" _} $name]
		} else {
			set tmp [file tail $::dvddir]
			if {$tmp != "VIDEO_TS" && [string first : $tmp]==-1} {
				set name $tmp
			} else {
				set tmp [file tail [file dirname $::dvddir]]
				if {[string first : $tmp]==-1} {
					set name $tmp
				}
			}
		}
		if {$name == ""} {
			set name DVD_Video
		}
		set default_pslfile [file join [file dirname $::plugins::psl2::default_logfile] $name.PSL]
		set output_pslfilename [tk_getSaveFile \
				-title "Save new PSL file for DVD Decrypter" \
				-initialdir  [file dirname $default_pslfile] \
				-initialfile [file tail $default_pslfile] \
				-parent $w \
				-filetypes {{"PSL files" {*.psl}} {"All files" {*.*}}}]
		if {$output_pslfilename == ""} {
			set step "Cancel"
		} else {
			set rc [saveFile $output_pslfilename auto $outputpsldata]
			if {$rc != 0} {
				set step "Error"
				set errmsg "Error saving \"$output_pslfilename\"!"
			} else {
				# force the default folder of the Open PSL dialog of DVDD to be
				# the folder used to save the PSL.
				if {! [catch {package require registry}]} {
					set keyname "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\LastVisitedMRU"
					if {! [catch {set mrulist [registry get $keyname MRUList]}]} {
						set found ""
						set string "D\000V\000D\000D\000e\000c\000r\000y\000p\000t\000e\000r\000.\000e\000x\000e\000\000\000"
						for {set idx 0} {$idx < [string length $mrulist]} {incr idx} {
							set item [string range $mrulist $idx $idx]
							if {! [catch {set entry [registry get $keyname $item]}]} {
								if {[string first $string $entry] != -1} {
									set found $item
								}
							}
						}
						if {$found == "" && $mrulist != ""} {
							set found [string range $mrulist 0 0]
						}
						if {$found != ""} {
							set dirname [file nativename [file dirname $output_pslfilename]]
							for {set i 0} {$i < [string length $dirname]} {incr i} {
								set string "$string[string range $dirname $i $i]\000"
							}
							set string "$string\000\000"
							catch {registry set $keyname $found $string binary}
						}
					}
				}

				set step 2_5_1
			}
		}
	}

	unset -nocomplain outputpsldata

	# Explain what to do with DVD Decrypter.
	if {$step == "2_5_1"} {
		wiz_$step $output_pslfilename
		set step [wiz_wait_button 2_5_2]
	}

	# Start DVD Decrypter again to rip the remaining files.
	if {$step == "2_5_2"} {
		# delete the IFO and BUP files to force DVDD to re-rip them
		if {! [catch {set ifofiles [glob -nocomplain -tails -directory $::dvddir -- *.IFO]}]} {
			foreach f $ifofiles {
				catch {file delete -force [file join $::dvddir $f]}
			}
		}
		if {! [catch {set ifofiles [glob -nocomplain -tails -directory $::dvddir -- *.BUP]}]} {
			foreach f $ifofiles {
				catch {file delete -force [file join $::dvddir $f]}
			}
		}
		unset -nocomplain ifofiles

		wiz_$step

		# remember the DVDD setups we need to change
		set_dvdd_regkeys set
		wm iconify $w	;# optional!
		refresh_gui
		set txt "Before launching the rip, don't forget to load the PSL file\n\"$output_pslfilename\"\n"
		set txt "${txt}with the menu \"File -> Protected Sectors List -> Import\"!\n\n"
		set txt "${txt}Remember:\n\nWhen asked if you want to replace the existing files,\n"
		set txt "${txt}answer \"No to All\".\n\n"
		set txt "${txt}When the rip is finished, close DVD Decrypter."
		set req_pid [open_background_msgbox "PSL2 Plugin - Wizard - Phase 2, step 5" $txt 2000]
		set rc [launch_dvdd FILE false $dvd_dirname]
		set_dvdd_regkeys restore
		set req_pid [close_background_msgbox $req_pid]
		wm deiconify $w
		if {$rc != 0} {
			set step "Error"
			set errmsg "Cannot launch DVD Decrypter!\nReinstalling DVD Decrypter may solve the problem."
		} else {
			set step 2_6
		}
	}

	# Reload the DVD in PgcEdit.
	if {$step == "2_6"} {
		set f [file join $::dvddir [format "VTS_%02d_0.IFO" $::pgcs(numvts)]]
		if {! [file exists $f]} {
			wm deiconify .
			set old_expertmode $::config(expert_mode)
			set ::config(expert_mode) true
			::main::Write_DVD false
			set ::config(expert_mode) $old_expertmode
			wm iconify .
			set step "Error"
			set errmsg "IFO file missing!\nThe plugin has resaved the IFO and BUP files.\n\nPlease redo phase 2 to recreate the missing VOB files."
		} else {
			wiz_$step
			set step [wiz_wait_button 3_1]
			if {$step != "Cancel"} {
				wm deiconify .
				if {[open_dvd_wo_messages false] != 0} {
					set step "Error"
					set errmsg "Error opening the DVD!"
				} else {
					set phase2done true
				}
				# remove old PgcEdit_backup, now obsolete, and create a new one
				set f [file join $::dvddir "PgcEdit_backup"]
				if {[file exists $f]} {
					catch {file delete -force $f}
				} else {
					set f [file join [file dirname $::dvddir] "PgcEdit_backup"]
					if {[file exists $f]} {
						catch {file delete -force $f}
					}
				}
				::main::Create_Backup false
				wm iconify .
			}
			wm deiconify $w
		}
	}

	######## Phase 3: Remove Crap ########

	# Open new DVD in PgcEdit. (Called only from the first screen of the wizard
	# or from phase 1 when there is no need to apply phase 2)
	if {$step == "3_prep1"} {
		while {1} {
			wm iconify .
			wm deiconify $w
			if {! $dvd_dirname_sure} {
				wiz_3_prep1_gui		;# If DVD folder might have changed, get it interactively.
			} else {
				wiz_3_prep1_load	;# else, load DVD directly
			}
			set step [wiz_wait_button 3_1]
			if {$step != "Cancel"} {
				wm deiconify .
				if {! $dvd_dirname_sure} {
					set rc [open_dvd_wo_messages true]
					if {$rc == 0 && ! [file writable [file join $::dvddir VIDEO_TS.IFO]]} {
						set txt "Error:  The DVD files are read-only!\n\nYou must select the DVD FOLDER RIPPED ON YOUR HARD DISC!\n\nPlease retry..."
						tk_messageBox -message $txt -icon error -title "PSL2 plugin"
						continue
					}
				} else {
					set ::dvddir $dvd_dirname
					set rc [open_dvd_wo_messages false]
					set ::config(initialdir) $dvd_dirname
				}
				if {$rc != 0} {
					set step "Error"
					set errmsg "Error opening the DVD!"
					break
				} else {
					set ::config(initialdir) $::dvddir
					set step 3_1
					break
				}
			} else {
				break
			}
		}
		wm iconify .
		wm deiconify $w
	}

	# Remove crap in the IFOs
	if {$step == "3_1"} {
		# generate list of protected PGCs
		set protectedtitleslist [find_protected_titles false false]
		if {[llength $protectedtitleslist] == 0} {
			set step "Error"
			set errmsg "Cannot find any protected title in this DVD!\n\nIf you see this window, the DVD is probably not protected by ARccOS.\nYou had just an ordinary I/O error. Try to clean the disc or replace it.\n(Or have you already applied phase 3 of the wizard or the Remove Crap manual pass?)"
		} else {
			# safety check to ensure that rip was successfull
			if {$::config(debug)==0} {
				foreach i $protectedtitleslist {
					set vobfilename [format VTS_%02d_1.VOB [lindex $i 0]]
					if {! [file exists [file join $::dvddir $vobfilename]]} {
						set step "Error"
						set errmsg "Cannot find $vobfilename !\nYou must do phases 1 and 2 of the wizard before phase 3 !"
						break
					}
				}
			}
		}
		if {$step != "Error"} {
			wiz_$step $protectedtitleslist
			set step [wiz_wait_button 3_2]
			if {$step != "Cancel"} {
				wm deiconify .
				set errorslist {}
				set ::plugins::psl2::cleaned_titles_list {}
				# for each PGC
				foreach protectedtitle $protectedtitleslist {
					set vts [lindex $protectedtitle 0]
					set pgc [lindex $protectedtitle 1]
					set unreferenced [lindex $protectedtitle 2]
					set numcells $::pgcs($vts,$pgc,numcells)

					# select the PGC in PgcEdit
					::main::display_pgc pgc $vts $pgc

					# If the current PGC is a fake PGC, simply remove it
					set rc false
					if {$unreferenced} {
						set rc [remove_if_fake_pgc $vts $pgc true]
					}

					if {! $rc} {
						# Clear cells selected flags
						for {set c 1} {$c <= $numcells} {incr c} {
							set ::plugins::psl2::cb($c) 0
						}
	
						# finds all cells to remove
						find_arcoss_cells $vts $pgc $numcells $::pgcs($vts,$pgc,cellpbtable) \
								::pgcs($vts,$pgc) $::pgcs($vts,$pgc,programmap) 1
						find_tiny_cells   $numcells $::pgcs($vts,$pgc,cellpbtable) \
								$::pgcs($vts,$pgc,programmap) true
	
						# and remove crap on the current PGC
						if {[remove_crap $vts $pgc 1 true] == ""} {
							lappend errorslist $protectedtitle
						}
					}
				}
				# Save IFOs, + incremental backup
				set old_expertmode $::config(expert_mode)
				set ::config(expert_mode) true
				if {[::main::Write_DVD false] != 0} {
					set step "Error"
					set errmsg "Error saving the IFOs!"
				} else {
					::main::Incremental_Backup
				}
				set ::config(expert_mode) $old_expertmode

				wm iconify .
				wm deiconify $w
				if {$errorslist != ""} {
					set errmsg "Error removing crap on the following PGCs:\n"
					foreach i $errorslist {
						set errmsg "${errmsg}VTST [lindex $i 0], [lindex $i 0],  "
					}
					set errmsg "[string trim $errmsg ",  "].\n\nContinuing anyway..."
					tk_messageBox -message $errmsg -icon error -title "PSL2 plugin"
				}
			}
		}
	}

	# Use FixVTS or VobBlanker to fix the VOBs?
	if {$step == "3_2"} {
		wiz_$step
		set step [wiz_wait_button x_1]
		if {$step != "Cancel"} {
			set step 3_3${::plugins::psl2::wiz::wizphase}_1
		}
	}

	######## use FixVTS to remove crap in the VOBs #########

	# Retrieve fixvts.exe filename
	if {$step == "3_3a_1"} {
		wiz_$step
		refresh_gui
		set fixvts_exe [find_exe FixVTS]
		if {$fixvts_exe == ""} {
			set step "Error"
			set errmsg "Cannot find the FixVTS program.\nYou should put it in the \"plugins\" or \"bin\" folder of PgcEdit!"
		} else {
			set step 3_3a_2
		}
	}

	# Use FixVTS to fix the VOBs
	if {$step == "3_3a_2"} {
		wiz_$step
		wm iconify $w
		set step 3_3a_3

		for {set i 1} {$i <= $::pgcs(numvts)} {incr i} {
			# searches in the list of cleaned titles if we need to process this VTS
			set cleanit false
			foreach cleaned $protectedtitleslist {
				if {$i == [lindex $cleaned 0]} {
					set cleanit true
					break
				}
			}
			if {$cleanit} {
				# launch fixvts
				set vobfilename [file nativename [file join $::dvddir [format "VTS_%02d_1.VOB" $i]]]
				if {$::plugins::psl2::wiz::full_process} {
					if {$::plugins::psl2::wiz::remove_cells} {
						if {$::plugins::psl2::wiz::in_place} {
							set rc [catch {exec -- $fixvts_exe -s -f -u -i $vobfilename} errstr]
						} else {
							set rc [catch {exec -- $fixvts_exe -s -f -u $vobfilename} errstr]
						}
					} else {
						if {$::plugins::psl2::wiz::in_place} {
							set rc [catch {exec -- $fixvts_exe -s -f -i $vobfilename} errstr]
						} else {
							set rc [catch {exec -- $fixvts_exe -s -f $vobfilename} errstr]
						}
					}
				} else {
					if {$::plugins::psl2::wiz::remove_cells} {
						if {$::plugins::psl2::wiz::in_place} {
							set rc [catch {exec -- $fixvts_exe -s -u -i $vobfilename} errstr]
						} else {
							set rc [catch {exec -- $fixvts_exe -s -u $vobfilename} errstr]
						}
					} else {
						if {$::plugins::psl2::wiz::in_place} {
							set rc [catch {exec -- $fixvts_exe -s -i $vobfilename} errstr]
						} else {
							set rc [catch {exec -- $fixvts_exe -s $vobfilename} errstr]
						}
					}
				}
				if {$rc} {
					set step "Error"
					set errmsg "FixVTS reported an error when fixing VTS $i!"
					break
				}
			}
		}
		wm deiconify $w
	}

	# Reopen DVD in PgcEdit + save DVD once again + incremental backup
	if {$step == "3_3a_3"} {
		wiz_$step
		refresh_gui
		focus $w
		set step 3_4
		if {$step != "Cancel"} {
			wm deiconify .
			if {[open_dvd_wo_messages false] != 0} {
				set step "Error"
				set errmsg "Error opening the DVD!"
			}
			set old_expertmode $::config(expert_mode)
			set ::config(expert_mode) true
			if {[::main::Write_DVD false] != 0} {
				set step "Error"
				set errmsg "Error saving the IFOs!"
			} else {
				::main::Incremental_Backup
			}
			set ::config(expert_mode) $old_expertmode
			wm iconify .
		}
		wm deiconify $w
		bell
	}

	######## use VobBlanker to remove crap in the VOBs #########

	# Retrieve VobBlanker.exe filename
	if {$step == "3_3b_1"} {
		wiz_$step
		refresh_gui
		set vobblanker_exe [find_exe VobBlanker]
		if {$vobblanker_exe == ""} {
			set step "Error"
			set errmsg "Cannot find the VobBlanker program.\nYou should install it!"
		} else {
			set step 3_3b_2
		}
	}

	# Instructions on how to use VobBlanker
	if {$step == "3_3b_2"} {
		wiz_$step $protectedtitleslist
		set step [wiz_wait_button 3_3b_3]
	}

	# Launch VobBlanker to fix the VOBs
	if {$step == "3_3b_3"} {
		wiz_$step
		set step 3_4
		refresh_gui
		wm iconify $w	;# optional!
		set txt "Remember:\n\nWhen VobBlanker has loaded the DVD, click \"PROCESS!!\".\n\n"
		set txt "${txt}When it has finished, click the \"QUIT\" button."
		set req_pid [open_background_msgbox "PSL2 Plugin - Wizard - Phase 3, step 3b" $txt 2500]
		set success [vobblanker_process $::plugins::psl2::cleaned_titles_list \
				$::plugins::psl2::wiz::force_fix_pointers $vobblanker_exe]
		set req_pid [close_background_msgbox $req_pid]
		wm iconify .
		wm deiconify $w
		if {! $success} {
			set step "Error"
			set errmsg "Error reported by VobBlanker, or operation aborted!"
		}
	}

	########### Launch DVD Shrink to verify the DVD? ###########

	if {$step == "3_4"} {
		set dvdshrink_exe [find_dvdshrink_exe]
		if {$dvdshrink_exe != ""} {
			wiz_$step
			set step [wiz_wait_button Finished]
			if {$step != "Cancel"} {
				if {$::plugins::psl2::wiz::dvdshrink} {
					set fname [file nativename [file join $::dvddir VIDEO_TS.IFO]]
					catch {exec -- $dvdshrink_exe $fname &}
				}
			}
		} else {
			set step Finished
		}
	}

	######## Finished! #########

	# End options
	if {$step == "Finished"} {
		if {$phase2done} {
			set del_log true
			set del_psl true
		} else {
			set del_log false
			set del_pal false
			set tmp [file join [file dirname $::dvddir] "DVDDecrypter.log"]
			if {[file exists $tmp]} {
				set del_log true
				set output_logfilename $tmp
			}
			if {! [catch {set tmp [glob -nocomplain -directory [file dirname $::dvddir] -- *.psl]}]} {
				set del_psl true
				set output_pslfilename [lindex $tmp 0]
			}
		}
		wiz_$step $del_log $del_psl
		set step [wiz_wait_button Done]
		if {$step != "Cancel"} {
			if {$::plugins::psl2::wiz::del_backups} {
				if {[file exists [file join $::dvddir FixVTSBackup]]} {
					catch {file delete -force [file exists [file join $::dvddir FixVTSBackup]]}
				}
				if {! [catch {set backupdirs [glob -nocomplain -directory $::dvddir -- VobBlanker_backup*]}]} {
					foreach f $backupdirs {
						catch {file delete -force $f}
					}
				}
				if {! [catch {set backupdirs [glob -nocomplain -directory [file dirname $::dvddir] -- VobBlanker_backup*]}]} {
					foreach f $backupdirs {
						catch {file delete -force $f}
					}
				}
				if {! [catch {set backupdirs [glob -nocomplain -directory $::dvddir -- PgcEdit_backup* VobBlanker_backup* FixVTSBackup*]}]} {
					foreach f $backupdirs {
						catch {file delete -force $f}
					}
				}
				if {$::config(backup_in_root)} {
					if {! [catch {set backupdirs [glob -nocomplain -directory [file dirname $::dvddir] -- PgcEdit_backup*]}]} {
						foreach f $backupdirs {
							catch {file delete -force $f}
						}
					}
				}
			}
			if {$::plugins::psl2::wiz::del_files} {
				catch {file delete -force $dvdd_logfilename}
				catch {file delete -force $output_logfilename}
			}
			if {$::plugins::psl2::wiz::del_outputpsl} {
				catch {file delete -force $output_pslfilename}
			}
			if {$::plugins::psl2::wiz::open_folder} {
				::macros::explore_dvd
			}
			if {$::plugins::psl2::wiz::close_pgcedit} {
				set step ClosePgcEdit
			} else {
				set step Done
			}
		}
	}

	######## Special step called upon error ########
	if {$step == "Error"} {
		wiz_Error $errmsg
		wiz_wait_button Next
	}

	######## Last step, or Cancel: cleanup and quit wizard ########

	set_dvdd_regkeys restore
	close_background_msgbox $req_pid
	if {$step == "ClosePgcEdit"} {
		cleanexit 0
	}
	destroy $w
	namespace delete ::plugins::psl2::wiz
	wm deiconify .
	return
}

# Wait for user action (Cancel, Next or close button clicked)
proc wiz_wait_button {next_step} {
	tkwait variable ::plugins::psl2::wiz::wizarddone
	if {$::plugins::psl2::wiz::wizarddone == "Next"} {
		return $next_step
	} else {
		return $::plugins::psl2::wiz::wizarddone
	}
}

# Configure the main wizard window
proc wiz_config_win {step txt {b1state normal} {b2state normal}} {
	if {$step != ""} {
		set steplist [split $step _]
		set title "PSL2 plugin - Wizard - Phase [lindex $steplist 0]"
		if {[llength $steplist] > 1} {
			set title "${title}, step [lindex $steplist 1]"
		}
		wm title .psl2wizard $title
	}
	set mf .psl2wizard.f1.mf
	destroy $mf
	pack [frame $mf] -side left -fill x -expand 1
	pack [label $mf.l -text $txt -justify left -anchor w] -anchor w
	.psl2wizard.bf.b1 configure -state $b1state
	.psl2wizard.bf.b2 configure -state $b2state
	if {$b1state == "normal"} {
		wm protocol .psl2wizard WM_DELETE_WINDOW {set ::plugins::psl2::wiz::wizarddone "Cancel"}
	} else {
		wm protocol .psl2wizard WM_DELETE_WINDOW {bell}
	}
}

######## Routines to Configure the Wizard for the various steps ########

proc wiz_0_1 {} {
	set txt "\nWelcome to the PSL2 (anti-ARccOS) wizard v $::plugins::psl2::version.\n"
	set txt "${txt}\n\nThis wizard will help you create a PSL file to decrypt an ARccOS protected DVD with DVD Decrypter v3.5.4.0,\n"
	set txt "${txt}and remove the crap stuff so that the DVD will be clean and accepted by applications such as DVD Shrink.\n\n\n\n\n"
	wiz_config_win "" $txt

	set mf .psl2wizard.f1.mf

	set txt "Automatically start the PSL2 wizard when PgcEdit is launched?"
	checkbutton $mf.r1 -variable ::plugins::psl2::autostart -onvalue true -offvalue false -text $txt
	pack $mf.r1 -anchor w

	wm title .psl2wizard "PSL2 plugin - Wizard - Welcome"
	.psl2wizard.bf.b2 configure -text "Next"
}

proc wiz_0_2 {} {
	set txt "Please select the phase you want to begin with.  When done, press the Next button."
	wiz_config_win "" $txt

	set mf .psl2wizard.f1.mf

	set txt "Phase 1:  Rip the DVD from scratch\nI haven't ripped the DVD yet.  I want to rip it now."
	radiobutton $mf.r1 -variable ::plugins::psl2::wiz::wizphase -value 1 -text $txt -justify left
	pack $mf.r1 -anchor w
	set txt "Phase 2:  Complete my failed rip\nI have ripped the DVD, but there are some VOB files with I/O errors that DVD Decrypter cannot rip.\n"
	set txt "${txt}I have saved the DVD Decrypter log file manually or when applying phase 1 of the wizard.\n"
	set txt "${txt}I want to generate the final PSL file for DVD Decrypter."
	radiobutton $mf.r2 -variable ::plugins::psl2::wiz::wizphase -value 2 -text $txt -justify left
	pack $mf.r2 -anchor w
	set txt "Phase 3:  Cleanup\nI have successfully ripped all files of the DVD.  I want to clean it now."
	radiobutton $mf.r3 -variable ::plugins::psl2::wiz::wizphase -value 3 -text $txt -justify left
	pack $mf.r3 -anchor w
	set ::plugins::psl2::wiz::wizphase 1

	wm title .psl2wizard "PSL2 plugin - Wizard - Select startup phase"
}

proc wiz_1_1 {} {
	set txt "IMPORTANT: If AnyDVD or DVD43 is running in the background, close it now !\n\n"
	set txt "${txt}IMPORTANT:  Do NOT start DVD Decrypter yourself !  If it is already running, close it now !\n\n"
	set txt "${txt}Insert your DVD in the drive now.  If Windows starts an application or opens its AutoPlay window, close it !\n\n"
	set txt "${txt}When done, press the Next button."
	wiz_config_win 1_1 $txt
}

proc wiz_1_2_a {} {
	set txt "DVD Decrypter will now be launched automatically by the wizard.\n\n"
	set txt "${txt}Do not change his settings (except, if you want, the source drive or destination folder.)\n"
	set txt "${txt}Be sure to rip the DVD in a new folder !\n\n"
	set txt "${txt}Start the rip by clicking on the large button.\n\n"
	set txt "${txt}IMPORTANT:  If DVD Decrypter encounters a read error, an I/O Error dialog will be shown.\n"
	set txt "${txt}In this case, click on the \"Abort\" button, and let DVD Decrypter continue to rip the remaining files.\n\n"
	set txt "${txt}IMPORTANT:  When the rip is finished, do not remove the DVD from the drive, but close DVD Decrypter.\n"
	set txt "${txt}The wizard will then continue automatically.\n\n"
	set txt "${txt}Now, click the Next button to launch DVD Decrypter."
	wiz_config_win 1_2_a $txt
}

proc wiz_1_2_b {} {
	set txt "\n\n\nDVD Decrypter should be running now.\n\n"
	set txt "${txt}When done, close it to continue."
	wiz_config_win 1_2_b $txt disabled disabled
}

proc wiz_1_3a_1 {} {
	set txt "Computing CRC32 value of tile VIDEO_TS.IFO.\nPlease wait..."
	wiz_config_win 1_3a_1 $txt disabled disabled
}

proc wiz_1_3a_2 {{logfilename ""}} {
	set txt "Since there were I/O errors during the rip, you need a PSL file to rip the protected files.\n\n\n"
	set txt "${txt}This step is the last one of the first phase.  Phase 2 will begin now.\n\n"
	wiz_config_win 1_3a_2 $txt

	set mf .psl2wizard.f1.mf
	set txt "Save DVD Decrypter log file in DVD output folder?\nIt will be saved as \"$logfilename\"."
	checkbutton $mf.r1 -variable ::plugins::psl2::wiz::save_log -text $txt -justify left
	pack $mf.r1 -anchor w
}

proc wiz_1_3b {{logfilename ""}} {
	set txt "Since no I/O error has been encountered during the rip, there is no need to generate a PSL file.\n\n\n"
	set txt "${txt}This step is the last one of the first phase.  Phase 2 is not needed, so phase 3 will begin now.\n\n"
	wiz_config_win 1_3b $txt

	set mf .psl2wizard.f1.mf
	set txt "Save DVD Decrypter log file in DVD output folder?\nIt will be saved as \"$logfilename\"."
	checkbutton $mf.r1 -variable ::plugins::psl2::wiz::save_log -text $txt -justify left
	pack $mf.r1 -anchor w
}

# Phase 2

proc wiz_2_prep1 {} {
	set txt "Please select the log file you have saved manually when ripping the DVD with DVD Decrypter, or the log saved\n"
	set txt "${txt}automatically by the plugin in the DVD folder during phase 1 of the wizard."
	wiz_config_win 2_prep.1 $txt disabled disabled
}

proc wiz_2_prep2 {{warning true}} {
	set txt "IMPORTANT: If AnyDVD or DVD43 is running in the background, close it now !\n\n"
	set txt "${txt}Insert your DVD in the drive now.  If Windows starts an application or opens its AutoPlay window, close it !\n\n"
	if {$warning} {
		set txt "${txt}ERROR:  No DVD-Video found in your drive(s) !  Please insert the original DVD and retry !\n\n"
	}
	set txt "${txt}When done, press the Next button."
	wiz_config_win 2_prep.2 $txt
}

proc wiz_2_prep3 {{driveslist {D: E: F: G:}} {sourceselected true}} {
	set txt "There are several DVD-Videos loaded in your drives.\n\n"
	set txt "${txt}Please select the source drive with your original DVD-Video :"
	if {$sourceselected} {
		set txt "${txt}\n(Note: The source drive used for the first rip has been automatically selected.)"
	}
	wiz_config_win 2_prep.3 $txt
	set mf .psl2wizard.f1.mf
	pack [frame $mf.f] -anchor w
	foreach i $driveslist {
		pack [radiobutton $mf.f.rb[string range $i 0 0] -text $i -bd 2 -relief groove \
				-variable ::plugins::psl2::wiz::drive -value $i -padx 2 -pady 2] -side left -padx 2
	}
	set txt "When done, press the Next button."
	label $mf.l2 -text $txt -justify left
	pack  $mf.l2 -anchor w -pady 8
}

proc wiz_2_1_gui {} {
	set txt "The RIPPED DVD will now be opened in PgcEdit.\n\n"
	set txt "${txt}Navigate to the FOLDER on your hard disk with the DVD files you have already ripped with DVD Decrypter,\n"
	set txt "${txt}and click Open.  Warning:  Cancelling the open DVD dialog will terminate the wizard!\n\n"
	set txt "${txt}PgcEdit may open several warning or information dialogs.  In this case, just click OK to skip them.\n"
	set txt "${txt}(Don't worry if PgcEdit says that you should save the DVD, it will be saved anyway later.)\n\n"
	set txt "${txt}When the DVD will be loaded, the wizard will continue.\n\n"
	set txt "${txt}Click Next below to load the DVD."
	wiz_config_win 2_1_gui $txt
}

proc wiz_2_1_load {} {
	set txt "The ripped DVD will now be opened in PgcEdit.\n\n"
	set txt "${txt}PgcEdit may open several warning or information dialogs.  In this case, just click OK to skip them.\n"
	set txt "${txt}(Don't worry if PgcEdit says that you should save the DVD, it will be saved anyway later.)\n\n"
	set txt "${txt}When the DVD will be loaded, the wizard will continue.\n\n"
	set txt "${txt}Click Next below to load the DVD."
	wiz_config_win 2_1_load $txt
}

proc wiz_2_2 {} {
	set txt "Please enter the DVD name and description here.\n"
	set txt "${txt}(Movie name - DVD zone - DVD country or language, edition, rental/store version...)"
	wiz_config_win 2_2 $txt normal normal

	set mf .psl2wizard.f1.mf

	entry $mf.e -textvariable ::plugins::psl2::wiz::name -validate key \
		-validatecommand {::plugins::psl2::validate_dvd_name {%P}}
	pack  $mf.e -expand 1 -fill x -padx 4 -pady 4

	set txt "\nNotes: This information is needed to identify the DVD without ambiguity, and is used by DVD Decrypter when\n"
	set txt "${txt}when PSL file is used for another DVD.  It's not very important if you don't want to distribute your PSL file.\n\n"
	set txt "${txt}You may not use comas (\",\") in the DVD name."
	label $mf.l2 -text $txt -justify left
	pack  $mf.l2 -anchor w

	focus -force $mf.e
	$mf.e icursor end
}

proc wiz_2_3 {} {
	set txt "Generating PSL file.\n\nPlease wait..."
	wiz_config_win 2_3 $txt normal disabled
}

proc wiz_2_4 {} {
	set txt "Please save the PSL file.\n"
	set txt "${txt}This is the PSL file you will need to rip the protected files remaining on the DVD.\n\n"
	set txt "${txt}Warning:  Cancelling the file open dialog will abort the wizard!"
	wiz_config_win 2_4 $txt disabled disabled
}

proc wiz_2_5_1 {{pslfilename ""}} {
	set txt "DVD Decrypter will be restarted now.\n\n"
	set txt "${txt}IMPORTANT:  Open the PSL file \"$pslfilename\"\n"
	set txt "${txt}with the menu \"File -> Protected Sectors List -> Import\", and then launch the rip!\n"
	set txt "${txt}(Be sure to rip the DVD in the same folder than the one used for the first rip!)\n\n"
	set txt "${txt}To speed up the rip process, when asked if you want to replace the existing files, answer \"No to All\".\n\n"
	set txt "${txt}NOTE:  The IFO files will be deleted now to force DVD Decrypter to recreate them!\n"
	set txt "${txt}Don't stop DVD Decrypter before the end of the rip or the DVD will be unreadable!\n\n"
	wiz_config_win 2_5_1 $txt
}

proc wiz_2_5_2 {} {
	set txt "DVD Decrypter should be running now..."
	wiz_config_win 2_5_2 $txt disabled disabled
}

proc wiz_2_6 {} {
	set txt "At this point, DVD Decrypter should have finished the rip process successfully!\n\n"
	set txt "${txt}(If you still got some I/O errors when ripping the files, it's probably because your DVD is dirty or scratched,\n"
	set txt "${txt}or because your drive is bad, or you forgot to import the PSL file.\n"
	set txt "${txt}In this case, you should cancel the wizard, and try to clean your DVD.  Then, retry phase 2.)\n\n"
	set txt "${txt}Since the original DVD is not needed anymore, remove it from the drive now!\n\n"
	set txt "${txt}The ripped DVD will now be reloaded in PgcEdit.  As usual, skip the dialogs.\n\n\n"
	set txt "${txt}This is the last step of phase 2 of the wizard.  Next phase: Remove crap!"
	wiz_config_win 2_6 $txt
}

# Phase 3

proc wiz_3_prep1_gui {} {
	set txt "The DVD FOLDER will now be opened in PgcEdit.\n\n"
	set txt "${txt}Click Next below to load the DVD.\n"
	set txt "${txt}Navigate to the FOLDER on your hard disk with the DVD files you have ripped with DVD Decrypter,\n"
	set txt "${txt}and click Open.  Warning:  Cancelling the open DVD dialog will terminate the wizard!\n\n"
	set txt "${txt}PgcEdit may open several warning or information dialogs.  In this case, just click OK.\n"
	set txt "${txt}(Don't worry if PgcEdit says that you should save the DVD, it will be saved anyway later.)\n\n"
	set txt "${txt}When the DVD will be loaded, the wizard will continue."
	wiz_config_win 3_prep.1_gui $txt
}

proc wiz_3_prep1_load {} {
	set txt "The DVD FOLDER will now be opened in PgcEdit.\n\n"
	set txt "${txt}Click Next below to load the DVD.\n\n"
	set txt "${txt}PgcEdit may open several warning or information dialogs.  In this case, just click OK.\n"
	set txt "${txt}(Don't worry if PgcEdit says that you should save the DVD, it will be saved anyway later.)\n\n"
	set txt "${txt}When the DVD will be loaded, the wizard will continue."
	wiz_config_win 3_prep.1_load $txt
}

proc wiz_3_1 {{protectedtitleslist {}}} {
	set numprotectedtitles [llength $protectedtitleslist]
	if {$numprotectedtitles > 10} {
		set protectedtitleslist [lrange $protectedtitleslist 0 8]
		set listnotcomplete 1
	} else {
		set listnotcomplete 0
	}
	set txt "The wizard will now remove the fake cells and the crap in the following PGC(s):\n\n"
	foreach i $protectedtitleslist {
		if {[lindex $i 2]} {
			set unref "    (Fake unreferenced title!)"
		} else {
			set unref ""
		}
		set txt "${txt}[::main::pgc_label $::config(global_font) pgc [lindex $i 0] [lindex $i 1]]$unref\n"
	}
	if {$listnotcomplete} {
		set txt "${txt}...  The list of title is too long!  Total: $numprotectedtitles titles to process.\n"
	}
	set txt "${txt}\nThen, the DVD will be saved, and an incremental backup will be created."
	wiz_config_win 3_1 $txt
}

proc wiz_3_2 {} {
	set txt "The IFO files are now hopefully clean.  Last step is to clean the VOB files."
	wiz_config_win 3_2 $txt

	set mf .psl2wizard.f1.mf

	frame $mf.pad1 -height 6
	pack  $mf.pad1

	set txt "Use FixVTS:  FixVTS has been specially designed to normalize bad VOB files and IFOs.\n"
	set txt "${txt}It is also very fast (specially if you chose the \"In Place\" option below.)\n"
	set txt "${txt}Since FixVTS is included in the distribution of the plugin, it should be available anyway."
	radiobutton $mf.r1 -variable ::plugins::psl2::wiz::wizphase -value a \
			-text $txt -justify left -bd 0 -pady 0 -command {
				.psl2wizard.f1.mf.cb1 configure -state normal
				.psl2wizard.f1.mf.cb2 configure -state normal
				.psl2wizard.f1.mf.cb3 configure -state normal
				.psl2wizard.f1.mf.cb4 configure -state disabled
			}
	pack $mf.r1 -anchor w -pady 0
	checkbutton $mf.cb1 -variable ::plugins::psl2::wiz::remove_cells \
		-text "Remove unreferenced and blank cells.  (Recommended, but slower)" -bd 0 -pady 0
	pack $mf.cb1 -anchor w -padx 20
	checkbutton $mf.cb2 -variable ::plugins::psl2::wiz::in_place \
		-text "Process VOBs \"in place\".  (Very fast, but FixVTS will not create a backup!)" -bd 0 -pady 0
	pack $mf.cb2 -anchor w -padx 20
	checkbutton $mf.cb3 -variable ::plugins::psl2::wiz::full_process \
		-text "Process all DVD files.  (Normally not needed, and slow)" -bd 0 -pady 0
	pack $mf.cb3 -anchor w -padx 20

	frame $mf.pad2 -height 8
	pack  $mf.pad2

	set txt "Use VobBlanker:  It is slower, but more powerful (you can blank also some stuff you don't want.)\n"
	set txt "${txt}However, VobBlanker may fail with some kind of ARccOS protections, rarely used.\n"
	set txt "${txt}To use VobBlanker, it must be installed on your computer."
	radiobutton $mf.r2 -variable ::plugins::psl2::wiz::wizphase -value b \
			-text $txt -justify left -bd 0 -pady 0 -command {
				.psl2wizard.f1.mf.cb1 configure -state disabled
				.psl2wizard.f1.mf.cb2 configure -state disabled
				.psl2wizard.f1.mf.cb3 configure -state disabled
				.psl2wizard.f1.mf.cb4 configure -state normal
			}
	pack $mf.r2 -anchor w -pady 0
	checkbutton $mf.cb4 -variable ::plugins::psl2::wiz::force_fix_pointers -state disabled \
		-text "Force \"Fix always VOB int pointers\" VobBlanker option.  (Slower but safer)" -bd 0 -pady 0
	pack $mf.cb4 -anchor w -padx 20

	set ::plugins::psl2::wiz::wizphase a
	set ::plugins::psl2::wiz::remove_cells 1
	set ::plugins::psl2::wiz::in_place 1
	set ::plugins::psl2::wiz::full_process 0
	set ::plugins::psl2::wiz::force_fix_pointers 0
}

proc wiz_3_3a_1 {} {
	set txt "Searching for the FixVTS program..."
	wiz_config_win 3_3a_1 $txt
}

proc wiz_3_3a_2 {} {
	set txt "FixVTS is now fixing the VOBs..."
	wiz_config_win 3_3a_2 $txt disabled disabled
}

proc wiz_3_3a_3 {} {
	set txt "FixVTS operation successfully completed.\n\n\n"
	set txt "${txt}The plugin reloads now the DVD in PgcEdit.  As usual, skip the dialogs.\n\n"
	set txt "${txt}Then, the DVD will be saved one more time."
	wiz_config_win 3_3a_3 $txt
}

proc wiz_3_3b_1 {} {
	set txt "Searching for the VobBlanker program..."
	wiz_config_win 3_3b_1 $txt
}

proc wiz_3_3b_2 {{vtslist {1 1}}} {
	set txt "VobBlanker will now be launched to process the titleset"
	if {[llength $vtslist] > 1} {
		set txt "${txt}s"
	}
	set last 0
	foreach i $vtslist {
		set t [lindex $i 0]
		if {$t != $last} {
			set txt "${txt} [format %02d $t],"
			set last $t
		}
	}
	set txt [string trim $txt ,]
	set txt "${txt}.\n\n"
	set txt "${txt}When VobBlanker is opened and has loaded the DVD, you must click yourself the \"PROCESS!!\" button.\n"
	set txt "${txt}VobBlanker will start processing the titlesets.  This operation may take several minutes.  Please be patient.\n\n"
	set txt "${txt}When finished, have a look at the last line of the log.  (If there are warnings or errors, it's a good idea to save it.)\n"
	set txt "${txt}Then, click the \"QUIT\" button.  The ripped DVD will be reloaded automatically in PgcEdit, and resaved.\n\n"
	set txt "${txt}Press the Next button below to launch VobBlanker."
	wiz_config_win 3_3b_2 $txt
}

proc wiz_3_3b_3 {} {
	set txt "VobBlanker should now be processing the VOBs."
	wiz_config_win 3_3b_3 $txt disabled disabled
}

proc wiz_3_4 {} {
	set txt "Your DVD should now be completely clean.\n\n\n"
	set txt "${txt}Since DVD Shrink 3.2 has been detected on your computer, it's a good idea to launch it now to verify the DVD.\n"
	set txt "${txt}If DVD Shrink fails, you should try to process the whole DVD with VobBlanker."
	wiz_config_win 3_4 $txt

	set mf .psl2wizard.f1.mf

	checkbutton $mf.cb3 -text "Launch DVD Shrink on the current DVD" \
		-variable ::plugins::psl2::wiz::dvdshrink
	pack  $mf.cb3 -anchor w
	set ::plugins::psl2::wiz::dvdshrink 1
}

# Phase Finished

proc wiz_Finished {{del_log true} {del_psl true}} {
	set txt "Done!\nYou have now completed the wizard.\n\nPlease select the additional options you want:"
	wiz_config_win Finished $txt
	.psl2wizard.bf.b2 configure -text "Finish"

	set mf .psl2wizard.f1.mf

	checkbutton $mf.cb1 -text "Delete all backup files.  (If you don't select this option, you will need to delete the backup folders yourself.)" \
		-variable ::plugins::psl2::wiz::del_backups
	pack  $mf.cb1 -anchor w
	checkbutton $mf.cb2 -text "Delete the DVD Decrypter LOG file." \
		-variable ::plugins::psl2::wiz::del_files
	pack  $mf.cb2 -anchor w
	checkbutton $mf.cb3 -text "Delete the new PSL file created by the plugin.  (Do it only if you have verified that your DVD works!)" \
		-variable ::plugins::psl2::wiz::del_outputpsl
	pack  $mf.cb3 -anchor w
	checkbutton $mf.cb4 -text "Open DVD folder.  (To view files and folders and/or launch a software player.)" \
		-variable ::plugins::psl2::wiz::open_folder
	pack  $mf.cb4 -anchor w
	checkbutton $mf.cb5 -text "Close PgcEdit when done.  (If you don't want to use it to examine or modify the DVD.)" \
		-variable ::plugins::psl2::wiz::close_pgcedit
	pack  $mf.cb5 -anchor w
	set ::plugins::psl2::wiz::del_backups   0
	set ::plugins::psl2::wiz::del_files     0
	set ::plugins::psl2::wiz::del_outputpsl 0
	set ::plugins::psl2::wiz::open_folder   1
	set ::plugins::psl2::wiz::close_pgcedit 1
	if {! $del_log} {
		$mf.cb2 configure -state disabled
	}
	if {! $del_psl} {
		$mf.cb3 configure -state disabled
	}
}

proc wiz_Error {{msg ""}} {
	console hide
	wm deiconify .psl2wizard
	.psl2wizard.bf.b1 configure -text "Close"
	set txt "Sorry.  An error has occured"
	if {$msg != ""} {
		set txt "${txt}:\n\n$msg\n\n\n"
		bell
	} else {
		set txt "${txt}.\n\n\n"
	}
	set txt "${txt}The wizard will now be closed."
	wiz_config_win "" $txt normal disabled
	wm title .psl2wizard "PSL2 plugin - Wizard - Error!"
}

# Display an information in a MessageBox in background using Windows Scripting Host
# (Note: not displayed if expert mode is on, or if wscript.exe is not available!)
# Return the pid of the process, or 0.
proc open_background_msgbox {title msg delay} {
	if {$::config(expert_mode)} {
		return 0
	}
	if {! [file exists [file join $::config(bindir) "PSKill.exe"]]} {
		return 0
	}
	set txt ""
	if {$delay > 0} {
		set txt "WScript.Sleep $delay\n"
	}
	set msg [string trim [string map {\n {"&chr(13)&"} \" {"&chr(34)&"}} [string trim $msg \n]] &]
	set txt "${txt}msg = \"$msg\"\n"
	set txt "${txt}result = MsgBox(msg, 64, \"$title\")\n"
	set filename [file join $::config(tempdir) "background_msgbox.vbs"]
	if {[catch {saveFile $filename auto $txt}]} {
		return 0
	}
	set pid 0
	catch {set pid [exec -- wscript $filename &]}
	return $pid
}

# Kill a process
proc close_background_msgbox {pid} {
	if {$pid > 0} {
		catch {exec -- [file join $::config(bindir) "PSKill.exe"] $pid}
	}
	return 0
}

################### Small Procs called from the PSL menu ###################
# Procedure called from the "Clean cell commands" menu.
proc clean_cell_commands {} {
	if {[catch {if {![test_version 0.6.1 "PSL2 Plugin"]} return}]} return

	if {![info exists ::dvddir] || $::scratchbook_active} {
		bell
		return
	}
	show_msg ""

	# retrieve cell commands of the current PGC
	set type [lindex $::editor_cmds_from 0]
	set indexes [split [lindex $::editor_cmds_from 1] ,]
	set vts [lindex $indexes 0]
	if {$type == "pgc"} {
		if {$vts == 0} {bell; show_msg "Dummy PGC!"; return}
		set pgc [lindex $indexes 1]
		set cellpbtable $::pgcs($vts,$pgc,cellpbtable)
		set numcells  $::pgcs($vts,$pgc,numcells)
		set cmds $::pgcs($vts,$pgc)
	} else {
		set lu  [lindex $indexes 1]
		set pgc [lindex $indexes 2]
		set cellpbtable $::menus($vts,$lu,$pgc,cellpbtable)
		set numcells  $::menus($vts,$lu,$pgc,numcells)
		set cmds $::menus($vts,$lu,$pgc)
	}
	set cellcmds [get_type_cmds cell cmds]

	# Clean cell commands
	set cleaned [clean_cell_commands_sub $type $vts $pgc \
			[string toupper [get_type_cmds pre cmds]] \
			[string toupper $cellcmds] $numcells $cellpbtable]
	set clean_cmds [lindex $cleaned 0]

	# Replace the cleaned cell commands in the current PGC
	if {$type == "pgc"} {
		set_type_cmds cell ::pgcs($vts,$pgc) $clean_cmds
		set ::pgcs($vts,$pgc,cellpbtable) [lindex $cleaned 1]
	} else {
		set_type_cmds cell ::menus($vts,$lu,$pgc) $clean_cmds
		set ::menus($vts,$lu,$pgc,cellpbtable) [lindex $cleaned 1]
	}

	# Refresh GUI
	refresh_current_pgc
	::main::dvd_modified true false

	# Done.
	if {!$::config(expert_mode)} {
		tk_messageBox -title "Clean cell commands" -icon info -message \
			"Done.\nRemoved [expr [llength $cellcmds]-[llength $clean_cmds]] unused or fake cell commands\nand all fake conditions."
	}
}

# Launch DVD Decrypter (Get filename from registry)
proc launch_dvdd {mode {background true} {destination ""}} {
	set rc 0
	if {! [catch {package require registry}]} {
		set keyname {HKEY_CURRENT_USER\Software\DVD Decrypter}
		if {! [catch {set exepath [registry get $keyname InstallDirectory]}] && \
					$exepath != "" && [file isdirectory $exepath]} {

			set exefile [file join $exepath "DVDDecrypter.exe"]

			if {[file size $exefile] >= 772608} {
				if {$destination == ""} {
					if {$background} {
						set rc [catch {exec -- $exefile /MODE $mode &}]
					} else {
						set rc [catch {exec -- $exefile /MODE $mode}]
					}
				} else {
					if {[file isdirectory [file join $destination "VIDEO_TS"]]} {
						set destination [file nativename [file join $destination "VIDEO_TS"]]
					}
					if {$background} {
						set rc [catch {exec -- $exefile /MODE $mode /DEST $destination &}]
					} else {
						set rc [catch {exec -- $exefile /MODE $mode /DEST $destination}]
					}
				}
			} else {
				set txt "Wrong version of DVD Decrypter!\nHis file size should be 772,608 bytes.\nPlease install v3.5.4.0 and retry.\n(See the psl2_install.txt file for a bunch\nof download links.)"
				tk_messageBox -title "PSL2 plugin" -icon error -message $txt
				set rc 1
			}
		} else {set rc 1}
	} else {set rc 1}
	if {$rc} bell
	return $rc
}

# Set the DVD Decrypter rip settings so that the rip will work fine
proc reset_dvdd_settings {} {
	if {[catch {package require registry}]} {
		bell
		return
	}

	set txt "This function resets the DVD Decrypter setups related to\n"
	set txt "${txt}the ripping process to reasonable defaults.\n"
	set txt "${txt}Use this function if you have trouble ripping a DVD.\n\n"
	set txt "${txt}Note that this function disables also the automatic check\n"
	set txt "${txt}for program update, because the DVD Decrypter site is\n"
	set txt "${txt}now the property of Macrovision, and may be monitored!\n\n"
	set txt "${txt}Important:  If DVD Decrypter is running, close it now!\n"
	set txt "${txt}Continue?"
	set r [tk_messageBox -title "PSL2 plugin" -icon question -message $txt \
					-type yesno -default yes]
	if {$r == "no"} return

	set keyname {HKEY_CURRENT_USER\Software\DVD Decrypter}
	if {[catch {registry get $keyname VersionMajor}]} {
		bell
		return
	}

	registry set $keyname GENERAL_CheckForStructureProtection 1 dword
	registry set $keyname GENERAL_BadSectorRemovalMethod 0 dword
	registry set $keyname GENERAL_HideExpectedPackHeaderWarningsInProtectedAreas 0 dword

	registry set $keyname FILE_RemoveStructureProtection 1 dword
	registry set $keyname FILE_FileSplitting 3 dword
	registry set $keyname FILE_IgnoreUnreferencedVTSSets 1 dword

	registry set $keyname IFO_RemoveStructureProtection 1 dword
	registry set $keyname IFO_UncheckSuspectCells 1 dword

	registry set $keyname ISOREAD_RemoveStructureProtection 1 dword

	registry set $keyname IO_SoftwareReadErrorRetries \
			$::plugins::psl2::DVDD_IO_SoftwareReadErrorRetries dword
	registry set $keyname IO_IgnoreReadErrors 0 dword

	registry set $keyname DEVICE_DisableMCN 1 dword
	registry set $keyname DEVICE_SetReadSpeed 1 dword
	registry set $keyname DEVICE_ReadSpeed 0 dword

	registry set $keyname EVENTS_DeleteIncompleteFiles 0 dword
	registry set $keyname EVENTS_SaveLogOnExit 1 dword
	registry set $keyname EVENTS_SaveLogOnExitOverwriteExisting 1 dword
	registry set $keyname EVENTS_CheckForProgramUpdate 0 dword

	# verify if the last version is used
	set v1 [registry get $keyname VersionMajor]
	set v2 [registry get $keyname VersionMinor]
	set v3 [registry get $keyname VersionRevision]
	set v4 [registry get $keyname VersionBuild]
	if {$v1 < 3 || $v2 < 5 || $v3 < 4} {
		set txt "WARNING: The DVD Decrypter version\n"
		set txt "${txt}stored in the registry is v$v1.$v2.$v3.$v4.\n"
		set txt "${txt}The latest and last version is v3.5.4.0.\n\n"
		set txt "${txt}Please update DVD Decrypter!"
		tk_messageBox -title "PSL2 plugin" -icon warning -message $txt
	}
}

# Set and backup or restore the DVD Decrypter registry keys
proc set_dvdd_regkeys {mode {number_of_retries -1}} {
	if {$number_of_retries == -1} {
		set number_of_retries $::plugins::psl2::DVDD_IO_SoftwareReadErrorRetries
	}
	set dvdd_keyname {HKEY_CURRENT_USER\Software\DVD Decrypter}
	set dvdd_regkeys {
		GENERAL_CheckForStructureProtection
		GENERAL_BadSectorRemovalMethod
		FILE_RemoveStructureProtection
		FILE_FileSplitting
		FILE_IgnoreUnreferencedVTSSets
		IO_SoftwareReadErrorRetries
		IO_IgnoreReadErrors
		DEVICE_DisableMCN
		DEVICE_SetReadSpeed
		DEVICE_ReadSpeed
		EVENTS_DeleteIncompleteFiles
		EVENTS_SaveLogOnExit
		EVENTS_SaveLogOnExitOverwriteExisting
	}
	if {$mode == "set"} {
		if {![namespace exists ::plugins::psl2::dvddkeys]} {
			namespace eval ::plugins::psl2::dvddkeys {}
			foreach key $dvdd_regkeys {
				catch {set ::plugins::psl2::dvddkeys::$key [registry get $dvdd_keyname $key]}
			}
			catch {set ::plugins::psl2::dvddkeys::EVENTS_SaveLogOnExitFolder \
					[registry get $dvdd_keyname EVENTS_SaveLogOnExitFolder]}
		}
		# set the DVDD settings to rip the DVD successfully
		registry set $dvdd_keyname GENERAL_CheckForStructureProtection 1 dword
		registry set $dvdd_keyname GENERAL_BadSectorRemovalMethod 0 dword
		registry set $dvdd_keyname FILE_RemoveStructureProtection 1 dword
		registry set $dvdd_keyname FILE_FileSplitting 3 dword
		registry set $dvdd_keyname FILE_IgnoreUnreferencedVTSSets 1 dword
		registry set $dvdd_keyname IO_SoftwareReadErrorRetries $number_of_retries dword
		registry set $dvdd_keyname IO_IgnoreReadErrors 0 dword
		registry set $dvdd_keyname DEVICE_DisableMCN 1 dword
		registry set $dvdd_keyname DEVICE_SetReadSpeed 1 dword
		registry set $dvdd_keyname DEVICE_ReadSpeed 0 dword
		registry set $dvdd_keyname EVENTS_DeleteIncompleteFiles 1 dword
		registry set $dvdd_keyname EVENTS_SaveLogOnExit 1 dword
		registry set $dvdd_keyname EVENTS_SaveLogOnExitOverwriteExisting 1 dword
		registry set $dvdd_keyname EVENTS_SaveLogOnExitFolder \
				"[file nativename $::config(tempdir)]\\"
		registry set $dvdd_keyname EVENTS_CheckForProgramUpdate 0 dword	;# for safety ;-)
	}
	if {$mode == "restore"} {
		if {[namespace exists ::plugins::psl2::dvddkeys]} {
			foreach key $dvdd_regkeys {
				catch {registry set $dvdd_keyname $key \
						[subst \$::plugins::psl2::dvddkeys::$key] dword}
			}
			catch {registry set $dvdd_keyname EVENTS_SaveLogOnExitFolder \
					$::plugins::psl2::dvddkeys::EVENTS_SaveLogOnExitFolder}
			namespace delete ::plugins::psl2::dvddkeys
		}
	}
	return
}

# Launch VobBlanker to process ALL titlesets cleaned by the remove crap function
# during the PgcEdit session.
proc clean_processed_vobs {} {
	if {[catch {if {![test_version 0.6.1 "PSL2 Plugin"]} return}]} return
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	if {! [info exists ::plugins::psl2::cleaned_titles_list] || \
		$::plugins::psl2::cleaned_titles_list == {}} {
		set txt "Nothing to process with VobBlanker!\n\n"
		set txt "${txt}You should use this function after applying the\n"
		set txt "${txt}Remove Crap pass on at least one title.\n\n"
		set txt "${txt}Unless the option to launch VobBlanker from the\n"
		set txt "${txt}Remove Crap GUI, this function removes the\n"
		set txt "${txt}unused cells from the VOBs of ALL titlesets you\n"
		set txt "${txt}have processed with the Remove Crap function\n"
		set txt "${txt}during the same PgcEdit session.\n"
		set txt "${txt}It is intended to be used with episodic DVDs with\n"
		set txt "${txt}ARccOS protections in several different titlesets.\n\n"
		set txt "${txt}(Don't quit PgcEdit and don't load another DVD in\n"
		set txt "${txt}PgcEdit between the Remove Crap passes, or the\n"
		set txt "${txt}list of processed titles will be lost!)\n"
		tk_messageBox -title "PSL2 plugin" -message $txt -icon error
		return false
	}
	return [vobblanker_process $::plugins::psl2::cleaned_titles_list]
}

# Launch FixVTS to process ALL titlesets cleaned by the remove crap function
# during the PgcEdit session.
proc fix_processed_vobs {} {
	if {[catch {if {![test_version 0.6.1 "PSL2 Plugin"]} return}]} return
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	if {! [info exists ::plugins::psl2::cleaned_titles_list] || \
		$::plugins::psl2::cleaned_titles_list == {}} {
		set txt "Nothing to process with FixVTS!\n\n"
		set txt "${txt}You should use this function after applying the\n"
		set txt "${txt}Remove Crap pass on at least one title.\n\n"
		set txt "${txt}This function removes the unused cells from the VOBs\n"
		set txt "${txt}of ALL titlesets you have processed with the Remove\n"
		set txt "${txt}Crap function during the same PgcEdit session.\n\n"
		set txt "${txt}(Don't quit PgcEdit and don't load another DVD in\n"
		set txt "${txt}PgcEdit between the Remove Crap passes, or the\n"
		set txt "${txt}list of processed titles will be lost!)\n"
		tk_messageBox -title "PSL2 plugin" -message $txt -icon error
		return false
	}
	return [fixvts_process $::plugins::psl2::cleaned_titles_list]
}

# launch VobBlanker on the current PGC (called from the menu)
proc vobblanker_current_domain {} {
	if {! [info exists ::dvddir]} {
		bell
		return false
	}
	set type [lindex $::editor_cmds_from 0]
	set indexes [split [lindex $::editor_cmds_from 1] ,]
	set vts  [lindex $indexes 0]
	set pgc  [lindex $indexes 1]
	if {$vts == 0 || $type == "menu"} {
		bell
		return false
	}
	return [vobblanker_process [list [list $vts $pgc 1]]]
}


# Launch VobBlanker, ready with the Title to process
proc vobblanker_process {cleaned_titles_list {CheckFixPointers 0} {vobblanker_exe ""}} {
	# Retrieve VobBlanker.exe filename
	if {$vobblanker_exe == ""} {
		set vobblanker_exe [find_exe VobBlanker]
		if {$vobblanker_exe == ""} {
			bell
			return false
		}
	}
	# Save DVD
	if {$::dvd_modified} {
		set old_expertmode $::config(expert_mode)
		set ::config(expert_mode) true
		set rc [::main::Write_DVD false]
		set ::config(expert_mode) $old_expertmode
		if {$rc > 0} {
			return
		}
	}
	# CheckFixPointers may be needed when a V/CID has been changed in the VOBs
	foreach cleaned $cleaned_titles_list {
		if {[lindex $cleaned 2]} {
			set CheckFixPointers 1
			break
		}
	}

	# Generate INI file for VobBlanker

	set out "\[General\]\n"
	set out "${out}InputPath=[file nativename $::dvddir]\n"
	set out "${out}OutputPath=[file nativename $::dvddir]\n"
	set out "${out}CheckMenu=0\n"
	set out "${out}CheckTitle=1\n"
	set out "${out}CheckIfoPuops=0\n"
	set out "${out}CheckVobPuops=0\n"
	# may be needed when a V/CID has been changed in the VOBs
	set out "${out}CheckFolder=1\n"

	# Options moved in subsections since VB 2.1,
	# but kept here also for compatibility with older versions
	#
	# In PGC options section:
	set out "${out}CheckPostPre=0\n"
	set out "${out}CheckPostPreForced=0\n"
	set out "${out}CheckNextMenu=0\n"
	set out "${out}CheckNextTitle=0\n"
	set out "${out}CheckPrevMenu=0\n"
	set out "${out}CheckPrevTitle=0\n"
	set out "${out}CheckAttrTitle=0\n"
	# In FixVOBPointers section:
	set out "${out}CheckFixPointers=$CheckFixPointers\n"

	# New section in VB 2.1
	set out "${out}\[PGC options\]\n"
	set out "${out}CheckPostPre=0\n"
	set out "${out}CheckPostPreForced=0\n"
	set out "${out}CheckNextMenu=0\n"
	set out "${out}CheckNextTitle=0\n"
	set out "${out}CheckPrevMenu=0\n"
	set out "${out}CheckPrevTitle=0\n"
	set out "${out}CheckAttrTitle=0\n"

	# New section in VB 2.1
	set out "${out}\[FixVOBPointers\]\n"
	set out "${out}CheckFixTitlePointers=$CheckFixPointers\n"
	set out "${out}CheckFixMenuPointers=$CheckFixPointers\n"

	set out "${out}\[ScanVOBUs\]\n"
	set out "${out}CheckAutoScan=0\n"

	set out "${out}\[VobLayout\]\n"
	set out "${out}CheckVobFill=0\n"
	set out "${out}CheckVobOrder=0\n"
	if {[min_version [find_vobblanker_version] 2 1]} {
		set out "${out}CheckVobMUnref=0\n"
		set out "${out}CheckVobLUnref=0\n"
		set out "${out}CheckVobRenumber=1\n"
	} else {
		set out "${out}CheckVobMUnref=1\n"
		set out "${out}CheckVobLUnref=0\n"
	}
	set out "${out}CheckClockDisc=1\n"

	set out "${out}\[VIDEO_TS\]\n"
	set out "${out}ActionVTS=SKIP\n"

	for {set i 1} {$i <= $::pgcs(numvts)} {incr i} {
		set out "${out}\[VTS[format %02d $i]\]\n"
		# searches in the list of cleaned titles if we need to process this VTS
		set vts_cleaned false
		foreach cleaned $cleaned_titles_list {
			if {$i == [lindex $cleaned 0]} {
				set vts_cleaned true
				break
			}
		}
		if {$vts_cleaned} {
			set out "${out}ActionVTS=PROCESS\n"
		} else {
			set out "${out}ActionVTS=SKIP\n"
		}
	}
	#set out "${out}\[Backup\]\n"
	#set out "${out}CheckBckVts=1\n"

	# Write INI file
	set inifilename [file join $::config(tempdir) "VobBlanker_PSL2.INI"]
	if {[saveFile $inifilename auto $out] != 0} {
		return false
	}
	# launch VobBlanker
	set oldvobsize [vob_files_size]
	wm iconify .
	set rc [catch {exec -- $vobblanker_exe $inifilename} errstr]
	wm deiconify .
	if {!$rc && $oldvobsize == [vob_files_size]} {
		set rc 1
		set errstr "There is something wrong: no change in VOB size!"
	}
	if {$rc} {
		tk_messageBox -title "VobBlanker error" -message $errstr -icon error
		return false
	} else {
		# Reload DVD in PgcEdit
		open_dvd_wo_messages false
		# fix PGC total time (VobBlanker has a bug here!) in all cleaned PGCs
		foreach cleaned $cleaned_titles_list {
			set vts [lindex $cleaned 0]
			set pgc [lindex $cleaned 1]
			if {[info exists ::pgcs($vts,$pgc,numcells)] && $::pgcs($vts,$pgc,numcells) > 0} {
				fix_pgc_time $vts $pgc $::pgcs($vts,$pgc,numcells) $::pgcs($vts,$pgc,cellpbtable)
			}
		}
		# and save DVD again
		set old_expertmode $::config(expert_mode)
		set ::config(expert_mode) true
		::main::Write_DVD false
		set ::config(expert_mode) $old_expertmode
		::main::Incremental_Backup
		return true
	}
}

# launch FixVTS on the current title domain
proc fixvts_current_domain {} {
	if {! [info exists ::dvddir]} {
		bell
		return false
	}
	set type [lindex $::editor_cmds_from 0]
	set vts  [lindex [split [lindex $::editor_cmds_from 1] ,] 0]
	if {$vts == 0 || $type == "menu"} {
		bell
		return false
	}
	return [fixvts_process [list [list $vts]]]
}

# Launch fixvts, ready with the Title to process
proc fixvts_process {cleaned_titles_list} {
	# Retrieve fixvts.exe filename
	set fixvts_exe [find_exe FixVTS]
	if {$fixvts_exe == ""} {
		bell
		return false
	}
	# Save DVD
	if {$::dvd_modified} {
		if {[::main::Write_DVD false] > 0} {
			return
		}
	}
	wm iconify .
	for {set i 1} {$i <= $::pgcs(numvts)} {incr i} {
		# searches in the list of cleaned titles if we need to process this VTS
		set vts_cleaned false
		foreach cleaned $cleaned_titles_list {
			if {$i == [lindex $cleaned 0]} {
				set vts_cleaned true
				break
			}
		}
		if {$vts_cleaned} {
			# launch fixvts
			set vobfilename [file nativename [file join $::dvddir [format "VTS_%02d_1.VOB" $i]]]
			set rc [catch {exec -- $fixvts_exe -u $vobfilename} errstr]
			if {$rc} {
				wm deiconify .
				tk_messageBox -title "FixVTS error" -message $errstr -icon error
				return false
			}
		}
	}
	wm deiconify .

	# Reload DVD in PgcEdit
	::main::ReRead_DVD false
	::main::Incremental_Backup
	return true
}

# Open the DVD in PgcEdit, but hides most of the dialogs and startup procedures.
proc open_dvd_wo_messages {open_file_dialog} {
	set pgcedit_options {
		open_trace_mode
		auto_create_backup
		auto_fix_streams
		expert_mode
		auto_launch_menubuttons_parser
	}

	# disable the various GUIs PgcEdit opens normally.
	foreach opt $pgcedit_options {
		set orig_$opt $::config($opt)
	}
	set pgcedit_options_saved true
	set ::config(open_trace_mode) false
	set ::config(auto_create_backup) true
	set ::config(auto_fix_streams) false
	set ::config(expert_mode) true
	set ::config(auto_launch_menubuttons_parser) false
	set ::do_not_save_pgc_comments true
	unset -nocomplain ::buttons::dvd_reopened

	if {$open_file_dialog} {
		set errcode [::main::Read_New_DVD]
	} else {
		set errcode [::main::ReRead_DVD false]
	}

	unset -nocomplain ::do_not_save_pgc_comments
	foreach opt $pgcedit_options {
		set ::config($opt) [subst \$orig_$opt]
	}
	return $errcode
}

# Get SCR (System Clock Reference) of the first Nav Pack of the current PGC, and display it.
proc display_starting_scr {} {
	if {[catch {if {![test_version 0.6.1 "PSL2 Plugin"]} return}]} return

	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]
	set vts [lindex [split $index ,] 0]
	if {[subst \$::${type}s($index,numcells)] == 0} {
		bell
		show_msg "This PGC is a dummy PGC and has no timing."
		return
	}
	set cellpbtable [subst \$::${type}s($index,cellpbtable)]
	set vobu_start [get_int cellpbtable 8 4]
	set pos [::menued::get_nav_pack_position $type $vts $vobu_start $::dvddir]
	set vobfile [lindex $pos 0]
	set lba [lindex $pos 1]
	set fh [::menued::openbinfile $vobfile r]
	if {$fh == ""} return
	set navpack [::menued::load_nav_pack $vobfile $fh $lba]
	close $fh
	if {$navpack == ""} return
	if {[expr [get_int navpack 0x48 1] & 0xC0] == [expr 0xC0]} {
		set fps 30
	} else {
		set fps 25
	}
	set scr [string range $navpack 4 9]
	unset navpack
	set txt       "The first Nav Pack of the current PGC\n"
	set txt "${txt}is at sector $lba of [file tail $vobfile].\n\n"
	set txt "${txt}His SCR (System Clock Reference) is:\n"
	set txt "${txt}[convert_SCR $scr $fps]\n\n"
	set txt "${txt}(You may need to substract this time\n"
	set txt "${txt}to synchronize the subpics extracted\n"
	set txt "${txt}by SubRip or similar programs.)"
	tk_messageBox -message $txt -title "PSL2 plugin"
}

# "Dump absolute cells LBAs" function
proc display_absolute_lbas {} {
	# safety check
	if {[catch {if {![test_version 0.6.1 "PSL2 Plugin"]} return}]} return

	if {![info exists ::dvddir] || $::scratchbook_active} {
		bell
		show_msg "Select the ARccOS protected Title!"
		return 0
	}

	show_msg ""

	# init variables
	set type [lindex $::editor_cmds_from 0]
	set indexes [split [lindex $::editor_cmds_from 1] ,]
	set vts [lindex $indexes 0]
	if {$type == "pgc"} {
		if {$vts == 0} {bell; show_msg "Dummy PGC!"; return 0}
		set pgc [lindex $indexes 1]
		set cellpbtable $::pgcs($vts,$pgc,cellpbtable)
		set numcells  $::pgcs($vts,$pgc,numcells)
		set vobfile "VTS_[format %02d $vts]_1.VOB"
		set pgclabel [::main::pgc_label "" pgc $vts $pgc]
	} else {
		set lu  [lindex $indexes 1]
		set pgc [lindex $indexes 2]
		set cellpbtable $::menus($vts,$lu,$pgc,cellpbtable)
		set numcells  $::menus($vts,$lu,$pgc,numcells)
		if {$vts == 0} {
			set vobfile "VIDEO_TS.VOB"
		} else {
			set vobfile "VTS_[format %02d $vts]_0.VOB"
		}
		set pgclabel [::main::pgc_label "" menu $vts $lu $pgc]
	}
	if {$numcells == 0} {
		bell
		show_msg "Dummy PGC!"
		return 0
	}

	# Get starting LBA from user (string requester of log file)
	set domainstart [input_starting_lba $vobfile]
	if {$domainstart == ""} {return 0}

	# generate output
	set out "# Cells absolute LBAs of $pgclabel\n"
	set out "${out}# Domain starting LBA (absolute start sector of $vobfile): $domainstart\n\n"
	for {set c 1} {$c <= $numcells} {incr c} {
		# get VOPBU start and end from the cell playback table, incremented by the start LBA of VOB.
		set cidx [expr (($c-1)*24)+8]
		set vobustart [expr [get_int cellpbtable $cidx 4] + $domainstart]
		incr cidx 12
		set vobuend   [expr [get_int cellpbtable $cidx 4] + $domainstart]
		set out "${out}Cell [format %3d $c]: [format %7d $vobustart] - [format %7d $vobuend]\n"
	}
	# Display output
	::main::showtext .psl2dump "Absolute LBA addresses" $out
	return 1
}

proc display_crc32 {} {
	set drive [lindex [find_dvd_video_drives] 0]
	set filename [tk_getOpenFile -title "Open the *ORIGINAL* VIDEO_TS.IFO file" \
				-initialdir [file join $drive VIDEO_TS] -initialfile "VIDEO_TS.IFO" \
				-filetypes {{"VIDEO_TS.IFO files" {VIDEO_TS.IFO}}}]
	if {$filename == ""} return
	set crc32 [crc32 $filename]
	if {$crc32 == ""} {
		set txt "Cannot retrieve CRC32 code of file\n\"$filename\"!"
		set icon error
	} else {
		set txt "The CRC32 code for file\n\"$filename\"\nis:  $crc32\n\n"
		set txt "${txt}Important:\n\nIf you have AnyDVD or DVD43,\nyou MUST disable it BEFORE\n"
		set txt "${txt}running the CRC32 tool, or the\nCRC32 code will be wrong!\n\n"
		set txt "${txt}Also, be sure to select the\nORIGINAL VIDEO_TS.IFO file,\n"
		set txt "${txt}from the original DVD!"
		set icon info
	}
	tk_messageBox -title "PSL2 Plugin" -icon $icon -message $txt
	return $crc32
}

########################### Windoze specific functions ################################

# return a list of the drives containing a DVD-Video
# The list is in the form D:\ E:\ ...
proc find_dvd_video_drives {} {
	set driveslist {}
	foreach drive [file volumes] {
		if {		[string range $drive 0 0] != "A"
				&&	[string range $drive 0 0] != "B"
				&& ![catch {file type $drive} blah]
				&& ![file writable $drive]
				&&	[lindex [file system $drive] 1] == "UDF"
				&&	[file exists [file join $drive "VIDEO_TS/VIDEO_TS.IFO"]]} {
			lappend driveslist [string map {/ \\} $drive]
		}
	}
	return $driveslist
}

# return the CRC32 (in hex) of the VIDEO_TS.IFO file
# Assumes CRC32.exe is in the path.
proc crc32 {filename} {
	set exename [::plugins::psl2::find_exe crc32]
	set fname [string map {\\ \\\\} [file nativename $filename]]
	if {![catch {set fh [open "|\"$exename\" \"$fname\" /R" r]}]} {
		fconfigure $fh -translation auto
		set buf [read $fh]
		close $fh
		return "0x[lindex $buf 3]"
	} else {
		return ""
	}
}

# Return the last saved PSL file (from the registry.)
proc find_last_pslfile {} {
	if {[info exists ::plugins::psl2::default_pslfile]} {
		return $::plugins::psl2::default_pslfile
	}
	if {! [catch {package require registry}]} {
		set keyname {HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU\PSL}
		if {! [catch {set mrulist [registry get $keyname MRUList]}] && $mrulist != ""} {
			if {! [catch {set filename [registry get $keyname [string range $mrulist 0 0]]}]} {
				return [string map {{_new} {}} $filename]
			}
		}
	}
	return "C:\\DVDVideo.PSL"
}

# Return the last saved LOG file (from the registry.)
proc find_last_logfile {} {
	if {[info exists ::plugins::psl2::default_logfile]} {
		return $::plugins::psl2::default_logfile
	}
	if {! [catch {package require registry}]} {
		set keyname {HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU\log}
		if {! [catch {set mrulist [registry get $keyname MRUList]}] && $mrulist != ""} {
			if {! [catch {set filename [registry get $keyname [string range $mrulist 0 0]]}]} {
				if {[file exists $filename]} {
					return $filename
				}
			}
		}
		set keyname {HKEY_CURRENT_USER\Software\DVD Decrypter}
		if {! [catch {set savelog [registry get $keyname EVENTS_SaveLogOnExit]}]} {
			if {$savelog && ! [catch {set overwrite [registry get $keyname EVENTS_SaveLogOnExitOverwriteExisting]}]} {
				if {$overwrite && ! [catch {set folder [registry get $keyname EVENTS_SaveLogOnExitFolder]}]} {
					set filename [file join $folder "DVDDecrypter.log"]
					if {[file exists $filename]} {
						return [file nativename $filename]
					}
				}
			}
		}
	}
	return "C:\\DVDDecrypter.log"
}

# Try to find an executable file in PgcEdit's install folder or in the registry.
proc find_exe {exe} {
	# search for exe in directory where PgcEdit is installed
	set filename [file join [file dirname $::argv0] $exe.exe]
	if {[file executable $filename]} {
		return $filename
	}
	set filename [file join [file join [file dirname $::argv0] plugins] "$exe.exe"]
	if {[file executable $filename]} {
		return $filename
	}
	set filename [file join [file join [file dirname $::argv0] bin] "$exe.exe"]
	if {[file executable $filename]} {
		return $filename
	}
	set filename [file join [file join [file dirname [file dirname $::argv0]] $exe] "$exe.exe"]
	if {[file executable $filename]} {
		return $filename
	}
	# If exe is not found, we need to use the registry
	if {! [catch {package require registry}]} {
		# search for filename.exe in <Program Files>/filename/
		if {[catch {set progfiles [registry get "HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion" "ProgramFilesDir"]}]} {
			set progfiles {C:\Program files}
		}
		set filename [file join [file join $progfiles $exe] $exe.exe]
		if {[file executable $filename]} {
			return $filename
		}
		# search for $exe.exe path in the registry
		set key {HKEY_CURRENT_USER\Software\Microsoft\Windows\ShellNoRoam\MUICache}
		foreach k [registry values $key "*$exe*"] {
			set value [registry get $key $k]
			if {[string toupper $value] == [string toupper $exe] && [file executable $k]} {
				return $k
			}
		}
	}
	# see if the psl2_plugin.cfg file exists
	set configfile [file join $::config(config_directory) "psl2_plugin.cfg"]
	if {! [catch {source $configfile}]} {
		if {[file executable ${exe}_exe_filename]} {
			return ${exe}_exe_filename
		}
	}
	# still not found: ask path interactively, and save it in config file
	set filename [tk_getOpenFile -title "First time setup: Please select $exe.exe" \
				-initialdir "C:\\" -initialfile "$exe.exe" \
				-filetypes {{"Program files" {*.exe}}}]
	if {$filename != "" && [file executable $filename]} {
		set txt [string trim [readFile $configfile auto] \n]
		set txt "$txt\nset ${exe}_exe_filename [list $filename]\n"
		saveFile $configfile auto $txt
		return $filename
	}
	return ""
}

# Try to find the DVDShrink path in the registry.
proc find_dvdshrink_exe {} {
	# we need to use the registry
	if {! [catch {package require registry}]} {
		# search for dvd shrink.exe path in the registry
		set key {HKEY_CURRENT_USER\Software\DVD Shrink\DVD Shrink 3.2\Preferences}
		if {! [catch {set path [registry get $key "InstallApp"]}]} {
			if {[file exists $path]} {
				return $path
			}
		}
	}
	return ""
}

# Retrieve VobBlanker's version numbers from the registry.
# Return {0 0 0 0} for pre-2.1 versions.
proc find_vobblanker_version {} {
	set version [list 0 0 0 0]
	# we need to use the registry
	if {! [catch {package require registry}]} {
		# search for dvd shrink.exe path in the registry
		set key {HKEY_CURRENT_USER\Software\jsoto's tools\VobBlanker\General}
		if {! [catch {set v [registry get $key "VobBlankerVersion"]}]} {
			set vl [split $v .]
			set version {}
			foreach i $vl {
				if {[string is integer $i]} {
					lappend version $i
				} else {
					set subver 0
					for {set len 0} {$len < [string length $i]} {incr len} {
						set s [string range $i 0 $len]
						if {[string is integer $s]} {
							set subver $s
						} else {
							break
						}
					}
					lappend version $subver
				}
			}
			while {[llength $version] < 4} {
				lappend $version 0
			}
		}
	}
	return $version
}

# return true if version in vlist is greather or equal to the version passed in arguments.
# Example usage: min_version [find_vobblanker_version] 2 1
# return true if VobBlanker 2.1.0.0 or above is installed.
proc min_version {vlist v0 {v1 0} {v2 0} {v3 0}} {
	while {[llength $vlist] < 4} {
		lappend vlist 0
	}
	if {$v0 > [lindex $vlist 0]} {return false}
	if {$v0 < [lindex $vlist 0]} {return true}
	if {$v1 > [lindex $vlist 1]} {return false}
	if {$v1 < [lindex $vlist 1]} {return true}
	if {$v2 > [lindex $vlist 2]} {return false}
	if {$v2 < [lindex $vlist 2]} {return true}
	if {$v3 > [lindex $vlist 3]} {return false}
	return true
}

##################### Support functions #####################

# reorder the lines of the PSL file so that the commands 0:, 1: and 6:
# are in increasing order (necessary for DVD Devrypter)
proc reorder_psl_file {out} {
	set out [string trim [string trim $out \n] "PSL2\n"]
	set old $out
	while 1 {
		set out [string map {\n\n \n} $out]
		if {$out == $old} break
		set old $out
	}
	set lst [lsort -dictionary [split $out \n]]
	set out "PSL2\n"
	set old_cmd "0:"
	set old_line ""
	set old_line6 ""
	foreach line $lst {
		if {$line == $old_line} continue
		set old_line $line
		set cmd [string range $line 0 1]
		if {$cmd != $old_cmd} {
			set out $out\n
			set old_cmd $cmd
		}
		if {$cmd == "6:"} {
			set lst6 [split $line ,]
			set line6 [lindex $lst6 0],[lindex $lst6 1],[lindex $lst6 2]
			if {$line6 != $old_line6} {
				set out $out\n
				set old_line6 $line6
			}
		}
		set out "$out$line\n"
	}
	return $out
}

# Fix PGC total time after remove cells and VobBlanker processing
proc fix_pgc_time {vts pgc numcells cellpbtable} {
	set end_time "\000\000\000\000"
	for {set c $numcells} {$c > 0} {incr c -1} {
		set cellpbidx [expr {($c - 1) * 24}]

		# add the cell time to the total PGC time, but only if the cell is angle 1.
		set celltype [get_int cellpbtable $cellpbidx 1]
		set cell_time [string range $cellpbtable [expr {$cellpbidx + 4}] [expr {$cellpbidx + 7}]]
		if {[expr {$celltype & 0x10}] == 0 || [expr {$celltype & 0xC0}] == 0x40} {
			set end_time [::pgced::AddBCDtime $cell_time $end_time]
		}
	}
	# Replace PGC's total time in PGC header
	set ::pgcs($vts,$pgc,header) "[string range $::pgcs($vts,$pgc,header) 0 3]$end_time[string range $::pgcs($vts,$pgc,header) 8 end]"
}

# Convert a SCR (System Clock Reference) to human readable format
# input: 6 bytes binary SCR value, frame per seconds value
proc convert_SCR {scr fps} {
	set ticks [expr {
		(([get_int scr 2 3] & 0x03FFF8) / 8) +
		((([get_int scr 0 3] & 0x03FFF8) / 8) * round(pow(2,15))) +
		((([get_int scr 0 1] & 0x38) / 8) * round(pow(2,30)))
	}]
	set remainder [expr ([get_int scr 4 2] & 0x03FE) / 2]
	set time_in_secs [expr ($ticks / 90000.0) + ($remainder / 27000000.0)]
	set sr [expr round(floor($time_in_secs))]
	set secs [expr $sr % 60]
	set mins [expr ($sr / 60) % 60]
	set hours [expr $sr / 3600]
	set frames [expr $time_in_secs * $fps]
	set txt "Time code: [format %02d:%02d:%02d.%02d $hours $mins $secs [expr round($frames) % $fps]] (at $fps fps)\n"
	set txt "${txt}Seconds:   [expr round($time_in_secs*1000)/1000.0]\n"
	set txt "${txt}Frames:    [expr round($frames)] (at $fps fps)"
	return $txt
}

# try to build the DVD name and description used to identify the PSL file
proc build_dvd_name {include_zones} {
	# Build the DVD name from the DVDText General Name, if any, or
	# get the DVD name from the directory name.
	set name [::utils::get_General_Name]
	if {$name == ""} {
		set tmp [file tail $::dvddir]
		if {$tmp != "VIDEO_TS" && [string first : $tmp]==-1} {
			set name $tmp
		} else {
			set tmp [file tail [file dirname $::dvddir]]
			if {[string first : $tmp]==-1} {
				set name $tmp
			}
		}
	}
	# remove "," in the DVD name: DVDD doesn't like them.
	set name [string map {, ""} $name]

	if {$include_zones} {
		# Concat "Zone # # #" after the filename
		set code [expr {[get_int ::raw_sectors(0,VMGM_MAT) 0X23 1] ^ 0XFF}]
		set regions ""
		for {set i 0} {$i < 6} {incr i} {
			if {[expr $code & round(pow(2,$i))] != 0} {
				set regions "$regions [expr $i+1]"
			}
		}
		set name "$name - Zone$regions"
	}
	return [string trim $name]
}

# Return total size of VOB files
proc vob_files_size {} {
	if {[catch {set vobfiles [glob -nocomplain -tails -directory $::dvddir -- *.VOB]}]} {
		return 0
	}
	set sectors 0
	foreach f $vobfiles {
		incr sectors [expr {[file size [file join $::dvddir $f]]/2048}]
	}
	return $sectors
}


##################### ARccOS support functions #####################

# search for protected PGCs in the DVD,
# If 'gui' arg is true, display the search results and select the first protected PGC.
# Warning: assumes some distinctive feature of the protected titles that might
# be changed in future protections!
proc find_protected_titles {from_cleanup {gui true}} {
	if {[catch {if {![test_version 0.6.1 "PSL2 Plugin"]} return}]} return
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	set pgcslist {}
	set numpgcs 0
	set curtype  [lindex $::editor_cmds_from 0]
	set curindex [split [lindex $::editor_cmds_from 1] ,]
	set curvts   [lindex $curindex 0]
	if {$curtype == "pgc"} {
		set curpgc [lindex $curindex 1]
	} else {
		set curpgc 0
	}
	set title_displayed 0

	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
			# number of cells > 1?
			if {$::pgcs($vts,$pgc,numcells) < 2} continue
			# number of cell commands > 0?
			if {[num_type_cmds cell ::pgcs($vts,$pgc)] == 0} continue
#			# first cell playback time < 1 sec?
#			if {[get_int ::pgcs($vts,$pgc,cellpbtable) 4 3] != 0} continue
			# first cell has the restricted flag set?
			if {[expr [get_int ::pgcs($vts,$pgc,cellpbtable) 1 1] & 0x20] == 0} continue

			# add current PGC to list of protected PGCs.
			incr numpgcs
			if {$numpgcs == 1} {
				set first_vts $vts
				set first_pgc $pgc
			}
			lappend pgcslist [list $vts $pgc false]
			if {$title_displayed == 0 && ($curvts < $vts || ($curvts == $vts && $curpgc < $pgc))} {
				set title_displayed $numpgcs
				::main::display_pgc pgc $vts $pgc
				set curpgclabel [::main::pgc_label {} pgc $vts $pgc]
			}
		}
	}
	# Find unreferenced titles at end of titlesets
	set numtitles [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]
	::utils::build_crossrefs -1
	for {set title $numtitles} {$title > 0} {incr title -1} {
		set idx [expr 8+(($title-1)*12)+6]
		set vts [get_int ::raw_sectors(0,VMG_TT_SRPT) $idx 1]
		incr idx
		set ttn [get_int ::raw_sectors(0,VMG_TT_SRPT) $idx 1]
		set pgc 0
		for {set p $::pgcs($vts,numpgcs)} {$p > 0} {incr p -1} {
			if {[info exists ::pgcs($vts,$p,TTN)] && $::pgcs($vts,$p,TTN)==$ttn} {
				set pgc $p
				break
			}
		}
		if {$pgc>0} {
			if {! [info exists ::crossrefs(pgcs,$vts,$pgc,numrefs)] || \
					$::crossrefs(pgcs,$vts,$pgc,numrefs) == 0} {
				set idx [lsearch [list $vts $pgc false] $pgcslist]
				if {$idx != -1} {
					set pgcslist [lreplace $pgclist $idx $idx]
				}
				lappend pgcslist [list $vts $pgc true]
				incr numpgcs
				set first_vts $vts
				set first_pgc $pgc
			} else {
				break
			}
		}
	}
	
	if {$gui} {
		if {$numpgcs > 0} {
			set pgcstext ""
			foreach i $pgcslist {
				if {[lindex $i 2]} {
					set txt "  (Unreferenced)"
				} else {
					set txt ""
				}
				set pgcstext "${pgcstext}[::main::pgc_label {} pgc [lindex $i 0] [lindex $i 1]]$txt\n"
			}
			if {$title_displayed == 0} {
				set title_displayed 1
				::main::display_pgc pgc $first_vts $first_pgc
				set curpgclabel [::main::pgc_label {} pgc $first_vts $first_pgc]
			}
			if {$numpgcs > 0} {set pluriel s} else {set pluriel ""}
			if {$from_cleanup} {
				set txt "Found $numpgcs other protected PGC$pluriel:\n$pgcstext\n"
				set txt "${txt}\nDo you want to clean also the PGC\n$curpgclabel ?\n(Recommended)"
				set r [tk_messageBox -title "PSL2 plugin" -icon question -message $txt \
						-type yesno -default yes]
				if {$r == "yes"} {
					update idletasks
					after 500 {::plugins::psl2::main_gui true}
				}
			} else {
				set txt "Found $numpgcs PGC$pluriel probably protected:\n$pgcstext\n"
				if {$title_displayed == 1} {
					set txt "${txt}(The first protected title has been selected.)"
				} else {
					set txt "${txt}(The next protected title has been selected.)"
				}
				tk_messageBox -title "PSL2 plugin" -icon info -message $txt
			}
		} else {
			if {$from_cleanup} {
				set txt "Cannot find any other protected title!\n\n"
				set txt "${txt}Cleanup done on the whole DVD."
			} else {
				set txt "Cannot find any protected title!\n\n"
				set txt "${txt}Since the protection method may have changed,\nyou should verify by yourself."
			}
			tk_messageBox -title "PSL2 plugin" -icon info -message $txt
		}
	}
	return $pgcslist
}

# Searches in the pre-commands if there are LinkCN commands jumping to
# a cell in the first program.  If any, return a list of the target cells.
proc find_possible_entry_cells {precmds numcells programmap} {
	# build a table of all cells with his program number
	for {set c 1} {$c <= $numcells} {incr c} {
		set cellprogram($c) 0
	}
	for {set i 0} {$i < [string length $programmap]} {incr i} {
		set cellprogram([get_int programmap $i 1]) [expr $i+1]
	}
	set oldprog 0
	for {set c 1} {$c <= $numcells} {incr c} {
		if {$cellprogram($c) == 0} {
			set cellprogram($c) $oldprog
		} else {
			set oldprog $cellprogram($c)
		}
	}

	# init return list to an empty list
	set possible_cells {}
	# for each LinkCN pre command
	foreach cc [lsearch -all -regexp $precmds "20 .7 .. .. .. .. .. .."] {
		# Get target cell number in the command
		set target [x2d [lindex [lindex $precmds $cc] 7]]
		# And add it in the list, if not already present, and if the cell is in program 1.
		if {$target != 1 && [lsearch $possible_cells $target]==-1 && $cellprogram($target)==1} {
			lappend possible_cells $target
		}
	}
	return [lsort $possible_cells]
}

# Find the unplayed cells by following the Link in the post commands.
# Uses the ::plugins::psl2::cb($cell) variable to store the result.
proc find_arcoss_cells {vts pgc numcells cellpbtable cmds_upvar programmap enter_cell_num} {
	upvar $cmds_upvar cmds

	# Clean cell commands and cell playback table a bit.
	set cleaned [clean_cell_commands_sub pgc $vts $pgc \
			[string toupper [get_type_cmds pre cmds]] \
			[string toupper [get_type_cmds cell cmds]] $numcells $cellpbtable]
	set ccmds [lindex $cleaned 0]
	set cellpbtable [lindex $cleaned 1]

	# begin by marking all cells as unplayed.
	for {set c 1} {$c <= $numcells} {incr c} {
		set ::plugins::psl2::cb($c) 1
		set cellprogram($c) 0
	}

	# build a table of all cells with his program number
	for {set i 0} {$i < [string length $programmap]} {incr i} {
		set c [get_int programmap $i 1]
		set cellprogram($c) [expr $i+1]
		set programentry([expr $i+1]) $c
	}
	set oldprog 0
	for {set c 1} {$c <= $numcells} {incr c} {
		if {$cellprogram($c) == 0} {
			set cellprogram($c) $oldprog
		} else {
			set oldprog $cellprogram($c)
		}
	}

	# enter the PGC at the cell number specified by the user (normally cell 1)
	set c $enter_cell_num
	# loop through cells, jumpimng to the cell pointed to by the cell command
	while {$c <= $numcells} {
		# If we encounter a cell that has already been cleaned, then we are in an
		# infinite loop.  Warning, and exit loop.
		if {$::plugins::psl2::cb($c) == 0} {
			tk_messageBox -title "Find ARccOS cells" -icon warning -message \
				"Something went wrong.\nInfinite loop encountered!\nThe Entry Cell value is probably wrong."
			break
		}
		# mark current cell as played (uncheck it in the GUI)
		set ::plugins::psl2::cb($c) 0
		# get cell command number
		set cellcmdnum [get_int cellpbtable [expr (($c-1)*24)+3] 1]
		# if no cell command, next cell is current cell + 1
		if {$cellcmdnum==0} {
			incr c
		} else {
			# retrieve cell command
			set cellcmd [lindex $ccmds [expr $cellcmdnum-1]]
			# analyze the hex code of the cell command (in the form
			# "00 00 00 00 00 00 00 00") to compute the target.
			switch [string range $cellcmd 0 0] {
				2 - 6 - 7 {
					# check Link and Set commands
					switch [string range $cellcmd 4 4] {
						1 {
							switch [x2d [lindex $cellcmd 7]] {
								2 {	# LinkNextCell -> next cell
									set target [expr $c+1]
								}
								3 {	# LinkPrevCell -> previous cell
									set target [expr $c-1]
								}
								6 {	# LinkNextPG -> next program
									set target $programentry([expr $cellprogram($c)+1])
								}
								7 {	# LinkPrevPG -> prev program
									set target $programentry([expr $cellprogram($c)-1])
								}
								9 - 10 - 11 - 12 - 13 { # LinkTailPGC,... -> End of playback
									set target [expr $numcells+1]
								}
								default {
									set target [expr $c+1]
								}
							}
						}
						5 - 6 {	set target $programentry([x2d [lindex $cellcmd 7]]) ;# LinkPTT and LinkPGN -> program's entry cell}
						7 {
							# LinkCN -> Retrieve target cell from command
							set target [x2d [lindex $cellcmd 7]]
							if {$target > $numcells} {
								if {[string range $cellcmd 3 3] == 0} {
									tk_messageBox -title "Find ARccOS cells" -icon warning -message \
										"Something went wrong.\nLink to inexistant cell $target encountered!"
									break
								} else {
									set target [expr $c+1]
								}
							}
						}
						default {
							# Assumes next cell for all other Links or Set commands
							set target [expr $c+1]
						}
					}
					set c $target
				}
				3 {
					# Check Jump commands
					if {[string range $cellcmd 4 4]==5} {
						# JumpVTS_PTT
						set c $programentry([x2d [lindex $cellcmd 3]])
					} elseif {[string range $cellcmd 4 4]==2} {
						# JumpTT
						set c 1
					} else {
						# Another jump: Assumes next cell
						incr c
					}
				}
				default {
					# Another command: Assumes next cell
					incr c
				}
			}
		}
	}
	# Remove the last cell if the previous cell has a LinkTailPGC cell command
	# (The last cell is used to skip to the end of the title with the remote.)
	if {$::plugins::psl2::cb($numcells)} {
		# retrieve cell command of previous cell
		set cellcmdnum 0
		catch {set cellcmdnum [get_int cellpbtable [expr (($numcells-2)*24)+3] 1]}
		if {$cellcmdnum > 0} {
			set cellcmd [string toupper [lindex $ccmds [expr $cellcmdnum-1]]]
			# if cell command is a LinkTailPGC, uncheck the last cell
			if {[lsearch -regexp [list $cellcmd] "20 .1 .. .. .. .. .. 0D"] != -1} {
				set ::plugins::psl2::cb($numcells) 0
			}
		}
	}
	# return list of unplayed cells
	set unplayed {}
	for {set c 1} {$c <= $numcells} {incr c} {
		if {! $::plugins::psl2::cb($numcells)} {
			lappend unplayed $c
		}
	}
	return $unplayed
}

# Clean cell commands
# Args: 1) cell commands as a Tcl List of hex codes
# 2) Number of cells.  3) Cell playback table, in binary.
proc clean_cell_commands_sub {type vts pgc precmds orig_cellcmds numcells cellpbtable} {


	# replace the SPRM in the conditions by his current value
	if {$type == "pgc"} {
		set linkcns [lsearch -regexp -all $orig_cellcmds "20 \[1-7\]. .. .. .. 8. .. .."]
		foreach c $linkcns {
			set cmd [lindex $orig_cellcmds $c]
			set sprm [x2d [string range $cmd 16 16]]
			set doit true
			switch $sprm {
				4 {
					if {[info exists ::pgcs($vts,$pgc,Title)]} {
						set val $::pgcs($vts,$pgc,Title)
					} else {
						set doit false
					}
				}
				5 {set val $::pgcs($vts,$pgc,TTN)}
				6 {set val $pgc}
				default {set doit false}
			}
			if {$doit} {
				set cond [expr [x2d [string range $cmd 3 3]] | 8]
				lset cmd 1 [format %X $cond][string range $cmd 4 4]
				lset cmd 5 [format %02X $val]
				lset orig_cellcmds $c $cmd
			}
		}
	}

	# Remove fake == conditons
	# Example: Converts "if { gprm(0) == gprm(0) } then LinkCN 3" to "LinkCN 3"

	# For all Link commands with an ==, <=, >= or & condition
	set linkcns [lsearch -regexp -all $orig_cellcmds "20 \[1246\]. .. .. .. .. .. .."]
	foreach c $linkcns {
		set cmd [lindex $orig_cellcmds $c]
		# if both operands of the condition are equal, then remove the condition
		if {[lindex $cmd 3]==[lindex $cmd 5]} {
			lset cmd 1 "0[string range $cmd 4 4]"
			lset cmd 3 "00"
			lset cmd 5 "00"
			lset orig_cellcmds $c $cmd
		}
	}

	# For all Link commands with an >= 0 condition, remove the condition
	set linkcns [lsearch -regexp -all $orig_cellcmds "20 C. .. .. 00 00 .. .."]
	foreach c $linkcns {
		set cmd [lindex $orig_cellcmds $c]
		lset cmd 1 "0[string range $cmd 4 4]"
		lset orig_cellcmds $c $cmd
	}

	# For all Link commands with a <= 0xFFFF condition, remove the condition
	set linkcns [lsearch -regexp -all $orig_cellcmds "20 E. .. .. FF FF .. .."]
	foreach c $linkcns {
		set cmd [lindex $orig_cellcmds $c]
		lset cmd 1 "0[string range $cmd 4 4]"
		lset cmd 5 "00"
		lset cmd 6 "00"
		lset orig_cellcmds $c $cmd
	}

	# Remove fake != conditons
	# Example: Converts "if { gprm(0) != gprm(0) } then LinkCN 3" to "NOP"

	# For all Link commands with an !=, < or > condition
	set linkcns [lsearch -regexp -all $orig_cellcmds "20 \[357\]. .. .. .. .. .. .."]
	foreach c $linkcns {
		set cmd [lindex $orig_cellcmds $c]
		# if both operands of the condition are equal, then replace the command by NOP
		if {[lindex $cmd 3]==[lindex $cmd 5]} {
			lset orig_cellcmds $c "00 00 00 00 00 00 00 00"
		}
	}

	# For all Link commands with an & 0 condition, remove the command
	set linkcns [lsearch -regexp -all $orig_cellcmds "20 9. .. .. 00 00 .. .."]
	foreach c $linkcns {
		lset orig_cellcmds $c "00 00 00 00 00 00 00 00"
	}

	# Remove fake Set commands
	# Example: Converts "Set gprm(0) = gprm(0) ; LinkCN 3" to "LinkCN 3"

	# For all Set commands with links
	set linkcns [lsearch -regexp -all $orig_cellcmds "\[67\]1 0\[1-7\] .. .. .. .. .. .."]
	foreach c $linkcns {
		set cmd [lindex $orig_cellcmds $c]
		# if the Set command is fake, convert it to plain Link command
		if {[lindex $cmd 3]==[lindex $cmd 5]} {
			lset cmd 0 "20"
			lset cmd 3 "00"
			lset cmd 5 "00"
			lset orig_cellcmds $c $cmd
		}
	}

	# Build new list of cell commands, sorted in cell order (first cell first),
	# and stripped of unused cell commands and NOPs.
	set cellcmds {}
	set newcmdnum 0
	set modified false
	set orig_numcmds [llength $orig_cellcmds]
	# for each cell
	for {set c 1} {$c <= $numcells} {incr c} {
		# retrieve cell command number from the cell playback table
		set cellcmdnum [get_int cellpbtable [expr (($c-1)*24)+3] 1]
		# If one command is defined
		if {$cellcmdnum>0} {
			set cellcmd [lindex $orig_cellcmds [expr $cellcmdnum-1]]
			# remove LinkNextCell and LinkCN commands when the command links
			# to the next cell
			if {[lsearch -regexp [list $cellcmd] "20 01 .. .. .. .. .. 02"] != -1 || \
					([lsearch -regexp [list $cellcmd] "20 07 .. .. .. .. .. .."] != -1 \
						&& [x2d [lindex $cellcmd 7]] == [expr $c+1])} {
				set cellcmd "00 00 00 00 00 00 00 00"
			}
			# If the cell command is not a NOP
			if {$cellcmd!="00 00 00 00 00 00 00 00" && \
					$cellcmdnum<=$orig_numcmds} {
				incr newcmdnum
				# keep cell command
				lappend cellcmds [lindex $orig_cellcmds [expr $cellcmdnum-1]]
				# replace the cell cmd number by the new one in the cell playback table
				set_int cellpbtable [expr (($c-1)*24)+3] $newcmdnum 1
				set modified true
			} else {
				# If cell command is a NOP, replace cell command number by 0,
				# and drop cell command
				set_int cellpbtable [expr (($c-1)*24)+3] 0 1
			}
		}
	}
	# Return the new list of cell commands and the modified cell playback table
	return [list $cellcmds $cellpbtable]
}

# routine called by the remove crap to warn the user if there are several PGCs
# using some or all of the removed cells
proc warn_if_more_pgcs {vts pgc numcells cellpbtable} {
	# Build a list of the start/end VOBUs
	set vobuend 0
	set state 0
	set protectedareas {}
	# for each cell
	for {set c 1} {$c <= $numcells} {incr c} {
		set old_vobuend $vobuend
		if {$::plugins::psl2::cb($c) == $state} continue
		# Get start and end LBAs of the cell from the cell playback table,
		# incremented by the start of VOB LBA.
		set cidx [expr (($c-1)*24)+8]
		set vobustart [get_int cellpbtable $cidx 4]
		incr cidx 12
		set vobuend   [get_int cellpbtable $cidx 4]
		if {$state == 0} {
			set start $vobustart
		} else {
			lappend protectedareas [list $start $old_vobuend]
		}
		set state $::plugins::psl2::cb($c)
	}
	# Add last cell LBA if the last cell is also marked as protected
	if {$state == 1} {
		lappend protectedareas [list $start $vobuend]
	}

	# searches in the other PGCs if there are cells in the same protected areas
	set found 0
	set pgclist ""
	set pgcnums {}
	for {set p 1} {$p <= $::pgcs($vts,numpgcs)} {incr p} {
		set foundinpgc false
		if {$p == $pgc} continue
		for {set c 1} {$c <= $::pgcs($vts,$p,numcells)} {incr c} {
			set cidx [expr (($c-1)*24)+8]
			set vobustart [get_int ::pgcs($vts,$p,cellpbtable) $cidx 4]
			incr cidx 12
			set vobuend   [get_int ::pgcs($vts,$p,cellpbtable) $cidx 4]
			foreach area $protectedareas {
				if {$vobustart >= [lindex $area 0] && $vobuend <= [lindex $area 1]} {
					incr found
					if {! $foundinpgc} {
						set foundinpgc true
						set pgclist "${pgclist}[::main::pgc_label {} pgc $vts $p]\n"
						lappend pgcnums $p
					}
				}
			}
		}
	}
	return [list $pgcnums $pgclist]
}

# patch IFOs and VOBs to change the VOB and cell IDs of the protected cells
# so that VobBlanker can strip them.
proc patch_vcids {vts start_sector end_sector new_vobid new_cellid} {
	if {![info exists ::dvddir]} {bell; return}

	set title "Change ARccOS VOB/Cell ID"

	# patch the IFO

	set modified false
	# patch cell position table is modified PGCs
	for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
		set cellpbtable $::pgcs($vts,$pgc,cellpbtable)
		for {set cell 1} {$cell <= $::pgcs($vts,$pgc,numcells)} {incr cell} {
			set cellpbidx [expr ($cell-1)*24]
			set start_vobu [get_int cellpbtable [expr $cellpbidx+0x08] 4]
			set end_vobu   [get_int cellpbtable [expr $cellpbidx+0x14] 4]
			if {$start_vobu >= $start_sector && $end_vobu <= $end_sector} {
				set modified true
				set cellposidx [expr ($cell-1)*4]
				set_int ::pgcs($vts,$pgc,cellpostable) $cellposidx $new_vobid 2
				incr cellposidx 3
				set_int ::pgcs($vts,$pgc,cellpostable) $cellposidx $new_cellid 1
			}
		}
	}
	if {$modified} {
		# patch VTS_C_ADT table
		set numvobids [get_int ::raw_sectors($vts,VTS_C_ADT) 0 2]
		incr numvobids
		set_int ::raw_sectors($vts,VTS_C_ADT) 0 $numvobids 2

		set_int ::raw_sectors($vts,VTS_C_ADT) 12 [expr $end_sector+1] 4

		set new_entry [string repeat \000 12]
		set_int new_entry 0 $new_vobid  2
		set_int new_entry 2 $new_cellid 1
		set_int new_entry 4 $start_sector 4
		set_int new_entry 8 $end_sector 4
		set ::raw_sectors($vts,VTS_C_ADT) $::raw_sectors($vts,VTS_C_ADT)$new_entry

		set end_byte_of_table [get_int ::raw_sectors($vts,VTS_C_ADT) 4 4]
		incr end_byte_of_table 12
		set_int ::raw_sectors($vts,VTS_C_ADT) 4 $end_byte_of_table 4

		::main::dvd_modified true false
	}

	# patch the VOBs

	show_msg "Patching the VOBs..."

	set ::plugins::psl2::numnavpacks 0
	set ::plugins::psl2::numfixed 0
	set vobnum 1
	set sector 0
	set filename [file join $::dvddir "VTS_[format %02d $vts]_$vobnum.VOB"]
	if {[file exists $filename]} {
		set vobsize [expr [file size $filename] / 2048]
		set fh [::menued::openbinfile $filename r+]
		if {$fh != ""} {
			while {1} {
				if {$sector >= $start_sector && $sector <= $end_sector} {
					set sector [patch_nav_pack_vcid $filename $fh $sector $new_vobid $new_cellid]
				} elseif {$sector < $end_sector} {
					set sector [next_nav_pack $filename $fh $sector]
				} else {
					break
				}
				if {$sector < 0} break
				if {$sector >= $vobsize} {
					close $fh
					set fh ""
					incr vobnum
					set filename [file join $::dvddir "VTS_[format %02d $vts]_$vobnum.VOB"]
					set sector [expr $sector - $vobsize]
					if {[file exists $filename]} {
						set vobsize [expr [file size $filename] / 2048]
						set fh [::menued::openbinfile $filename r+]
						if {$fh == ""} break
					} else {
						break
					}
				}
			}
			if {$fh != ""} {close $fh}
		}
	}
	show_msg "VOB and Cell IDs changed.  $::plugins::psl2::numnavpacks nav packs scanned.  $::plugins::psl2::numfixed nav packs fixed."
	return $::plugins::psl2::numfixed
}

# change VOB and Cell ID in current nav pack, and return next valid nav pack LBA
proc patch_nav_pack_vcid {filename fh sector vobid cellid} {
	# load Nav Pack
	set navpack [::menued::load_nav_pack $filename $fh $sector]

	# if current nav pack is zeroed, find next valid nav pack
	if {$navpack == ""} {
		show_msg "*** Nav Pack at sector $sector is invalid!  Searching for next valid Nav Pack..." -1
		set l [search_next_valid_nav_pack $filename $fh $sector]
		set sector  [lindex $l 0]
		set navpack [lindex $l 1]
		if {$navpack == ""} {
			show_msg "*** No more valid nav pack found!" -1
			tk_messageBox  -title "Change ARccOS VOB/Cell ID" -icon error \
					-message "No more valid nav pack found\nin \"[file tail $filename]\":\nAborting current domain!\nTry to strip or blank the bad VOB cell(s)\nwith IfoEdit."
			return -11
		}
		show_msg "A new valid nav pack was found at sector $sector.  Continuing..." -1
	}
	incr ::plugins::psl2::numnavpacks

	# fix VOB and Cell ID
	set cur_vobid  [get_int navpack 0x041F 2]
	set cur_cellid [get_int navpack 0x0422 1]
	if {$cur_vobid != $vobid || $cur_cellid != $cellid} {
		set_int navpack 0x041F $vobid  2
		set_int navpack 0x0422 $cellid 1
		if {! [::menued::save_nav_pack $filename $fh $sector $navpack]} {return -12}
		incr ::plugins::psl2::numfixed
	}

	# Compute next nav pack LBA
	set end_of_vobu [get_int navpack 0x040F 4]
	if {$end_of_vobu == 0} {return -1}
	return [expr $sector + $end_of_vobu + 1]
}

# return next valid nav pack LBA
proc next_nav_pack {filename fh sector} {
	set navpack [::menued::load_nav_pack $filename $fh $sector]
	if {$navpack == ""} {
		show_msg "*** Nav Pack at sector $sector is invalid!  Searching for next valid Nav Pack..." -1
		set l [search_next_valid_nav_pack $filename $fh $sector]
		set sector  [lindex $l 0]
		set navpack [lindex $l 1]
		if {$navpack == ""} {
			show_msg "*** No more valid nav pack found!" -1
			tk_messageBox  -title "Change ARccOS VOB/Cell ID" -icon error \
					-message "No more valid nav pack found\nin \"[file tail $filename]\":\nAborting current domain!\nTry to strip or blank the bad VOB cell(s)\nwith IfoEdit."
			return -11
		}
		show_msg "A new valid nav pack was found at sector $sector.  Continuing..." -1
	}
	incr ::plugins::psl2::numnavpacks
	set end_of_vobu [get_int navpack 0x040F 4]
	if {$end_of_vobu == 0} {return -1}
	return [expr $sector + $end_of_vobu + 1]
}

# skip zeroed sectors and return next valid nav pack LBA and nav pack contents
proc search_next_valid_nav_pack {filename fh sector} {
	# searches sector by sector for a valid nav pack
	set vobsize [expr [file size $filename] / 2048]
	incr sector
	if {$sector > $vobsize} {return [list $sector ""]}
	set navpack [::menued::load_nav_pack $filename $fh $sector]
	while {$navpack == ""} {
		incr sector
		if {$sector > $vobsize} {return [list $sector ""]}
		set navpack [::menued::load_nav_pack $filename $fh $sector]
	}
	return [list $sector $navpack]
}

# If the PGC is a fake, unreferenced title, remove it
proc remove_if_fake_pgc {vts pgc quiet} {
	if {$pgc == $::pgcs($vts,numpgcs) && \
			[info exists ::pgcs($vts,$pgc,Title)] &&
			$::pgcs($vts,$pgc,Title) == [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]} {
		::utils::build_crossrefs $vts
		if {$::crossrefs(pgcs,$vts,$pgc,numrefs) == 0} {
			if {$quiet} {
				set r "yes"
			} else {
				set txt "The current title is unreferenced and is the last\n"
				set txt "${txt}title number in the DVD.\n"
				set txt "${txt}The best way to clean it is to delete it!\n\n"
				set txt "${txt}Do you want to delete title $::pgcs($vts,$pgc,Title)? (Recommended)"
				set r [tk_messageBox -title "PSL2 plugin" -icon question -message $txt \
						-type yesno -default yes]
			}
			if {$r == "yes"} {
				set warningexists [info exists ::warning_delete_last_pgc]
				set ::warning_delete_last_pgc true
				::utils::delete_last_pgc
				if {!$warningexists} {
					unset ::warning_delete_last_pgc
				}
				lappend ::plugins::psl2::cleaned_titles_list [list $vts $pgc false]
				return "true"
			}
		}
	}
	return false
}

#################### Generate PSL data main function ####################

# generate PSL output.
proc generate_psl2_data {vts pgc domainlba entry_cell remap_cell remap_cellnum quiet} {
	set numcells    $::pgcs($vts,$pgc,numcells)
	set cellpbtable $::pgcs($vts,$pgc,cellpbtable)

	# Build output data
	set out1 ""
	set out6 ""
	set state 0
	set vobuend 0
	set protectedareas {}

	# for each cell
	for {set c 1} {$c <= $numcells} {incr c} {
		set old_vobuend $vobuend
		# if cell is marked as protected...
		if {$::plugins::psl2::cb($c)} {
			# Add line 6:*,*,*,2,1 anyway in output.
			set out6 "$out6\n6:$vts,$pgc,$c,2,1"
			# If the remap cell option is checked...
			if {$remap_cell} {
				#  add also the 6:*,*,*,3,1 and 6:*,*,*,4,<remapping_cell-1> lines
				set out6 "$out6\n6:$vts,$pgc,$c,3,1"
				set out6 "$out6\n6:$vts,$pgc,$c,4,$remap_cellnum"
			}
		}
		# Get start and end LBAs of the cell from the cell playback table,
		# incremented by the start of VOB LBA.
		set cidx [expr (($c-1)*24)+8]
		set vobustart [expr [get_int cellpbtable $cidx 4] + $domainlba]
		incr cidx 12
		set vobuend   [expr [get_int cellpbtable $cidx 4] + $domainlba]

		# Keep line with first LBA to end LBA, but only for block of contiguous cells
		if {$::plugins::psl2::cb($c) == $state} continue
		if {$state == 0} {
			set start $vobustart
		} else {
			set out1 "$out1\n1:$start,$old_vobuend"
			lappend protectedareas [list $start $old_vobuend]
		}
		set state $::plugins::psl2::cb($c)
	}
	# Add last cell LBA if the last cell is also marked as protected
	if {$state == 1} {
		set out1 "$out1\n1:$start,$vobuend"
		lappend protectedareas [list $start $vobuend]
	}

	# searches in the other PGCs if there are cells in the same protected areas
	set found 0
	set pgclist ""
	for {set p 1} {$p <= $::pgcs($vts,numpgcs)} {incr p} {
		set foundinpgc false
		if {$p == $pgc} continue
		for {set c 1} {$c <= $::pgcs($vts,$p,numcells)} {incr c} {
			set cidx [expr (($c-1)*24)+8]
			set vobustart [expr [get_int ::pgcs($vts,$p,cellpbtable) $cidx 4] + $domainlba]
			incr cidx 12
			set vobuend   [expr [get_int ::pgcs($vts,$p,cellpbtable) $cidx 4] + $domainlba]
			foreach area $protectedareas {
				if {$vobustart >= [lindex $area 0] && $vobuend <= [lindex $area 1]} {
					incr found
					if {! $foundinpgc} {
						set foundinpgc true
						set pgclist "${pgclist}[::main::pgc_label {} pgc $vts $p]\n"
					}
					# Add line 6:*,*,*,2,1 anyway in output.
					set out6 "$out6\n6:$vts,$p,$c,2,1"
					# If the remap cell option is checked...
					if {$remap_cell} {
						#  add also the 6:*,*,*,3,1 and 6:*,*,*,4,<remapping_cell-1> lines
						set out6 "$out6\n6:$vts,$p,$c,3,1"
						set out6 "$out6\n6:$vts,$p,$c,4,$remap_cellnum"
					}
				}
			}
		}
	}
	if {$found > 0 && !$quiet} {
		set txt "Found $found cells in the same protected area in the following PGC(s):\n$pgclist\n"
		set txt "${txt}The commands to process those cells will been added to the PSL file.\n\n"
		set txt "${txt}When doing the Remove Crap pass, don't forget to process also those PGC(s)."
		tk_messageBox -title "PSL2 plugin" -icon warning -message $txt
	}

	# Build full output by concatenating parts together, and strip multiple blank lines
	return "$out1\n$out6"
}

####################### main remove crap function ######################

# remove crap.
proc remove_crap {vts pgc entry_cell quiet} {
	set cellpbtable $::pgcs($vts,$pgc,cellpbtable)
	set programmap  $::pgcs($vts,$pgc,programmap)
	set numcells    $::pgcs($vts,$pgc,numcells)

	# Retrieve cell commands (as Tcl List of hex codes)
	set cellcommands [get_type_cmds cell ::pgcs($vts,$pgc)]

	set need_fix_pointers false	;# for VobBlanker

	######## trick used as a workaround for the same reused V/CID ######
	######## problem (cf. Seinfeld season 4, Disc 4, VTS 3.)      ######

	# used only to display the "VOB modified: Save DVD" warning
	set num_nav_packs_fixed 0

	# verify if the VOB/Cell IDs of the removed cells are not shared by some good cells
	for {set c 1} {$c <= $numcells} {incr c} {
		if {$::plugins::psl2::cb($c)} {
			set cellposidx [expr ($c-1)*4]
			set vid [get_int ::pgcs($vts,$pgc,cellpostable) $cellposidx 2]
			incr cellposidx 3
			set cid [get_int ::pgcs($vts,$pgc,cellpostable) $cellposidx 1]
			set vcid_unused($vid,$cid) 1
			set last_removed_cell $c
		}
	}
	set vcid_reused false
	for {set c 1} {$c <= $numcells} {incr c} {
		if {! $::plugins::psl2::cb($c)} {
			set cellposidx [expr ($c-1)*4]
			set vid [get_int ::pgcs($vts,$pgc,cellpostable) $cellposidx 2]
			incr cellposidx 3
			set cid [get_int ::pgcs($vts,$pgc,cellpostable) $cellposidx 1]
			if {[info exists vcid_unused($vid,$cid)]} {
				set vcid_reused true
				break
			}
		}
	}
	# if there are some V/C IDs reused in the played cells...
	if {$vcid_reused} {
		# we must be sure that there are no cells with a reused V/CID left unchecked
		# before the last checked cell!
		for {set c 1} {$c <= $numcells} {incr c} {
			if {! $::plugins::psl2::cb($c)} {
				set cellposidx [expr ($c-1)*4]
				set vid [get_int ::pgcs($vts,$pgc,cellpostable) $cellposidx 2]
				incr cellposidx 3
				set cid [get_int ::pgcs($vts,$pgc,cellpostable) $cellposidx 1]
				if {[info exists vcid_unused($vid,$cid)] && $c < $last_removed_cell} {
					if {! $quiet} {
						set msg "The title you are processing is somewhat special.\n"
						set msg "${msg}The same VOB/Cell ID is used in several cells.\n"
						set msg "${msg}For this reason, it is necessary to select ALL tiny\n"
						set msg "${msg}ARccOS cells (starting at cell 1 and ending at the\n"
						set msg "${msg}last unused cell.)\n\n"
						set msg "${msg}I cannot continue with your current settings.\n\n"
						set msg "${msg}Please retry.\n"
						set msg "${msg}Use the preview to locate the first cell with video\n"
						set msg "${msg}contents, and mark ALL cells before that one."
						tk_messageBox -title "PSL2 plugin" -icon error -message $msg
					}
					return ""
				}
			}
		}
		unset vcid_unused

		# we must also be sure that all checked cells are contiguous.
		# and we must compute the start and end sector of the cells collection
		set invcid 0
		set start_sector 999999999999
		set end_sector 0
		set vobuend_set false
		for {set c 1} {$c <= $numcells} {incr c} {
			set cidx [expr (($c-1)*24)+8]
			if {$::plugins::psl2::cb($c)} {
				set vobustart [get_int cellpbtable $cidx 4]
				if {$vobustart < $start_sector} {
					set start_sector $vobustart
				}
				# Doesn't work if the ARccOS cells have been remapped
				incr cidx 12
				set vobuend [get_int cellpbtable $cidx 4]
				if {$vobuend > $end_sector} {
					set end_sector $vobuend
				}
				if {$invcid == 0} {
					set invcid 1
				} elseif {$invcid == 2} {
					set msg "The title you are processing is somewhat special.\n"
					set msg "${msg}The same VOB/Cell ID is used in several cells.\n"
					set msg "${msg}For this reason, you cannot leave gaps between\n"
					set msg "${msg}the checked cells.\n\n"
					set msg "${msg}Please retry."
					tk_messageBox -title "PSL2 plugin" -icon error -message $msg
					return ""
				}
			} else {
				# compute VOBU End from first kept cell
				set tmp [get_int cellpbtable $cidx 4]
				incr tmp -1
				if {! $vobuend_set} {
					set vobuend_set true
					if {$tmp > $end_sector} {
						set end_sector $tmp
					}
				}
				if {$invcid == 1} {
					set invcid 2
				}
			}
		}
		# search for the first unused VOB ID
		set first_unused_vid 0
		set adtidx 8
		while {$adtidx < [string length $::raw_sectors($vts,VTS_C_ADT)]} {
			set vid [get_int ::raw_sectors($vts,VTS_C_ADT) $adtidx 2]
			if {$vid > $first_unused_vid} {
				set first_unused_vid $vid
			}
			incr adtidx 12
		}
		incr first_unused_vid
		show_msg "Patching IFOs and VOBs..."
		set need_fix_pointers true
		# skip Bad Nav Pack error messages (we will certainly have some of them!)
		set skip [info exists ::menued::load_nav_pack_error_skip]
		set ::menued::load_nav_pack_error_skip 1
		# Patch the IFOs and VOBs to assign a new VOB/Cell ID to the removed cells
		set num_nav_packs_fixed [patch_vcids $vts $start_sector $end_sector $first_unused_vid 1]
		# restore skip Bad Nav Pack error messages state
		if {!$skip} {
			unset ::menued::load_nav_pack_error_skip
		}
	}
	# we must also concatenate all remaining cells with the same V/CID in one single cell
	show_msg "Concatenating cells with same Vob/Cell ID in one single cell..."
	set cells_reused_exists false
	for {set c 1} {$c <= $numcells} {incr c} {
		if {! $::plugins::psl2::cb($c)} {
			set cellposidx [expr ($c-1)*4]
			set vid [get_int ::pgcs($vts,$pgc,cellpostable) $cellposidx 2]
			incr cellposidx 3
			set cid [get_int ::pgcs($vts,$pgc,cellpostable) $cellposidx 1]
			set cellpbidx [expr {(($c - 1) * 24)+4}]
			if {! [info exists vcid_reused_cell($vid,$cid)]} {
				set vcid_reused_cell($vid,$cid) 1
				set vcid_reused_cell($vid,$cid,kept) $c
				set vcid_reused_cell($vid,$cid,timing) \
						[string range $cellpbtable $cellpbidx [expr $cellpbidx+3]]
			} else {
				set cells_reused_exists true
				incr vcid_reused_cell($vid,$cid)
				set vcid_reused_cell($vid,$cid,timing) [::pgced::AddBCDtime $vcid_reused_cell($vid,$cid,timing) \
						[string range $cellpbtable $cellpbidx [expr $cellpbidx+3]]]
				incr cellpbidx 12
				set vcid_reused_cell($vid,$cid,last_vobu_start) [get_int cellpbtable $cellpbidx 4]
				incr cellpbidx 4
				set vcid_reused_cell($vid,$cid,last_vobu_end)   [get_int cellpbtable $cellpbidx 4]
			}
		}
	}
	# Fixes the VTS_C_ADT table
	show_msg "Fixing VTS_C_ADT table..."
	set adtidx 8
	while {$adtidx < [string length $::raw_sectors($vts,VTS_C_ADT)]} {
		set vid [get_int ::raw_sectors($vts,VTS_C_ADT) $adtidx 2]
		set cid [get_int ::raw_sectors($vts,VTS_C_ADT) [expr $adtidx+2] 1]
		if {[info exists vcid_reused_cell($vid,$cid,last_vobu_end)]} {
			if {$vcid_reused_cell($vid,$cid,last_vobu_end) >
					[get_int ::raw_sectors($vts,VTS_C_ADT) [expr $adtidx+8] 4]} {
				set_int ::raw_sectors($vts,VTS_C_ADT) [expr $adtidx+8] \
						$vcid_reused_cell($vid,$cid,last_vobu_end) 4
			}
		}
		incr adtidx 12
	}

	# Fixes the VTS_VOBU_ADMAP table.
	# Remove the fake VOBU_ADMAP pointers pointing at the end of the VOBs
	show_msg "Fixing VTS_VOBU_ADMAP table..."
	set table $::raw_sectors($vts,VTS_VOBU_ADMAP)
	set len [string length $table]
	set last_vobu [expr 0xFFFFFFFF]
	set numremoved 0
	for {set idx [expr $len-4]} {$idx>0} {incr idx -4} {
		set vobu [get_int table $idx 4]
		if {$vobu>$last_vobu} {
			set table [string range $table 0 [expr $idx-1]][string range $table [expr $idx+4] end]
			incr numremoved
		} else {
			set last_vobu $vobu
		}
	}
	if {$numremoved > 0} {
		set endbyte [expr [get_int table 0 4] - ($numremoved*4)]
		set_int table 0 $endbyte 4
		set ::raw_sectors($vts,VTS_VOBU_ADMAP) $table
	}

	# If a start sector of a cell referenced in the VTS_C_ADT table
	# is missing in the VTS_VOBU_ADMAP table, DVD Shrink reports an
	# Illegal Navigation error.  So, we must search for the missing
	# entries and add them at the right place. (This error is fixed
	# anyway by the VobBlanker pass,  but we want to be able to use
	# the DVD w/o this long pass!)
	set vobutablelen [string length $::raw_sectors($vts,VTS_VOBU_ADMAP)]
	set adtidx 12
	while {$adtidx < [string length $::raw_sectors($vts,VTS_C_ADT)]} {
		set entrysec [get_int ::raw_sectors($vts,VTS_C_ADT) $adtidx 4]
		set entrysecbin [string range $::raw_sectors($vts,VTS_C_ADT) $adtidx [expr $adtidx+3]]
		incr adtidx 12
		# to speedup the process, searches for the entry in binary
		set idx 3
		set found false
		while {$idx < $vobutablelen} {
			set idx [string first $entrysecbin $::raw_sectors($vts,VTS_VOBU_ADMAP)]
			if {$idx == -1} break
			if {[expr $idx % 4] == 0} {
				set found true
				break
			}
		}
		# if entry is not present, add it
		if {!$found} {
			set idx 4
			while {$idx < $vobutablelen} {
				if {[get_int ::raw_sectors($vts,VTS_VOBU_ADMAP) $idx 4] > $entrysec} {
					set ::raw_sectors($vts,VTS_VOBU_ADMAP) \
						[string range $::raw_sectors($vts,VTS_VOBU_ADMAP) 0 [expr $idx-1]]$entrysecbin[string range $::raw_sectors($vts,VTS_VOBU_ADMAP) $idx end]
					break
				}
				incr idx 4
			}
		}
	}

	# Remove fake cells
	show_msg "Removing fake cells..."
	if {$cells_reused_exists} {
		set totalcells $numcells
		for {set c $numcells} {$c > 0} {incr c -1} {
			# if cell is not an arccos cell and V/CID is reused
			if {! $::plugins::psl2::cb($c)} {
				set cellposidx [expr ($c-1)*4]
				set vid [get_int ::pgcs($vts,$pgc,cellpostable) $cellposidx 2]
				incr cellposidx 3
				set cid [get_int ::pgcs($vts,$pgc,cellpostable) $cellposidx 1]
				set cellpbidx [expr {($c - 1) * 24}]
				if {$vcid_reused_cell($vid,$cid) > 1} {
					if {$vcid_reused_cell($vid,$cid,kept) == $c} {
						set cellpbtable [string range $cellpbtable 0 [expr $cellpbidx+3]]$vcid_reused_cell($vid,$cid,timing)[string range $cellpbtable [expr $cellpbidx+8] end]
						incr cellpbidx 16
						set_int cellpbtable $cellpbidx $vcid_reused_cell($vid,$cid,last_vobu_start) 4
						incr cellpbidx 4
						set_int cellpbtable $cellpbidx $vcid_reused_cell($vid,$cid,last_vobu_end) 4
					} else {
						incr totalcells -1
						# remove cell from cell playback table
						set cellpbtable "[string range $cellpbtable 0 [expr {$cellpbidx-1}]][string range $cellpbtable [expr {$cellpbidx+24}] end]"
						# and from cell position table
						set cellposidx [expr {($c - 1) * 4}]
						set ::pgcs($vts,$pgc,cellpostable) "[string range $::pgcs($vts,$pgc,cellpostable) 0 [expr {$cellposidx-1}]][string range $::pgcs($vts,$pgc,cellpostable) [expr {$cellposidx+4}] end]"

						# fix program map table
						for {set i 0} {$i < [string length $::pgcs($vts,$pgc,programmap)]} {incr i} {
							set pc [get_int ::pgcs($vts,$pgc,programmap) $i 1]
							if {$pc>$c && $pc!=0} {
								set_int ::pgcs($vts,$pgc,programmap) $i [expr $pc-1] 1
							}
						}

						# Fix LinkCN commands in the pre/post and cell commands to jump to the new cell
						foreach i [lsearch -regexp -all $::pgcs($vts,$pgc) "20 .7 .. .. .. .. .. .."] {
							set cmd [lindex $::pgcs($vts,$pgc) $i]
							set target [x2d [lindex $cmd 7]]
							if {$target >$c} {
								lset cmd 7 [format %02X [expr $target-1]]
								lset ::pgcs($vts,$pgc) $i $cmd
							}
						}
					}
				}
			}
		}
		# fix PgcEdit's internal variable holding the number of cells of the current PGC
		set ::pgcs($vts,$pgc,numcells) $totalcells
		set numcells $totalcells
	}
	########  End of trick  ########

	# clean the cell commands
	show_msg "Cleaning cell commands (pass 1)..."
	set cleaned [clean_cell_commands_sub pgc $vts $pgc \
			[string toupper [get_type_cmds pre ::pgcs($vts,$pgc)]] \
			[string toupper $cellcommands] $numcells $cellpbtable]
	set_type_cmds cell ::pgcs($vts,$pgc) [lindex $cleaned 0]
	# retrieve the cell playback table (as raw binary)
	set cellpbtable [lindex $cleaned 1]


	# Clear the seamless flag and set the STC discontinuity flag on cells
	# AFTER the marked ones.
	set lastremoved true
	for {set c 1} {$c <= $numcells} {incr c} {
		if {$::plugins::psl2::cb($c)} {
			set lastremoved true
		} else {
			if {$lastremoved} {
				set cellpbidx [expr {($c - 1) * 24}]
				set_int cellpbtable [expr $cellpbidx] [expr [get_int cellpbtable [expr $cellpbidx] 1] & 0xF7 | 0x02] 1
				set lastremoved false
			}
		}
	}

	# Remove marked cells
	show_msg "Removing marked cells..."
	set totalcells $numcells
	for {set c $numcells} {$c > 0} {incr c -1} {
		set cellpbidx [expr {($c - 1) * 24}]
		# if cell is marked for removal
		if {$::plugins::psl2::cb($c)} {
			incr totalcells -1
			# remove cell from cell playback table
			set cellpbtable "[string range $cellpbtable 0 [expr {$cellpbidx-1}]][string range $cellpbtable [expr {$cellpbidx+24}] end]"
			# and from cell position table
			set cellposidx [expr {($c - 1) * 4}]
			set ::pgcs($vts,$pgc,cellpostable) "[string range $::pgcs($vts,$pgc,cellpostable) 0 [expr {$cellposidx-1}]][string range $::pgcs($vts,$pgc,cellpostable) [expr {$cellposidx+4}] end]"

			# fix program map table
			for {set i 0} {$i < [string length $::pgcs($vts,$pgc,programmap)]} {incr i} {
				set pc [get_int ::pgcs($vts,$pgc,programmap) $i 1]
				if {$pc>$c && $pc!=0} {
					set_int ::pgcs($vts,$pgc,programmap) $i [expr $pc-1] 1
				}
			}

			# Fix LinkCN commands in the pre/post and cell commands to jump to the new cell
			foreach i [lsearch -regexp -all $::pgcs($vts,$pgc) "20 .7 .. .. .. .. .. .."] {
				set cmd [lindex $::pgcs($vts,$pgc) $i]
				set target [x2d [lindex $cmd 7]]
				if {$target >$c} {
					lset cmd 7 [format %02X [expr $target-1]]
					lset ::pgcs($vts,$pgc) $i $cmd
				}
			}
		} else {
			# Remove the restricted bit on remaining cells
			set_int cellpbtable [expr $cellpbidx+1] [expr [get_int cellpbtable [expr $cellpbidx+1] 1] & 0xDF] 1
		}
	}
	# fix PgcEdit's internal variable holding the number of cells of the current PGC
	set ::pgcs($vts,$pgc,numcells) $totalcells

	# Clean the cell commands again (since we have deleted cells, most of the cell commands will be deleted)
	show_msg "Cleaning cell commands (pass 2)..."
	set cleaned [clean_cell_commands_sub pgc $vts $pgc \
			[string toupper [get_type_cmds pre ::pgcs($vts,$pgc)]] \
			[string toupper [get_type_cmds cell ::pgcs($vts,$pgc)]] $totalcells $cellpbtable]
	set_type_cmds cell ::pgcs($vts,$pgc) [lindex $cleaned 0]
	set ::pgcs($vts,$pgc,cellpbtable) [lindex $cleaned 1]

	# fix PGC total time
	fix_pgc_time $vts $pgc $totalcells $cellpbtable
	# refresh GUI.  Done.
	refresh_current_pgc
	::main::init_tree_menu_no_move
	::main::display_pgc pgc $vts $pgc
	::main::dvd_modified true false
	update idletasks
	show_msg ""

	# remember which PGC to post-process with VobBlanker
	lappend ::plugins::psl2::cleaned_titles_list [list $vts $pgc $need_fix_pointers]

	return $num_nav_packs_fixed
}

# Remove the last unreferenced titles
proc remove_unref_titles {} {
	if {![info exists ::dvddir]} {bell; return}

	# generate list of protected PGCs
	# Abort if no ARccOS titles
	set protectedtitleslist [find_protected_titles false false]
	if {[llength $protectedtitleslist] == 0} {
		tk_messageBox -title "PSL2 plugin" -icon error -message "Cannot find any protected title!"
		return
	}

	# extract the list of unreferenced titles from the list of protected titles
	# and build the confirmation message
	set unreftitleslist ""
	set txt "The following Titles are unreferenced,\nand are at the end of a Title domain.\nThey are plausibly fake ARccOS titles.\nYou may probably remove them safely.\n\n"
	set numentries 0
	foreach protectedtitle $protectedtitleslist {
		if {[lindex $protectedtitle 2]} {
			lappend unreftitleslist $protectedtitle
			if {$numentries < 30} {
				incr numentries
				set txt "${txt}[::main::pgc_label $::config(global_font) pgc \
						[lindex $protectedtitle 0] [lindex $protectedtitle 1]]\n"
			}
		}
	}
	if {$numentries != [llength $unreftitleslist]} {
		set txt "${txt}...  The list of unreferenced titles is too long!\n"
		set txt "${txt}Total: [llength $unreftitleslist] unreferenced titles to remove.\n"
	}
	set txt "${txt}\nRemove those unreferenced titles?"

	# abort if no unref titles
	if {[llength $unreftitleslist] == 0} {
		tk_messageBox -title "PSL2 plugin" -icon info -message "Cannot find any unreferenced title\nat the end of a title domain!"
		return
	}

	# display first title to remove and ask confirmation
	set vts [lindex [lindex $unreftitleslist 0] 0]
	set pgc [lindex [lindex $unreftitleslist 0] 1]
	::main::display_pgc pgc $vts $pgc
	set r [tk_messageBox -title "PSL2 plugin" -icon question -message $txt \
			-type yesno -default yes]
	if {$r == "no"} return

	# remove the unreferenced titles
	foreach t $unreftitleslist {
		set vts [lindex $t 0]
		set pgc [lindex $t 1]
		# select and display the PGC in PgcEdit
		::main::display_pgc pgc $vts $pgc
		after 100 {set ::plugins::psl2::pause 1}
		tkwait variable ::plugins::psl2::pause

		# Be sure that the current PGC is the last one of the domain!
		if {$pgc == $::pgcs($vts,numpgcs)} {
			# remove the title
			set warningexists [info exists ::warning_delete_last_pgc]
			set ::warning_delete_last_pgc true
			::utils::delete_last_pgc
			if {!$warningexists} {
				unset ::warning_delete_last_pgc
			}
			# remember the VTS to clean for FixVTS or VobBlanker final cleanup.
			lappend ::plugins::psl2::cleaned_titles_list [list $vts $pgc false]
		}
	}
	unset -nocomplain ::plugins::psl2::pause
}


##### autostart?
catch {source [file join $::config(config_directory) "psl2_plugin.cfg"]}
if {![info exists ::plugins::psl2::autostart]} {
	if {[info exists ::config(auto_open_new_dvd)] && $::config(auto_open_new_dvd)} {
		set ::plugins::psl2::autostart false
	} else {
		set ::plugins::psl2::autostart true
	}
}
if {$::plugins::psl2::autostart} {
	after idle {::plugins::psl2::wizard}
}


}	;# end of namespace ::plugins::psl2
